//--------------------------------------------------------------------------------------
// Copyright 2011,2012,2013 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//--------------------------------------------------------------------------------------
#include "AVSM.h"
#include "CPUTRenderTarget.h"
#include "CPUTTextureDX11.h"

#include ".\AVSMTechnique\ShaderUtils.h"
#include ".\AVSMTechnique\ListTexture.h"
#include ".\AVSMTechnique\Texture2D.h"
// 'slice' debug view
#include "AVSMDebugView.h"

#include "DefaultShader.h"

// Intel extensions detector
#include ".\IGFXExtensions\IGFXExtensionsHelper.h"
#include <memory> // for shared_ptr<>

extern char *gpDefaultShaderSource;

const UINT SHADOW_WIDTH_HEIGHT = 2048;

// Application entry point.  Execution begins here.
//-----------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Prevent unused parameter compiler warnings
    UNREFERENCED_PARAMETER(hInstance);
    UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(nCmdShow);

#ifdef DEBUG
    // tell VS to report leaks at any exit of the program
    _CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif
    CPUTResult result=CPUT_SUCCESS;
    int returnCode=0;

    // create an instance of my sample
    AVSMSample *pSample = new AVSMSample();
    
    // We make the assumption we are running from the executable's dir in
    // the CPUT SampleStart directory or it won't be able to use the relative paths to find the default
    // resources    
    cString ResourceDirectory;
    CPUTOSServices::GetOSServices()->GetExecutableDirectory(&ResourceDirectory);
    ResourceDirectory.append(_L(".\\CPUT\\resources\\"));
    
    // Initialize the system and give it the base CPUT resource directory (location of GUI images/etc)
    // For now, we assume it's a relative directory from the executable directory.  Might make that resource
    // directory location an env variable/hardcoded later
    pSample->CPUTInitialize(ResourceDirectory); 

    // window and device parameters
    CPUTWindowCreationParams params;
    params.deviceParams.refreshRate         = 0; // 60 if you want to frame-lock it
    params.deviceParams.swapChainBufferCount= 1;
    params.deviceParams.swapChainFormat     = DXGI_FORMAT_R8G8B8A8_UNORM_SRGB;
    params.deviceParams.swapChainUsage      = DXGI_USAGE_RENDER_TARGET_OUTPUT | DXGI_USAGE_SHADER_INPUT;

    // parse out the parameter settings or reset them to defaults if not specified
    cString AssetFilename;
    cString CommandLine(lpCmdLine);
    pSample->CPUTParseCommandLine(CommandLine, &params, &AssetFilename);       

    // create the window and device context
    result = pSample->CPUTCreateWindowAndContext(_L("AVSM"), params);
    ASSERT( CPUTSUCCESS(result), _L("CPUT Error creating window and context.") );
    
    // start the main message loop
    returnCode = pSample->CPUTMessageLoop();

	pSample->DeviceShutdown();

    // cleanup resources
    SAFE_DELETE(pSample);

    return returnCode;
}

// Handle creation events
//-----------------------------------------------------------------------------
void AVSMSample::Create()
{   
    // Detect and report Intel extensions on this system
    HRESULT hr = IGFX::Init( this->mpD3dDevice );
	if ( FAILED(hr) )
	{
		printf("Failed hardware detection initialization: incorrect vendor or device.\n\n");
	}
    // detect the available extensions
    IGFX::Extensions extensions = IGFX::getAvailableExtensions( this->mpD3dDevice );

    mAVSMExtensionAvailable = extensions.PixelShaderOrdering;  
    
    // report and disable the AVSM extension method if the hardware/driver does not support Pixel Shader Ordering feature
    if ( !extensions.PixelShaderOrdering ) 
    {
        CPUTOSServices::GetOSServices()->OpenMessageBox(_L("Pixel Shader Ordering feature not found"), _L("Your hardware or graphics driver does not support the pixel shader ordering feature.  Please update your driver or run on a system that supports the required feature to see that option.  Only the pure DirectX version will be enabled.\n\nSee this link for information and a video about Pixel Shader Ordering in this sample:\nhttp://software.intel.com/gamecode"));      
    }
    
    // Create the GUI controls
    CPUTAssetLibrary *pAssetLibrary = CPUTAssetLibrary::GetAssetLibrary();
    CPUTGuiControllerDX11 *pGUI = CPUTGetGuiController();
    pGUI->CreateButton(_L("Fullscreen"), ID_FULLSCREEN_BUTTON, ID_MAIN_PANEL, &mpFullScreen);

    pGUI->CreateText(_L(" "), ID_IGNORE_CONTROL_ID, ID_MAIN_PANEL);
    pGUI->CreateText(_L("Volume Shadowing Method"), ID_IGNORE_CONTROL_ID, ID_MAIN_PANEL);
    pGUI->CreateDropdown(_L("Disabled"), ID_VOLUME_SHADOWING_METHOD, ID_MAIN_PANEL, &mpVolumeShadowingMethod);
    mpVolumeShadowingMethod->AddSelectionItem(_L("DirectX11 Method"), true);
    if(true == mAVSMExtensionAvailable)
    {
        mpVolumeShadowingMethod->AddSelectionItem(_L("Pixel Shader Ordering"), true);
    }

    pGUI->CreateText(_L(" "), ID_IGNORE_CONTROL_ID, ID_MAIN_PANEL);
    pGUI->CreateText(_L("Shadow Texture Dimension"), ID_IGNORE_CONTROL_ID, ID_MAIN_PANEL);
    pGUI->CreateDropdown(_L("512x512"), ID_SHADOW_TEXTURE_DIMENSION, ID_MAIN_PANEL, &mpShadowTextureDimension);
    mpShadowTextureDimension->AddSelectionItem(_L("256x256"), true);
    mpShadowTextureDimension->AddSelectionItem(_L("128x128"));

    pGUI->CreateText(_L("AVSM Node Count"), ID_IGNORE_CONTROL_ID, ID_MAIN_PANEL);
    pGUI->CreateDropdown(_L("8 nodes per sample"), ID_AVSM_NODE_COUNT, ID_MAIN_PANEL, &mpAVSMNodeCount);
    mpAVSMNodeCount->AddSelectionItem(_L("4 nodes per sample"));

    pGUI->CreateText(_L("\t "), ID_IGNORE_CONTROL_ID, ID_MAIN_PANEL);
    pGUI->CreateText(_L("\t "), ID_IGNORE_CONTROL_ID, ID_MAIN_PANEL);
    pGUI->CreateSlider(_L("Particle Size"), ID_PARTICLE_SIZE, ID_MAIN_PANEL, &mpParticleSize);
	mpParticleSize->SetScale(5, 50, 20);
	mpParticleSize->SetValue(25);

    pGUI->CreateText(_L("Pixel shader ordering feature not supported on this system.  Option disabled."), ID_IGNORE_CONTROL_ID, ID_MAIN_PANEL, &mpExtensionWarning);
    mpExtensionWarning->SetAutoArranged(false);
    int x,y, Winwidth,Winheight;
    CPUTOSServices::GetOSServices()->GetClientDimensions(&x,&y, &Winwidth, &Winheight);
    mpExtensionWarning->GetDimensions(x,y);
    mpExtensionWarning->SetPosition(0,Winheight-y-2);
    CPUTColor4 color;
    color.r=1.0f; color.g=0.0f; color.b=0.0f; color.a=1.0f;
    mpExtensionWarning->SetColor(color);
    if(mAVSMExtensionAvailable)
    {
        mpExtensionWarning->SetVisibility(false);
    }
    
    pGUI->CreateSlider(_L("Particle Opacity"), ID_PARTICLE_OPACITY, ID_MAIN_PANEL, &mpParticleOpacity);
	mpParticleOpacity->SetScale(5, 100, 20);
	mpParticleOpacity->SetValue(50);

    pGUI->CreateCheckbox(_L("Pause Particles"), ID_PAUSE_PARTICLES, ID_MAIN_PANEL, &mpPauseParticles);
#ifdef AVSM_METRICS
    pGUI->CreateCheckbox(_L("Enable Stats"), ID_ENABLE_STATS, ID_MAIN_PANEL, &mpEnableStats);
#endif
	// Extended Vertex shader functionality
    pGUI->CreateCheckbox(_L("WireFrame Particles"), ID_WIREFRAME_PARTICLES, ID_MAIN_PANEL, &mpWireFrameParticles);
    pGUI->CreateCheckbox(_L("Tessellated VS-based particle AVSM"), ID_VERTEX_SHADE_PARTICLES, ID_MAIN_PANEL, &mpVertexShadeParticles);
    mpVertexShadeParticles->SetCheckboxState( CPUT_CHECKBOX_CHECKED );

    pGUI->CreateSlider(_L("Particle Tessellation"), ID_PARTICLETESSELLATION, ID_MAIN_PANEL, &mpParticleTessellation);
	mpParticleTessellation->SetScale(1, 30, 31);
	mpParticleTessellation->SetValue(5);

    pGUI->CreateSlider(_L("Light longitude"), ID_LIGHT_ORBIT_LONG_SLIDER, ID_MAIN_PANEL, &mpLightLongitudeSlider);
    mpLightLongitudeSlider->SetScale( -180.0f, 180.0f, 361 );
    mpLightLongitudeSlider->SetValue( 29.0f );
    pGUI->CreateSlider(_L("Light latitude"), ID_LIGHT_ORBIT_LAT_SLIDER, ID_MAIN_PANEL, &mpLightLatitudeSlider);
    mpLightLatitudeSlider->SetScale( 10.0f, 90.0f, 81 );
    mpLightLatitudeSlider->SetValue( 48.0f );

    pGUI->CreateCheckbox(_L("Show debug view"), ID_DEBUG_DISPLAY, ID_MAIN_PANEL, &mpDebugViewShow);

   

    // Create help text panel    
    pGUI->CreateText( _L("F1 for Help"), ID_IGNORE_CONTROL_ID, ID_SECONDARY_PANEL);
    pGUI->CreateText( _L("[Escape] to quit application"), ID_IGNORE_CONTROL_ID, ID_SECONDARY_PANEL);
    pGUI->CreateText( _L("A,S,D,F - move camera position"), ID_IGNORE_CONTROL_ID, ID_SECONDARY_PANEL);
    pGUI->CreateText( _L("Q - camera position up"), ID_IGNORE_CONTROL_ID, ID_SECONDARY_PANEL);
    pGUI->CreateText( _L("E - camera position down"), ID_IGNORE_CONTROL_ID, ID_SECONDARY_PANEL);
    pGUI->CreateText( _L("mouse + right click - camera look location"), ID_IGNORE_CONTROL_ID, ID_SECONDARY_PANEL);
#ifdef AVSM_METRICS
	pGUI->CreateText( _L("Volumetric shadow map render time (ms):"), ID_SMRT, ID_MAIN_PANEL,&mpSMRT);
    mpSMRT->SetAutoArranged(false);
	mpSMRT->SetPosition(5,50);

	mpSMRT->SetVisibility(CPUT_CHECKBOX_CHECKED == mpEnableStats->GetCheckboxState());

    pGUI->CreateText( _L("Solid geometry render time (ms):"), ID_WST, ID_MAIN_PANEL,&mpWST);
    mpWST->SetAutoArranged(false);
	mpWST->SetPosition(5,65);
	mpWST->SetVisibility(CPUT_CHECKBOX_CHECKED == mpEnableStats->GetCheckboxState());

	pGUI->CreateText( _L("Particles (...) render time (ms): "), ID_PST, ID_MAIN_PANEL,&mpPST);
    mpPST->SetAutoArranged(false);
	mpPST->SetPosition(5,80);
	mpWST->SetVisibility(CPUT_CHECKBOX_CHECKED == mpEnableStats->GetCheckboxState());
#endif

    pGUI->SetActivePanel(ID_MAIN_PANEL);
    pGUI->DrawFPS(true);
    
    // create the 'slice' debug view
    SAFE_RELEASE(mpAVSMDebugView);
    mpAVSMDebugView = new AVSMDebugView();
    mpAVSMDebugView->Create(ID_MAIN_PANEL, ID_DEBUG_DISPLAY_SLICE_SLIDER, CPUT_CHECKBOX_CHECKED == mpDebugViewShow->GetCheckboxState());
    
    // Add our programatic (and global) material parameters
    pAssetLibrary->SetMediaDirectoryName(    _L("Media\\"));
    CPUTMaterial::mGlobalProperties.AddValue( _L("cbPerFrameValues"), _L("$cbPerFrameValues") );
    CPUTMaterial::mGlobalProperties.AddValue( _L("cbPerModelValues"), _L("#cbPerModelValues") );
    CPUTMaterial::mGlobalProperties.AddValue( _L("_Shadow"), _L("$shadow_depth") );

    // Create default shaders
    CPUTPixelShaderDX11  *pPS       = CPUTPixelShaderDX11::CreatePixelShaderFromMemory(            _L("$DefaultShader"), CPUT_DX11::mpD3dDevice,          _L("PSMain"), _L("ps_4_0"), gpDefaultShaderSource );
    CPUTPixelShaderDX11  *pPSNoTex  = CPUTPixelShaderDX11::CreatePixelShaderFromMemory(   _L("$DefaultShaderNoTexture"), CPUT_DX11::mpD3dDevice, _L("PSMainNoTexture"), _L("ps_4_0"), gpDefaultShaderSource );
    CPUTVertexShaderDX11 *pVS       = CPUTVertexShaderDX11::CreateVertexShaderFromMemory(          _L("$DefaultShader"), CPUT_DX11::mpD3dDevice,          _L("VSMain"), _L("vs_4_0"), gpDefaultShaderSource );
    CPUTVertexShaderDX11 *pVSNoTex  = CPUTVertexShaderDX11::CreateVertexShaderFromMemory( _L("$DefaultShaderNoTexture"), CPUT_DX11::mpD3dDevice, _L("VSMainNoTexture"), _L("vs_4_0"), gpDefaultShaderSource );

    // We just want to create them, which adds them to the library.  We don't need them any more so release them, leaving refCount at 1 (only library owns a ref)
    SAFE_RELEASE(pPS);
    SAFE_RELEASE(pPSNoTex);
    SAFE_RELEASE(pVS);
    SAFE_RELEASE(pVSNoTex);

    // load shadow casting material+sprite object
    cString ExecutableDirectory;
    CPUTOSServices::GetOSServices()->GetExecutableDirectory(&ExecutableDirectory);
    pAssetLibrary->SetMediaDirectoryName(  ExecutableDirectory+_L(".\\Media\\"));
    mpShadowRenderTarget = new CPUTRenderTargetDepth();
    mpShadowRenderTarget->CreateRenderTarget( cString(_L("$shadow_depth")), SHADOW_WIDTH_HEIGHT, SHADOW_WIDTH_HEIGHT, DXGI_FORMAT_D16_UNORM );

    // create the debug view sprite
     mpAVSMDebugView->CreateSprite();

    // Call ResizeWindow() because it creates some resources we'll be needing (e.g., the back buffer)
    int width, height;
    CPUTOSServices::GetOSServices()->GetClientDimensions(&width, &height);
    ResizeWindow(width, height);

    CPUTRenderStateBlockDX11 *pBlock = new CPUTRenderStateBlockDX11();
    CPUTRenderStateDX11 *pStates = pBlock->GetState();

    // Override default sampler desc for our default shadowing sampler
    pStates->SamplerDesc[1].Filter         = D3D11_FILTER_COMPARISON_MIN_MAG_LINEAR_MIP_POINT;
    pStates->SamplerDesc[1].AddressU       = D3D11_TEXTURE_ADDRESS_BORDER;
    pStates->SamplerDesc[1].AddressV       = D3D11_TEXTURE_ADDRESS_BORDER;
    pStates->SamplerDesc[1].ComparisonFunc = D3D11_COMPARISON_LESS; // D3D11_COMPARISON_GREATER

    // what mpAVSMTechnique->mAVSMSampler used to be
    {
        CD3D11_SAMPLER_DESC desc(D3D11_DEFAULT);
        desc.Filter = D3D11_FILTER_MIN_MAG_MIP_POINT;
        desc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
        desc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
        desc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
        pStates->SamplerDesc[2] = desc;
    }

    // what mpAVSMTechnique->mAVSMGenSampler used to be (gAVSMGenCtrlSurfaceSampler in shaders)
    {
        float borderColor[4] = {0.f, 0.f, 0.f, 0.f};
        CD3D11_SAMPLER_DESC desc(D3D11_DEFAULT);
        desc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
        desc.AddressU = D3D11_TEXTURE_ADDRESS_BORDER;
        desc.AddressV = D3D11_TEXTURE_ADDRESS_BORDER;
        desc.AddressW = D3D11_TEXTURE_ADDRESS_BORDER;
        *desc.BorderColor = *borderColor;
        pStates->SamplerDesc[3] = desc;
    }

    pBlock->CreateNativeResources();
    CPUTAssetLibrary::GetAssetLibrary()->AddRenderStateBlock( _L("$DefaultRenderStates"), pBlock );
    pBlock->Release(); // We're done with it.  The library owns it now.
    
    
    // initialize the AVSM subsystem     
    unsigned int shadowTextureDim     = SHADOW_WIDTH_HEIGHT;
    unsigned int avsmShadowTextureDim = AVSM_SHADOW_TEXTURE_DIM;    //  Shadow Texture Dimension - ID_SHADOW_TEXTURE_DIMENSION

    mAppOptions.scene = GROUND_PLANE_SCENE;
    mAppOptions.enableParticles = true;
    mAppOptions.enableAutoBoundsAVSM = false;
    mAppOptions.enableShadowPicking = false;
    mAppOptions.NodeCount = AVSM_START_NODE_COUNT;
    mAppOptions.enableTransmittanceCurve = false;
    mAppOptions.enableVolumeShadowCreation = true;
    mAppOptions.pickedX = 0;
    mAppOptions.pickedY = 0;    

    // create the technique class
    mpAVSMTechnique = new AVSMTechnique(mpD3dDevice, mpContext, mAppOptions.NodeCount, shadowTextureDim, avsmShadowTextureDim);
        
    //  Load the custom shaders needed for the AVSM technique
    //-------------------------------------------------------
    bool ReturnCode = mpAVSMTechnique->LoadAVSMShaders(mpD3dDevice, mpContext, mAppOptions.NodeCount, shadowTextureDim, avsmShadowTextureDim);
    ASSERT( (true==ReturnCode), _L("CPUT Error loading AVSM shaders.") );
    
    //  Create the render states needed for the AVSM technique
    //-------------------------------------------------------
    ReturnCode = mpAVSMTechnique->CreateAVSMRenderStates(mpD3dDevice, mpContext, mAppOptions.NodeCount, shadowTextureDim, avsmShadowTextureDim);
    ASSERT( (true==ReturnCode), _L("CPUT Error creating AVSM render states.") );

    //  Create the buffers needed for the AVSM technique
    //-------------------------------------------------------
    ReturnCode = mpAVSMTechnique->CreateAVSMBuffers(mpD3dDevice, mpContext, mAppOptions.NodeCount, avsmShadowTextureDim);
    ASSERT( (true==ReturnCode), _L("CPUT Error creating AVSM buffers.") );

    // We wrap buffers used by the AVSM technique when they are used by CPUT because it allows
    // the CPUT shader auto-bind system to automatically bind the buffers when they're used
    //-------------------------------------------------------
    mpAVSMTechnique->WrapBuffers();

    // Set the debug view material
    mpAVSMDebugView->SetMaterial( pAssetLibrary->GetMaterial( L"AVSMDebugView" ));


    // Load .set file that was specified on the command line
    // Otherwise, load the default object if no .set was specified

    cString MediaPath = _L(".\\Media\\");
    pAssetLibrary->SetMediaDirectoryName( MediaPath );

    // Load assets 
    mpSkyBox    = pAssetLibrary->GetAssetSet( _L("sky") );
    mpAssetSet  = pAssetLibrary->GetAssetSet( _L("roofTop") );  
    

	// If no cameras were created from the model sets then create a default simple camera
	// and add it to the camera array.
    if( mpAssetSet && mpAssetSet->GetCameraCount() )
    {
        mpCamera = mpAssetSet->GetFirstCamera();
        mpCamera->AddRef(); 
    } 
    else
    {
        mpCamera = new CPUTCamera();
        CPUTAssetLibraryDX11::GetAssetLibrary()->AddCamera( _L("SampleStart Camera"), mpCamera );

        float factorfactor = 1.5f;
        mpCamera->SetPosition( 70.44f/factorfactor, 13.27f/factorfactor, -18.786f/factorfactor); 
        // Set the projection matrix for all of the cameras to match our window.        
        mpCamera->SetAspectRatio(((float)width)/((float)height));

        mpCamera->SetFov(XMConvertToRadians(60.0f)); 
        mpCamera->SetFarPlaneDistance(10000.0f);    
        mpCamera->LookAt( 18.823f, 1.077f, -0.038f); 
        mpCamera->Update();
    }

   
    mpShadowCamera = new CPUTCamera();
    CPUTAssetLibraryDX11::GetAssetLibrary()->AddCamera( _L("ShadowCamera"), mpShadowCamera );

    mpCameraController = new CPUTCameraControllerFPS();
    mpCameraController->SetCamera(mpCamera);
    mpCameraController->SetLookSpeed(0.004f);
    mpCameraController->SetMoveSpeed(70.0f);

    // Initialize with the current surface description
    D3D11_RENDER_TARGET_VIEW_DESC BackBufferRTVDesc;    
    mpBackBufferRTV->GetDesc(&BackBufferRTVDesc);
    
    ID3D11Resource* pBackBufferRTV=NULL;
    mpBackBufferRTV->GetResource(&pBackBufferRTV);

    // get the texture
    ID3D11Texture2D* pBackBufferTexture=NULL;
    pBackBufferRTV->QueryInterface(__uuidof(ID3D11Texture2D), (void**)&pBackBufferTexture);

    // get surface desc
    D3D11_TEXTURE2D_DESC surfDesc;
    pBackBufferTexture->GetDesc(&surfDesc);
    SAFE_RELEASE(pBackBufferTexture);
    SAFE_RELEASE(pBackBufferRTV);

    D3DXMatrixIdentity( &gWorldMatrix );
    
    // Create a particle system for this specific scene
    CreateParticles(mpD3dDevice);

    // get the current gui state
    GetGUIState(&mCurrentGUIState);

    SetAmbientColor( float3( 0.05f, 0.05f, 0.05f ) );
    SetLightColor( float3( 1.2f, 1.2f, 1.2f ) );

#ifdef SSAO_ENABLE
    mSSAO.Create( mpD3dDevice, mpContext );
#endif
}

// Create our very simple particle system.  Could use much more fancy system
// with wind physics/etc - but this works for our demonstration purposes.
//-----------------------------------------------------------------------------
void AVSMSample::CreateParticles(ID3D11Device *d3dDevice)
{
    SAFE_DELETE(gParticleSystem);

    AVSMGuiState GUIState;
    std::string shadowAAStr;
    std::string avsmNodeCountStr;
    std::ostringstream oss;

    // get the current GUI settings to make sure we have most updated
    // particle size/opacity values
    GetGUIState(&GUIState);

    // convert to strings
    oss << GUIState.NumberOfNodes;
    shadowAAStr = oss.str();
    oss << GUIState.ShadowTextureDimensions;
    avsmNodeCountStr = oss.str();

    // far left low mushroom plume
    ParticleEmitter emitter0;
    {
        emitter0.mDrag           = 0.1f;
        emitter0.mGravity        = -0.1f;
        emitter0.mRandScaleX     = 1.0f;
        emitter0.mRandScaleY     = 0.1f;
        emitter0.mRandScaleZ     = 1.0f;
        emitter0.mLifetime       = 15.0f;
        emitter0.mStartSize      = 3.0f;
        emitter0.mSizeRate       = 0.05f;
        emitter0.mpPos[0]        = 12.15f;
        emitter0.mpPos[1]        = -5.578f; 
        emitter0.mpPos[2]        = -9.48f;
        emitter0.mpVelocity[0]   = 0.0f;
        emitter0.mpVelocity[1]   = 1.0f;
        emitter0.mpVelocity[2]   = 0.5f;
    }
    
    // tall front plume
    ParticleEmitter emitter1;
    {
        emitter1.mDrag           = 0.2f;
        emitter1.mGravity        = 0.0f;
        emitter1.mRandScaleX     = 1.5f;
        emitter1.mRandScaleY     = 0.5f;
        emitter1.mRandScaleZ     = 1.5f;
        emitter1.mLifetime       = 18.0f;
        emitter1.mStartSize      = 4.0f;
        emitter1.mSizeRate       = 0.05f;
        emitter1.mpPos[0]        = 21.932f;
        emitter1.mpPos[1]        = -0.899f; 
        emitter1.mpPos[2]        = 12.227f;
        emitter1.mpVelocity[0]   = -1.0f;
        emitter1.mpVelocity[1]   = 3.0f;
        emitter1.mpVelocity[2]   = 0.0f;
    }
    

    // horizontal steam jet
    ParticleEmitter emitter2;
    {
        emitter2.mDrag           = 0.2f;
        emitter2.mGravity        = 0.5f;
        emitter2.mRandScaleX     = 0.7f;
        emitter2.mRandScaleY     = 0.7f;
        emitter2.mRandScaleZ     = 0.7f;
        emitter2.mLifetime       = 6.0f;
        emitter2.mStartSize      = 5.0f;
        emitter2.mSizeRate       = 0.05f;
        emitter2.mpPos[0]        = 27.6f; // front to back
        emitter2.mpPos[1]        = -1.5f; // height
        emitter2.mpPos[2]        = 5.4f;  // left/right
        emitter2.mpVelocity[0]   = -5.0f;
        emitter2.mpVelocity[1]   = 1.2f;
        emitter2.mpVelocity[2]   = -3.0f;
    }           

    const ParticleEmitter* emitters[3] = {&emitter0, &emitter1, &emitter2};
    const UINT maxNumPartices = 5000;
    D3D10_SHADER_MACRO shaderDefines[] = {
        {"AVSM_NODE_COUNT", avsmNodeCountStr.c_str()},
        {0, 0}
    };

    // create our simple particle system with 3 emitters
    gParticleSystem = new ParticleSystem(maxNumPartices);
    gParticleSystem->InitializeParticles(emitters, 3, d3dDevice, 8, 8, shaderDefines);

}


// System shutdown
//-----------------------------------------------------------------------------
void AVSMSample::Shutdown()
{
    // destroy the particle system
    SAFE_DELETE(gParticleSystem);

    // destroy the AVSMTechnique
    SAFE_DELETE(mpAVSMTechnique);

    // destroy the debug view
    SAFE_RELEASE(mpAVSMDebugView);
}

// Update routine before rendering our frame
//-----------------------------------------------------------------------------
void AVSMSample::Update(double deltaSeconds)
{
    if( mpCameraController != NULL )
    { 
        mpCameraController->Update((float)deltaSeconds);
    }

    // Set up the shadow camera (a camera that sees what the light sees)
    {
        float3 lookAtPoint(0.0f, 0.0f, 0.0f);
        float3 half(1.0f, 1.0f, 1.0f);
        if( mpAssetSet )
        {
            float3 center;
            mpAssetSet->GetBoundingBox( &center, &half );
        }
        float length = half.length();

        mpShadowCamera->SetFov(XMConvertToRadians(45));
        mpShadowCamera->SetAspectRatio(1.0f);
        float fov = mpShadowCamera->GetFov();
        float tanHalfFov = tanf(fov * 0.5f);
        float cameraDistance = length/tanHalfFov;
        float nearDistance = cameraDistance * 0.1f;
        mpShadowCamera->SetNearPlaneDistance(nearDistance);
        mpShadowCamera->SetFarPlaneDistance(1000.0f * cameraDistance);

        float shadowCameraDistance = 140.0f;
        float3 shadowCameraPosition; // = float3( 85.0f, 38.0f, 42.0f);

        // always update these guys so we can track at realtime
        mpLightLatitudeSlider->GetValue( mCurrentGUIState.LightLatitude );
        mpLightLongitudeSlider->GetValue( mCurrentGUIState.LightLongitude );

        float longRad = mCurrentGUIState.LightLongitude / 180.0f * 3.14159265f;
        float latRad = mCurrentGUIState.LightLatitude / 180.0f * 3.14159265f;

        // Convert from long/lat to 3D spherical coordinates
        shadowCameraPosition.x = cosf( longRad ) * cosf( latRad );
        shadowCameraPosition.y = sinf( latRad ); // y is up
        shadowCameraPosition.z = sinf( longRad ) * cosf( latRad );

        shadowCameraPosition *= shadowCameraDistance;

        mpShadowCamera->SetPosition( shadowCameraPosition );
        mpShadowCamera->LookAt( lookAtPoint.x, lookAtPoint.y, lookAtPoint.z );
        mpShadowCamera->Update();
    }
}

// Handle keyboard events
//-----------------------------------------------------------------------------
CPUTEventHandledCode AVSMSample::HandleKeyboardEvent(CPUTKey key)
{
    static bool panelToggle = false;
    CPUTEventHandledCode    handled = CPUT_EVENT_UNHANDLED;
    cString fileName;
    CPUTGuiControllerDX11*  pGUI = CPUTGetGuiController();

    switch(key)
    {
    case KEY_F1:
        panelToggle = !panelToggle;
        if(panelToggle)
        {
            pGUI->SetActivePanel(ID_SECONDARY_PANEL);
        }
        else
        {
            pGUI->SetActivePanel(ID_MAIN_PANEL);
        }
        handled = CPUT_EVENT_HANDLED;
        break;
    case KEY_L:
        {
            static int cameraObjectIndex = 0;
            CPUTRenderNode *pCameraList[] = { mpCamera, mpShadowCamera };
            cameraObjectIndex = (++cameraObjectIndex) % (sizeof(pCameraList)/sizeof(*pCameraList));
            CPUTRenderNode *pCamera = pCameraList[cameraObjectIndex];
            mpCameraController->SetCamera( pCamera );
        }
        handled = CPUT_EVENT_HANDLED;
        break;
    case KEY_ESCAPE:
        handled = CPUT_EVENT_HANDLED;
        CPUT_DX11::Shutdown();
        break;
    }

    // pass it to the camera controller
    if(handled == CPUT_EVENT_UNHANDLED)
    {
        handled = mpCameraController->HandleKeyboardEvent(key);
    }
    return handled;
}

// Handle mouse events
//-----------------------------------------------------------------------------
CPUTEventHandledCode AVSMSample::HandleMouseEvent(int x, int y, int wheel, CPUTMouseState state)
{
    if(mpDebugViewShow)
    {
        mLButtonDown = (state & CPUT_MOUSE_LEFT_DOWN)!=0;
        mRButtonDown = (state & CPUT_MOUSE_RIGHT_DOWN)!=0;

        if( CPUT_CHECKBOX_CHECKED == mpDebugViewShow->GetCheckboxState() )
        {
            // Send this mouse event to the debug view
            if(CPUT_EVENT_HANDLED == mpAVSMDebugView->HandleMouseEvent(x,y,wheel,state, mLButtonDown, mRButtonDown ))
            {
                return CPUT_EVENT_HANDLED;
            }           
        }

        // camera movement
        if( mpCameraController )
        {
            return mpCameraController->HandleMouseEvent(x, y, wheel, state);
        }
    }
    return CPUT_EVENT_UNHANDLED;
}

// Handle any control callback events
//-----------------------------------------------------------------------------
void AVSMSample::HandleCallbackEvent( CPUTEventID Event, CPUTControlID ControlID, CPUTControl *pControl )
{
    UNREFERENCED_PARAMETER(Event);
    UNREFERENCED_PARAMETER(pControl);
    
    AVSMGuiState NewGUIState;

    switch(ControlID)
    {
    case ID_FULLSCREEN_BUTTON:
        CPUTToggleFullScreenMode();
        break;

    case ID_VOLUME_SHADOWING_METHOD:
        // update the current state to match new state
        GetGUIState(&NewGUIState);
        mCurrentGUIState = NewGUIState;
        break;

	case ID_SHADOW_TEXTURE_DIMENSION:
        GetGUIState(&NewGUIState);
        if(mCurrentGUIState.ShadowTextureDimensions != NewGUIState.ShadowTextureDimensions)
		{
		    bool ReturnCode=0;

			mpAVSMTechnique->mAVSMShadowTextureDim = NewGUIState.ShadowTextureDimensions;

			ReturnCode = mpAVSMTechnique->DeleteAVSMBuffers(mpD3dDevice, mpContext, mAppOptions.NodeCount);
			ASSERT( (true==ReturnCode), _L("CPUT Error Deleting AVSM buffers.") );

			ReturnCode = mpAVSMTechnique->CreateAVSMBuffers(mpD3dDevice, mpContext, mAppOptions.NodeCount, mpAVSMTechnique->mAVSMShadowTextureDim);
			ASSERT( (true==ReturnCode), _L("CPUT Error creating AVSM buffers.") );
		}
        mCurrentGUIState = NewGUIState;
		break;
	case ID_WIREFRAME_PARTICLES:
	case ID_AVSM_NODE_COUNT:
	case ID_ENABLE_STATS:
	case ID_PAUSE_PARTICLES:
	case ID_PARTICLE_SIZE:
	case ID_PARTICLE_OPACITY:
	case ID_PARTICLETESSELLATION:
	case ID_VERTEX_SHADE_PARTICLES:
    case ID_LIGHT_ORBIT_LAT_SLIDER:
    case ID_LIGHT_ORBIT_LONG_SLIDER:
        GetGUIState(&NewGUIState);
        mCurrentGUIState = NewGUIState;
		break;
    case ID_DEBUG_DISPLAY:
        mpAVSMDebugView->SetVisibility( CPUT_CHECKBOX_CHECKED == mpDebugViewShow->GetCheckboxState() );        
        break;
	default:
        break;
    }
}

// Get state of current gui controls
//-----------------------------------------------------------------------------
void AVSMSample::GetGUIState(AVSMGuiState* pState)
{
    cString SelectedItem;
    ASSERT(pState, _L("Null GUI state object"));

    // AVSM method
    mpVolumeShadowingMethod->GetSelectedItem(SelectedItem);
    if(_L("Disabled") == SelectedItem)
    {
        pState->Method = AVSM_DISABLED;
    }
    else if(_L("DirectX11 Method") == SelectedItem)
    {
        pState->Method = AVSM_METHOD_DIRECTX;
    }
    else if(_L("Pixel Shader Ordering") == SelectedItem)
    {
        pState->Method = AVSM_METHOD_INTEL_EXT;
    }
    else
    {
        ASSERT(0, _L("Unknown AVSM state"));
    }

    // shadow texture size
    mpShadowTextureDimension->GetSelectedItem(SelectedItem);
    if(L"512x512" == SelectedItem)
    {
        pState->ShadowTextureDimensions = 512;
    }
    else if(L"256x256" == SelectedItem)
    {
        pState->ShadowTextureDimensions = 256;
    }
    else if(L"128x128" == SelectedItem)
    {
        pState->ShadowTextureDimensions = 128;
    }
    else
    {
        ASSERT(0, _L("Unknown shadow texture size"));
    }

    // shadow texture size
    mpAVSMNodeCount->GetSelectedItem(SelectedItem);
    if(L"8 nodes per sample" == SelectedItem)
    {
        pState->NumberOfNodes = 8;
    }
    else if(L"4 nodes per sample" == SelectedItem)
    {
        pState->NumberOfNodes = 4;
    }
    else
    {
        ASSERT(0, _L("Unknown Node size"));
    }

    // particle sliders
    mpParticleSize->GetValue(pState->ParticleSize);
	pState->ParticleSize /= 20.0;
    mpParticleOpacity->GetValue(pState->ParticleOpacity);

    // paused checkbox
    pState->ParticlesPaused = false;
    if( CPUT_CHECKBOX_CHECKED == mpPauseParticles->GetCheckboxState())
    {
         pState->ParticlesPaused = true;
    }
#ifdef AVSM_METRICS
    // enable stats checkbox
    pState->EnableStats = false;
    if( CPUT_CHECKBOX_CHECKED == mpEnableStats->GetCheckboxState())
    {
         pState->EnableStats = true;
    }
#endif
    //  enablewireframe
    pState->WireFrame = false;
    if( CPUT_CHECKBOX_CHECKED == mpWireFrameParticles->GetCheckboxState())
    {
         pState->WireFrame = true;
    }
    // enable Vertex Shade Particles
    pState->vertexShaderShadowLookup = false;
    if( CPUT_CHECKBOX_CHECKED == mpVertexShadeParticles->GetCheckboxState())
    {
         pState->vertexShaderShadowLookup = true;
    }

    mpParticleTessellation->GetValue(pState->TessellationDensity);

	pState->tessellate = (pState->TessellationDensity)? ( CPUT_CHECKBOX_CHECKED == mpVertexShadeParticles->GetCheckboxState()) : 0;

    mpLightLatitudeSlider->GetValue( pState->LightLatitude );
    mpLightLongitudeSlider->GetValue( pState->LightLongitude );
}


// Handle resize events
//-----------------------------------------------------------------------------
void AVSMSample::ResizeWindow(UINT width, UINT height)
{
    if(0==width || 0==height)
        return;

#ifdef SSAO_ENABLE
    mSSAO.ResizeWindow( width, height );
#endif
    
    CPUTAssetLibrary *pAssetLibrary = CPUTAssetLibrary::GetAssetLibrary();

    // Before we can resize the swap chain, we must release any references to it.
    // We could have a "AssetLibrary::ReleaseSwapChainResources(), or similar.  But,
    // Generic "release all" works, is simpler to implement/maintain, and is not performance critical.
    pAssetLibrary->ReleaseTexturesAndBuffers();

    mpAVSMDebugView->ReleaseTexturesAndBuffers();

    // Resize the CPUT-provided render target
    CPUT_DX11::ResizeWindow( width, height );

    // Resize any application-specific render targets here
    if( mpCamera ) mpCamera->SetAspectRatio(((float)width)/((float)height));

    pAssetLibrary->RebindTexturesAndBuffers();

    mpAVSMDebugView->Resize(width, height);

    // save the new screen width/height
	mScreenWidth = (float)width;
	mScreenHeight = (float)height;

    // move the warning text
    int TextWidth,TextHeight;    
    mpExtensionWarning->GetDimensions(TextWidth,TextHeight);
    mpExtensionWarning->SetPosition(0,height-TextHeight-2);
}

#ifdef AVSM_METRICS
// Display various timings/etc
//-----------------------------------------------------------------------------
void AVSMSample::DisplayStats(AVSMTechnique::FrameContext & frameContext, const FrameStats &frameStats)
{
   const AVSMTechnique::Options & options                 = frameContext.Options;
   UIConstants * ui                        = &frameContext.GPUUIConstants;

    wchar_t wcstring[CPUT_MAX_STRING_LENGTH];
    swprintf_s(wcstring, CPUT_MAX_STRING_LENGTH, _L("AVSM shadow map create time (ms): \t%.2f"), frameStats.VolumetricShadowTime);

    mpSMRT->SetText(wcstring);

	if (ui->volumeShadowMethod && options.enableVolumeShadowCreation) 
	{
		mpSMRT->SetVisibility(CPUT_CHECKBOX_CHECKED == mpEnableStats->GetCheckboxState());
	}
	else
	{
		mpSMRT->SetVisibility(false);
	}

	swprintf_s(wcstring, CPUT_MAX_STRING_LENGTH, _L("Solid geometry shading time (ms): \t%.2f"), frameStats.SolidGeometryTime);
	mpWST->SetText(wcstring);
	mpWST->SetVisibility(CPUT_CHECKBOX_CHECKED == mpEnableStats->GetCheckboxState());

    int numberOfParticles = gParticleSystem->GetParticleCount();
    swprintf_s(wcstring, CPUT_MAX_STRING_LENGTH, _L("Particles (%04d) shading time (ms): \t%.2f"), numberOfParticles, frameStats.ParticlesTime);
	mpPST->SetText(wcstring);
	mpPST->SetVisibility(CPUT_CHECKBOX_CHECKED == mpEnableStats->GetCheckboxState());
}
#endif

// routine to translate GUi state to AVSMFrameContext.GPUUIConstant.* parameters
//-----------------------------------------------------------------------------
void AVSMSample::SetAVSMFrameContextData(AVSMTechnique::FrameContext& AVSMFrameContext)
{
    // update render method:
    switch( mCurrentGUIState.Method )
    {
    case( AVSM_DISABLED ):          AVSMFrameContext.GPUUIConstants.volumeShadowMethod = VOL_SHADOW_NO_SHADOW;  break;
    case( AVSM_METHOD_DIRECTX ):    AVSMFrameContext.GPUUIConstants.volumeShadowMethod = VOL_SHADOW_AVSM;       break;
    case( AVSM_METHOD_INTEL_EXT ):  AVSMFrameContext.GPUUIConstants.volumeShadowMethod = VOL_SHADOW_AVSM_GEN;   break;
    }

    AVSMFrameContext.GPUUIConstants.enableStats = mCurrentGUIState.EnableStats;
    AVSMFrameContext.GPUUIConstants.pauseParticleAnimaton = mCurrentGUIState.ParticlesPaused;
    AVSMFrameContext.GPUUIConstants.particleSize = mCurrentGUIState.ParticleSize;
    AVSMFrameContext.GPUUIConstants.particleOpacity = (unsigned int)mCurrentGUIState.ParticleOpacity;
    AVSMFrameContext.GPUUIConstants.wireframe = mCurrentGUIState.WireFrame;
    AVSMFrameContext.GPUUIConstants.vertexShaderShadowLookup = mCurrentGUIState.vertexShaderShadowLookup;
    AVSMFrameContext.GPUUIConstants.tessellate = mCurrentGUIState.tessellate;
    AVSMFrameContext.GPUUIConstants.TessellationDensity = 1.0f/(31.0f - mCurrentGUIState.TessellationDensity);

    // set the current size
    AVSMFrameContext.mScreenWidth = mScreenWidth;
    AVSMFrameContext.mScreenHeight = mScreenHeight;
}

//-----------------------------------------------------------------------------
void AVSMSample::Render(double deltaSeconds)
{
    CPUTGPUProfilerDX11_AutoScopeProfile timerAll( mGPUTimerAll, mCurrentGUIState.EnableStats );
    CPUTRenderParametersDX renderParams(mpContext);

    if( mpAVSMTechnique == NULL )
    {
        const float clearColor[] = { 1.0f, 0.0f, 0.0f, 1.0f };
        mpContext->ClearRenderTargetView( mpBackBufferRTV,  clearColor );
        return;
    }    


    // Set the viewport back to full window size
    D3D11_VIEWPORT viewport;
    viewport.Width    = (float)mScreenWidth;
    viewport.Height   = (float)mScreenHeight;
    viewport.MinDepth = 0.0f;
    viewport.MaxDepth = 1.0f;
    viewport.TopLeftX = 0.0f;
    viewport.TopLeftY = 0.0f;

    // 1. Initialize AVSM structures and update particle movement (but don't do any actual rendering yet)
    
    // Tell the materials the number of nodes to be used in the technique
    if( mCurrentGUIState.NumberOfNodes == 4 )
        CPUTMaterialDX11::GlobalSetCurrentSubMaterial( 0 );
    else if( mCurrentGUIState.NumberOfNodes == 8 )
        CPUTMaterialDX11::GlobalSetCurrentSubMaterial( 1 );
    else
    {
        ASSERT( false, L"Only 4 or 8 supported for mCurrentGUIState.NumberOfNodes" );
    }
    
    // initialize the frame context 
    AVSMTechnique::FrameContext AVSMFrameContext;
    mpAVSMTechnique->InitializeFrameContext( AVSMFrameContext, mAppOptions, mpContext, gParticleSystem, gWorldMatrix,
        mpCamera, mpShadowCamera, viewport );

    CPUTBufferDX11 * pMainDepthBuffer = (CPUTBufferDX11 *)CPUTAssetLibrary::GetAssetLibrary()->GetBuffer( L"$DepthBuffer" );
    if( pMainDepthBuffer != NULL )
    {
        AVSMFrameContext.DepthBufferSRV = pMainDepthBuffer->GetShaderResourceView();
        pMainDepthBuffer->Release();
    }

    // translate the GUI state to AVSMFrameContext parameters
    SetAVSMFrameContextData(AVSMFrameContext);

    // update the particle positions (but do not render)
    mpAVSMTechnique->UpdateParticles( AVSMFrameContext );

    // set up the number of nodes
    mpAVSMTechnique->SetNodeCount(mCurrentGUIState.NumberOfNodes);

    // 2. Draw the shadowed scene using a standard shadow map from the light's point of view  
    // one could also use cascades shadow maps or other shadowing techniques.  We choose simple
    // shadow mapping for simplicity/demonstration purposes
    CPUTCamera *pLastCamera = mpCamera;
    mpCamera = renderParams.mpCamera = mpShadowCamera;
    mpShadowRenderTarget->SetRenderTarget( renderParams, 0, 1.0f, true );

    if( mpAssetSet ) { mpAssetSet->RenderShadowRecursive( renderParams ); }

    mpShadowRenderTarget->RestoreRenderTarget(renderParams);
    mpCamera = renderParams.mpCamera = pLastCamera;
    renderParams.mpShadowCamera = mpShadowCamera;

    // Clear render target and depth buffer
    const float clearColor[] = { 0.0993f, 0.0993f, 0.1993f, 1.0f };
    mpContext->ClearRenderTargetView( mpBackBufferRTV,  clearColor );
    mpContext->ClearDepthStencilView( mpDepthStencilView, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1.0f, 0);

    // Draw scene    
    {
        // Save our main RTV/DSV
        ID3D11RenderTargetView* pRTV = CPUTRenderTargetColor::GetActiveRenderTargetView();
        ID3D11DepthStencilView* pDSV = CPUTRenderTargetDepth::GetActiveDepthStencilView();
  
        // 3. create the AVSM shadow map and capture particle fragments into the AVSM render targets
        {
            CPUTGPUProfilerDX11_AutoScopeProfile timer( mGPUTimerAVSMCreation, mCurrentGUIState.EnableStats );
            mpAVSMTechnique->CreateAVSMShadowMap( AVSMFrameContext, mpAssetSet, &renderParams );
        }

        // 4. set the standard render target and draw the regular scene geometry
        {
            CPUTGPUProfilerDX11_AutoScopeProfile timer( mGPUTimerSolidGeometryRendering, mCurrentGUIState.EnableStats );
         
            // draw CPUT loaded geometry/scene
            // set the render targets and viewports
            mpContext->OMSetRenderTargets( 1, &pRTV, pDSV );  
            mpContext->RSSetViewports(1, &viewport);
        
            // set all 'external' states for drawing regular scene geometry
            mpAVSMTechnique->UpdateStatesAndConstantsForAVSMUse( AVSMFrameContext, true );

            // render scene geometry - passing in the AVSM buffers
            renderParams.mAVSMMethod = mCurrentGUIState.Method;

            // render the skybox - which does not accept AVSM shadows
            if( mpSkyBox ) { mpSkyBox->RenderRecursive(renderParams); }

            // render the scene - accepting AVSM shadows
            if( mpAssetSet ) { mpAssetSet->RenderAVSMShadowedRecursive(renderParams); }
        }

        // 5. draw screen space ambient occlusion
#ifdef SSAO_ENABLE        
        {
            mSSAO.UpdateConsts( mpContext, mpCamera );
            mSSAO.GenerateSSAO( mpContext );
            mSSAO.FullscreenApplySSAO( mpContext );
        }
#endif

        // 6. Render the shaded particles into the standard render target
        {
            CPUTGPUProfilerDX11_AutoScopeProfile timer( mGPUTimerParticlesRendering, mCurrentGUIState.EnableStats );

            // remove depth buffer since we're doing soft depth blend in the pixel shader
            mpContext->OMSetRenderTargets( 1, &pRTV, NULL );

            // set all 'external' states for drawing 'external' geometry (particles)
            mpAVSMTechnique->UpdateStatesAndConstantsForAVSMUse( AVSMFrameContext, false );

            // render the shadowed/lit particles
            mpAVSMTechnique->RenderShadedParticles( AVSMFrameContext, pRTV, pDSV );
        }

        // restore our main RTV/DSV
        mpContext->OMSetRenderTargets( 1, &pRTV, pDSV );       
    }

    // Draw the debug view
    if( CPUT_CHECKBOX_CHECKED == mpDebugViewShow->GetCheckboxState() )
    {
        mpAVSMDebugView->Render(mpContext, mpAVSMTechnique, &AVSMFrameContext, &renderParams);
    }

    // update stats
    if( mCurrentGUIState.EnableStats )
    {
        AVSMFrameContext.Stats.VolumetricShadowTime = (float)(mGPUTimerAVSMCreation.GetAvgTime() * 1000.0);
        AVSMFrameContext.Stats.SolidGeometryTime    = (float)(mGPUTimerSolidGeometryRendering.GetAvgTime()* 1000.0);
        AVSMFrameContext.Stats.ParticlesTime        = (float)(mGPUTimerParticlesRendering.GetAvgTime()* 1000.0);
        AVSMFrameContext.Stats.TotalTime            = (float)(mGPUTimerAll.GetAvgTime() * 1000.0);
    }
    else
    {
        memset( &AVSMFrameContext.Stats, 0, sizeof(AVSMFrameContext.Stats) );
    }

#ifdef AVSM_METRICS
    // update frame stat text
    DisplayStats(AVSMFrameContext, AVSMFrameContext.Stats);
#endif

    // draw the GUI elements
    CPUTDrawGUI();
}


