//--------------------------------------------------------------------------------------
// Copyright 2013 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//--------------------------------------------------------------------------------------
#include "AVSMDebugView.h"
#include "CPUTGuiControllerDX11.h"


// utility function for debug view
//-----------------------------------------------------------------------------
static inline float clamp( double value, double low, double high )
{
    return (float) (value < low ? low : (value > high ? high : value));
}

// Constructor
//-----------------------------------------------------------------------------
AVSMDebugView::AVSMDebugView():mpDebugSprite(NULL),
    mpAVSMDebugViewConsts(NULL),
    mpAVSMDebugViewMaterial(NULL),
    mpExtensionWarning(NULL),
    mpDebugViewSliceDepth(NULL),
    mAVSMDebugViewUVZoomMulAdd( 1.0f, 1.0f, 0.0f, 0.0f ),
    mAVSMDebugViewMarkerPos( -1.0f, -1.0f, 0.0f, 0.0f ),
    mMouseDragStartPos( 0.0f, 0.0f, 0.0f )
{

}

// Destructor
//-----------------------------------------------------------------------------
AVSMDebugView::~AVSMDebugView()
{
    SAFE_DELETE( mpDebugSprite);
    SAFE_RELEASE( mpAVSMDebugViewConsts );
    SAFE_RELEASE( mpAVSMDebugViewMaterial );
}

// create the buffers, display sliders and other resources needed by this view
//-----------------------------------------------------------------------------
void AVSMDebugView::Create(const CPUTControlID PanelID, const CPUTControlID SliderID, bool Visible)
{
    CPUTAssetLibrary *pAssetLibrary = CPUTAssetLibrary::GetAssetLibrary();

    CPUTGuiControllerDX11 *pGUI = CPUTGuiControllerDX11::GetController();

    pGUI->CreateText(_L("\t "), -1, PanelID);
    pGUI->CreateSlider(_L("Debug Slice Position"), SliderID, PanelID, &mpDebugViewSliceDepth);
    mpDebugViewSliceDepth->SetScale( 100.0f, 160.0f, 500 );
    mpDebugViewSliceDepth->SetValue( 120.0f );
    mpDebugViewSliceDepth->SetAutoArranged(false);

    // set the visibility of the debug view sprite/slider
    SetVisibility(Visible);

    // create the debug view constant buffer
    D3D11_BUFFER_DESC bd = {0};
    bd.ByteWidth = sizeof(AVSMDebugViewConsts);
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.Usage = D3D11_USAGE_DYNAMIC;
    bd.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    ID3D11Buffer * pOurConstants;
    HRESULT hr = (CPUT_DX11::GetDevice())->CreateBuffer( &bd, NULL, &pOurConstants );
    ASSERT( !FAILED(hr), _L("Failed to create the depth debug view."));

    // add the constant buffer to the asset library for auto-binding
    cString name = _L("$cbAVSMDebugViewConsts");
    mpAVSMDebugViewConsts = new CPUTBufferDX11( name, pOurConstants );
    CPUTAssetLibrary::GetAssetLibrary()->AddConstantBuffer( name, mpAVSMDebugViewConsts );
    SAFE_RELEASE(pOurConstants); // We're done with it.  The CPUTBuffer now owns it.
        
    // add the name for the constant buffer
    pAssetLibrary->SetMediaDirectoryName(    _L("Media\\"));
    CPUTMaterial::mGlobalProperties.AddValue( _L("cbAVSMDebugViewConsts"), _L("$cbAVSMDebugViewConsts") );
}

// create the actual sprite object the debug view is displayed on
//-----------------------------------------------------------------------------
void AVSMDebugView::CreateSprite()
{
    int width, height;
    CPUTOSServices::GetOSServices()->GetClientDimensions(&width, &height);
    mpDebugSprite = new CPUTSprite();
    mpDebugSprite->CreateSprite( c_DebugSpriteX, c_DebugSpriteY, c_DebugSpriteWidth, c_DebugSpriteHeight, _L("Sprite") );
}

// set the material that is used when rendering the debug view
//-----------------------------------------------------------------------------
void AVSMDebugView::SetMaterial(CPUTMaterial *pMaterial)
{
    mpAVSMDebugViewMaterial = pMaterial;
}

// Handle mouse events for the debug view
//-----------------------------------------------------------------------------
CPUTEventHandledCode AVSMDebugView::HandleMouseEvent(int x, int y, int wheel, CPUTMouseState state, bool mLButtonDown, bool mRButtonDown )
{
    int width, height;
    CPUTOSServices::GetOSServices()->GetClientDimensions(&width, &height);
    float spriteUVX = (1.0f + c_DebugSpriteX) * 0.5f * width;
    float spriteUVY = (c_DebugSpriteY + 1.0f) * 0.5f * height;
    float spriteUVW = c_DebugSpriteWidth * 0.5f  * width;
    float spriteUVH = c_DebugSpriteHeight * 0.5f * height;
    float2 mouseUVOrig = float2( (x - spriteUVX)/spriteUVW, (y - spriteUVY)/spriteUVH );
    float2 mouseUV;
    mouseUV.x = mouseUVOrig.x * mAVSMDebugViewUVZoomMulAdd.x + mAVSMDebugViewUVZoomMulAdd.z;
    mouseUV.y = mouseUVOrig.y * mAVSMDebugViewUVZoomMulAdd.y + mAVSMDebugViewUVZoomMulAdd.w;

    if( mMouseDragStartPos.z == 1.0f ) // if isDragging
    {
        mMouseDragStartPos.z = (mRButtonDown)?(1.0f):(0.0f);

        if( mMouseDragStartPos.z == 1.0f ) // still isDragging
        {
            float deltaUVX = mMouseDragStartPos.x - mouseUV.x;
            float deltaUVY = mMouseDragStartPos.y - mouseUV.y;

            mAVSMDebugViewUVZoomMulAdd.z += deltaUVX;
            mAVSMDebugViewUVZoomMulAdd.w += deltaUVY;

            mMouseDragStartPos.x = mouseUV.x + deltaUVX;
            mMouseDragStartPos.y = mouseUV.y + deltaUVY;
            return CPUT_EVENT_HANDLED;
        }
    }

    if( (mouseUVOrig.x >= 0) && (mouseUVOrig.y >= 0) && (mouseUVOrig.x <= 1) && (mouseUVOrig.y <= 1) )
    {
        if( wheel != 0 ) // zoom in/out
        {
            float wheelDelta = (float)wheel / 1200.0f;

            // zoom
            mAVSMDebugViewUVZoomMulAdd.x *= (1.0f-wheelDelta);
            mAVSMDebugViewUVZoomMulAdd.y *= (1.0f-wheelDelta);

            // clamp to min/max
            mAVSMDebugViewUVZoomMulAdd.x = clamp( mAVSMDebugViewUVZoomMulAdd.x, 1.0f/32.0f, 1.0f );
            mAVSMDebugViewUVZoomMulAdd.y = clamp( mAVSMDebugViewUVZoomMulAdd.y, 1.0f/32.0f, 1.0f );

            // make it zoom around mouse position
            float2 mouseUVPrev = mouseUV;
            mouseUV.x = mouseUVOrig.x * mAVSMDebugViewUVZoomMulAdd.x + mAVSMDebugViewUVZoomMulAdd.z;
            mouseUV.y = mouseUVOrig.y * mAVSMDebugViewUVZoomMulAdd.y + mAVSMDebugViewUVZoomMulAdd.w;
            float deltaUVX = mouseUVPrev.x - mouseUV.x;
            float deltaUVY = mouseUVPrev.y - mouseUV.y;
            mAVSMDebugViewUVZoomMulAdd.z += deltaUVX;
            mAVSMDebugViewUVZoomMulAdd.w += deltaUVY;
            mouseUV.x = mouseUVOrig.x * mAVSMDebugViewUVZoomMulAdd.x + mAVSMDebugViewUVZoomMulAdd.z;
            mouseUV.y = mouseUVOrig.y * mAVSMDebugViewUVZoomMulAdd.y + mAVSMDebugViewUVZoomMulAdd.w;

            return CPUT_EVENT_HANDLED;
        }
        if( mLButtonDown ) // position marker
        {
            mAVSMDebugViewMarkerPos.x = mouseUV.x;
            mAVSMDebugViewMarkerPos.y = mouseUV.y;
        }
        if( mRButtonDown ) // start drag move
        {
            mMouseDragStartPos.x = mouseUV.x;
            mMouseDragStartPos.y = mouseUV.y;
            mMouseDragStartPos.z = 1.0f;
        }

        return CPUT_EVENT_HANDLED;
    }
    return CPUT_EVENT_UNHANDLED;
}

// release all textures/buffers (for resize events, etc)
//-----------------------------------------------------------------------------
void AVSMDebugView::ReleaseTexturesAndBuffers()
{
    if( mpAVSMDebugViewMaterial != NULL )
    {
        mpAVSMDebugViewMaterial->ReleaseTexturesAndBuffers( true );
    }
}

// Rebind textures and resize and re-position debug view when window resizes
//-----------------------------------------------------------------------------
void AVSMDebugView::Resize(UINT width, UINT height)
{
    if( mpAVSMDebugViewMaterial != NULL )
    {
        mpAVSMDebugViewMaterial->RebindTexturesAndBuffers( true );
    }
    
    float sprX = (1.0f + c_DebugSpriteX) * 0.5f * width;
    float sprY = (c_DebugSpriteY + 1.0f) * 0.5f * height;
    float sprW = c_DebugSpriteWidth * 0.5f  * width;
    float sprH = c_DebugSpriteHeight * 0.5f * height;

    // todo: set the position of the debug
    mpDebugViewSliceDepth->SetPosition( (int)sprX, (int)sprY - 34 );
    mpDebugViewSliceDepth->SetOverrideWidth( (int)(sprW) - 10 );
}

// Draw the debug view
//-----------------------------------------------------------------------------
void AVSMDebugView::Render(ID3D11DeviceContext* pContext, AVSMTechnique* pAVSMTechnique, AVSMTechnique::FrameContext* pAVSMFrameContext, CPUTRenderParametersDX* pRenderParams)
{
    if( NULL != mpAVSMDebugViewConsts)
    {
        ID3D11Buffer *pBuffer = mpAVSMDebugViewConsts->GetNativeBuffer();

        // update parameters of constant buffer
        D3D11_MAPPED_SUBRESOURCE mapInfo;
        pContext->Map( pBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mapInfo );
        {
            AVSMDebugViewConsts & consts = *((AVSMDebugViewConsts*)mapInfo.pData);

            mpDebugViewSliceDepth->GetValue( consts.SliceDepth );

            consts.MarkerX = mAVSMDebugViewMarkerPos.x;
            consts.MarkerY = mAVSMDebugViewMarkerPos.y;

            consts.UVZoomMulAdd = mAVSMDebugViewUVZoomMulAdd;

            pContext->Unmap(pBuffer,0);

            pContext->VSSetConstantBuffers( 4, 1, &pBuffer );
            pContext->PSSetConstantBuffers( 4, 1, &pBuffer );
            pContext->CSSetConstantBuffers( 4, 1, &pBuffer );
        }

        pAVSMTechnique->UpdateDebugViewD3D11( *pAVSMFrameContext );

        // set all 'external' states for drawing CPUT geometry
        pAVSMTechnique->UpdateStatesAndConstantsForAVSMUse( *pAVSMFrameContext, true );

        mpDebugSprite->DrawSprite( *pRenderParams, *mpAVSMDebugViewMaterial );

        ID3D11ShaderResourceView* nullViews[32] = {0};
        pContext->PSSetShaderResources(0, 32, nullViews);
    }
}