//--------------------------------------------------------------------------------------
// Copyright 2011-2013 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//--------------------------------------------------------------------------------------
#include "AVSMTechnique.h"
#include "..\media\Shader\AVSM_def.h"
#include "..\media\Shader\AVSM_Gen_def.h"
#include "ListTexture.h"
#include "Partitions.h"
#include "ShaderUtils.h"
#include <fstream>
#include <iostream>
#include <limits>
#include <sstream>
#include "..\CPUT\CPUT\CPUTTextureDX11.h"
#include "..\CPUT\CPUT\CPUTBufferDX11.h"

#undef max // std::numeric_limits<unsigned int>::max();
#undef min // std::numeric_limits<unsigned int>::min();                      

using std::tr1::shared_ptr;

using namespace ShaderUtils;

static const float EMPTY_NODE = 65504.0f; // max half prec

// Constructor
//-----------------------------------------------------------------------------
AVSMTechnique::AVSMTechnique(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount,
         unsigned int shadowTextureDim,
         unsigned int avsmShadowTextureDim)
    : mShadowTextureDim(shadowTextureDim)
    , mAVSMShadowTextureDim(avsmShadowTextureDim)
    // 50% coverage of screen (assuming resolution of 1680x1050
    // with average depth complexity of 50. Should be more than plenty.
    , mLisTexNodeCount(1 << 24)
    , mLastTime( 0 ) 
{
    // initialize all members
	for (UINT i = 0; i < MAX_SHADER_VARIATIONS; ++i)
		 mAVSMGenDataSRVCPUT[i]=NULL;

    mAVSMGenCtrlSurfaceCPUT = NULL;
    mAVSMTexturesCPUT = NULL;
    for(int ii=0; ii<MAX_SHADER_VARIATIONS; ii++)
    {
       mAVSMGenDataSRVCPUT[ii] = NULL;
    }
    mAVSMConstantBufferCPUT = NULL;
    mAVSMConstantsCPUT = NULL;
    mPerFrameConstantsCPUT = NULL;   
    mParticlePerFrameConstantsCPUT = NULL;

    // constant buffers
    mAVSMConstants = NULL;
    mPerFrameConstants = NULL;
    mParticlePerFrameConstants = NULL;

    mParticleDepthStencilState = NULL;
    mDefaultDepthStencilState = NULL;
    mAVSMCaptureDepthStencilState = NULL;
    mAVSMSampler = NULL;
    mAVSMGenSampler = NULL;
    mParticlePerPassConstants = NULL;
    mListTextureConstants = NULL;
    mLightingBlendState = NULL;
    mGeometryBlendState = NULL;
    mParticleBlendState = NULL;
    mShadowRasterizerState = NULL;
    mRasterizerState = NULL;
    mDoubleSidedRasterizerState = NULL;
    mParticleWireFrameState = NULL;

    mParticleRasterizerState = NULL;
    mFullScreenTriangleVS = NULL;
    mFullScreenTriangleVSReflector = NULL;
    mParticleShadingVS = NULL;
    mParticleShadingVSReflector = NULL;
    mAVSMClearStructuredBufPS = NULL;
    mAVSMClearStructuredBufPSReflector = NULL;
    mParticleAVSMCapturePS = NULL;
    mParticleAVSMCapturePSReflector = NULL;
    mListTexSegmentNodes = NULL;
    mListTexSegmentNodesUAV = NULL;
    mListTexSegmentNodesSRV = NULL;
    mListTexVisibilityNodes = NULL;
    mListTexVisibilityNodesUAV = NULL;
    mListTexVisibilityNodesSRV = NULL;
    mAVSMStructBuf = NULL;

    mAVSMTexturesDebug = NULL;
    mAVSMStructBufUAV = NULL;
    mAVSMStructBufSRV = NULL;

    mParticleShadingTessellationVS = NULL;
    mParticleShadingTessellationVSReflector = NULL;
    mParticleShadingHS = NULL;
    mParticleShadingHSReflector = NULL;
    mParticleShadingPS = NULL;
    mParticleShadingPSReflector = NULL;

    for (int i = 0; i < MAX_SHADER_VARIATIONS; ++i) 
    {
       mAVSMInsertionSortResolvePS[i] = NULL;
       mAVSMInsertionSortResolvePSReflector[i] = NULL;
       mAVSMSinglePassInsertPS[i] = NULL;
       mAVSMSinglePassInsertPSReflector[i] = NULL;
       mParticleShadingPerPixelPS[i] = NULL;
       mParticleShadingPerPixelPSReflector[i] = NULL;
       mParticleAVSM_GenPS[i] = NULL;
       mParticleAVSM_GenPSReflector[i] = NULL;
       mParticleShadingDS[i] = NULL;
       mParticleShadingDSReflector[i] = NULL;
       mParticleShadingFinalVS[i] = NULL;
       mParticleShadingFinalVSReflector[i] = NULL;

       mAVSMGenData[i] = NULL;
       mAVSMGenDataUAV[i] = NULL;
       mAVSMGenDataSRV[i] = NULL;
    }
}

// Destructor
//-----------------------------------------------------------------------------
AVSMTechnique::~AVSMTechnique() 
{
    SAFE_RELEASE(mAVSMGenCtrlSurfaceCPUT);
    SAFE_RELEASE(mAVSMTexturesCPUT);
    for(int ii=0; ii<MAX_SHADER_VARIATIONS; ii++)
    {
        SAFE_RELEASE(mAVSMGenDataSRVCPUT[ii]);
    }
    SAFE_RELEASE(mAVSMConstantBufferCPUT);
    SAFE_RELEASE(mAVSMConstantsCPUT);
    SAFE_RELEASE(mPerFrameConstantsCPUT);    
    SAFE_RELEASE(mParticlePerFrameConstantsCPUT);

    SAFE_RELEASE(mParticleDepthStencilState);
    SAFE_RELEASE(mDefaultDepthStencilState);
    SAFE_RELEASE(mAVSMCaptureDepthStencilState);
    SAFE_RELEASE(mAVSMSampler);
	SAFE_RELEASE(mAVSMGenSampler);
    SAFE_RELEASE(mPerFrameConstants); 
    SAFE_RELEASE(mParticlePerFrameConstants);
    SAFE_RELEASE(mParticlePerPassConstants);
    SAFE_RELEASE(mListTextureConstants);
    SAFE_RELEASE(mAVSMConstants);     
    SAFE_RELEASE(mLightingBlendState);
    SAFE_RELEASE(mGeometryBlendState);
    SAFE_RELEASE(mParticleBlendState);
    SAFE_RELEASE(mShadowRasterizerState);
    SAFE_RELEASE(mRasterizerState);
    SAFE_RELEASE(mDoubleSidedRasterizerState);
	SAFE_RELEASE(mParticleWireFrameState);

    SAFE_RELEASE(mParticleRasterizerState);
    SAFE_RELEASE(mFullScreenTriangleVS);
    SAFE_RELEASE(mFullScreenTriangleVSReflector);
    SAFE_RELEASE(mParticleShadingVS);
    SAFE_RELEASE(mParticleShadingVSReflector);
    SAFE_RELEASE(mAVSMClearStructuredBufPS);
    SAFE_RELEASE(mAVSMClearStructuredBufPSReflector);
    SAFE_RELEASE(mParticleAVSMCapturePS);
    SAFE_RELEASE(mParticleAVSMCapturePSReflector);
    SAFE_RELEASE(mListTexSegmentNodes);
    SAFE_RELEASE(mListTexSegmentNodesUAV);
    SAFE_RELEASE(mListTexSegmentNodesSRV);
    SAFE_RELEASE(mListTexVisibilityNodes);
    SAFE_RELEASE(mListTexVisibilityNodesUAV);
    SAFE_RELEASE(mListTexVisibilityNodesSRV);
    SAFE_RELEASE(mAVSMStructBuf);
    
    SAFE_RELEASE(mAVSMTexturesDebug);
    SAFE_RELEASE(mAVSMStructBufUAV);
    SAFE_RELEASE(mAVSMStructBufSRV);

    SAFE_RELEASE(mParticleShadingTessellationVS);
    SAFE_RELEASE(mParticleShadingTessellationVSReflector);
    SAFE_RELEASE(mParticleShadingHS);
    SAFE_RELEASE(mParticleShadingHSReflector);
    SAFE_RELEASE(mParticleShadingPS);
    SAFE_RELEASE(mParticleShadingPSReflector);

    for (int i = 0; i < MAX_SHADER_VARIATIONS; ++i) {
        SAFE_RELEASE(mAVSMInsertionSortResolvePS[i]);
        SAFE_RELEASE(mAVSMInsertionSortResolvePSReflector[i]);
        SAFE_RELEASE(mAVSMSinglePassInsertPS[i]);
        SAFE_RELEASE(mAVSMSinglePassInsertPSReflector[i]);
        SAFE_RELEASE(mParticleShadingPerPixelPS[i]);
        SAFE_RELEASE(mParticleShadingPerPixelPSReflector[i]);
        SAFE_RELEASE(mParticleAVSM_GenPS[i]);
        SAFE_RELEASE(mParticleAVSM_GenPSReflector[i]);
		SAFE_RELEASE(mParticleShadingDS[i]);
		SAFE_RELEASE(mParticleShadingDSReflector[i]);
		SAFE_RELEASE(mParticleShadingFinalVS[i]);
		SAFE_RELEASE(mParticleShadingFinalVSReflector[i]);

		SAFE_RELEASE(mAVSMGenData[i]);
		SAFE_RELEASE(mAVSMGenDataUAV[i]);
		SAFE_RELEASE(mAVSMGenDataSRV[i]);
    }
}

//  Load the custom shaders needed for the AVSM technique
//-----------------------------------------------------------------------------
bool AVSMTechnique::LoadAVSMShaders(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount,
         unsigned int shadowTextureDim,
         unsigned int avsmShadowTextureDim)
{
    mAVSMNodeCount = nodeCount;
    mShaderIdx = (nodeCount+3) / 4 - 1;
    
    // Create lighting pass shaders
    ShaderUtils::CreateVertexShaderFromCompiledObj(d3dDevice,
                                      L".\\media\\Shader\\ShaderObjs\\FullScreenTriangleVS.fxo",
                                      &mFullScreenTriangleVS,
                                      &mFullScreenTriangleVSReflector);

    // Create AVSM shaders
    ShaderUtils::CreatePixelShadersFromCompiledObjs(d3dDevice, MAX_SHADER_VARIATIONS, 
                                       ".\\media\\Shader\\ShaderObjs\\ParticleAVSM_Gen", 
                                       "PS.fxo", 
                                       mParticleAVSM_GenPS, 
                                       mParticleAVSM_GenPSReflector);

    ShaderUtils::CreatePixelShaderFromCompiledObj(d3dDevice,
                                     L".\\media\\Shader\\ShaderObjs\\ParticleAVSMCapturePS.fxo",
                                     &mParticleAVSMCapturePS,
                                     &mParticleAVSMCapturePSReflector);

    ShaderUtils::CreatePixelShadersFromCompiledObjs(d3dDevice, MAX_SHADER_VARIATIONS, 
                                       ".\\media\\Shader\\ShaderObjs\\AVSM", 
                                       "InsertionSortResolvePS.fxo", 
                                       mAVSMInsertionSortResolvePS, 
                                       mAVSMInsertionSortResolvePSReflector);

    // Create other volume shadowing techniques shaders
	ShaderUtils::CreatePixelShaderFromCompiledObj(d3dDevice,  
                                       L".\\media\\Shader\\ShaderObjs\\DynamicParticlesShadingPS.fxo", 
                                       &mParticleShadingPS, 
                                       &mParticleShadingPSReflector);


    ShaderUtils::CreateVertexShaderFromCompiledObj(d3dDevice,
                                      L".\\media\\Shader\\ShaderObjs\\DynamicParticlesShadingVS.fxo",
                                      &mParticleShadingVS,
                                      &mParticleShadingVSReflector); 

    ShaderUtils::CreatePixelShadersFromCompiledObjs(d3dDevice, MAX_SHADER_VARIATIONS, 
                                       ".\\media\\Shader\\ShaderObjs\\DynamicParticlesPerPixelShading", 
                                       "PS.fxo", 
                                       mParticleShadingPerPixelPS, 
                                       mParticleShadingPerPixelPSReflector);

    ShaderUtils::CreatePixelShadersFromCompiledObjs(d3dDevice, MAX_SHADER_VARIATIONS, 
                                       ".\\media\\Shader\\ShaderObjs\\ParticleAVSM", 
                                       "SinglePassInsertPS.fxo", 
                                       mAVSMSinglePassInsertPS, 
                                       mAVSMSinglePassInsertPSReflector);

    ShaderUtils::CreatePixelShaderFromCompiledObj(d3dDevice,
                                    L".\\media\\Shader\\ShaderObjs\\AVSMClearStructuredBufPS.fxo",
                                    &mAVSMClearStructuredBufPS,
                                    &mAVSMClearStructuredBufPSReflector); 

	// Alterative path with Tessellation
	ShaderUtils::CreateVertexShadersFromCompiledObjs(d3dDevice, MAX_SHADER_VARIATIONS,
                                      ".\\media\\Shader\\ShaderObjs\\DynamicParticlesShadingFinal",
                                       "VS.fxo", 
                                      mParticleShadingFinalVS,
                                      mParticleShadingFinalVSReflector); 


    ShaderUtils::CreateVertexShaderFromCompiledObj(d3dDevice,
                                      L".\\media\\Shader\\ShaderObjs\\DynamicParticlesShadingTessellationVS.fxo",
                                      &mParticleShadingTessellationVS,
                                      &mParticleShadingTessellationVSReflector); 

    ShaderUtils::CreateDomainShadersFromCompiledObjs(d3dDevice, MAX_SHADER_VARIATIONS, 
                                      ".\\media\\Shader\\ShaderObjs\\DynamicParticlesShading",
                                       "DS.fxo", 
                                      mParticleShadingDS,
                                      mParticleShadingDSReflector); 

	ShaderUtils::CreateHullShaderFromCompiledObj(d3dDevice,
                                      L".\\media\\Shader\\ShaderObjs\\DynamicParticlesShadingHS.fxo",
                                      &mParticleShadingHS,
                                      &mParticleShadingHSReflector); 

    return true;
}


// Create the render states required for the AVSM technique
//-----------------------------------------------------------------------------
bool AVSMTechnique::CreateAVSMRenderStates(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount,
         unsigned int shadowTextureDim,
         unsigned int avsmShadowTextureDim)
{
    // Create standard rasterizer state
    {
        CD3D11_RASTERIZER_DESC desc(D3D11_DEFAULT);
        d3dDevice->CreateRasterizerState(&desc, &mRasterizerState);
    }

    // LD: Create particle wireframe rasterizer state
    {
        CD3D11_RASTERIZER_DESC desc(D3D11_DEFAULT);
        desc.CullMode = D3D11_CULL_NONE;
        desc.DepthClipEnable = false;
		desc.FillMode = D3D11_FILL_WIREFRAME;
        d3dDevice->CreateRasterizerState(&desc, &mParticleWireFrameState);
    }

    // Create double-sided standard rasterizer state
    {
        CD3D11_RASTERIZER_DESC desc(D3D11_DEFAULT);
        desc.CullMode = D3D11_CULL_NONE;
        d3dDevice->CreateRasterizerState(&desc, &mDoubleSidedRasterizerState);
    }

    // Shadow rasterizer state has no back-face culling and multisampling enabled
    {
        CD3D11_RASTERIZER_DESC desc(D3D11_DEFAULT);
        desc.CullMode = D3D11_CULL_NONE;
        desc.MultisampleEnable = true;
        desc.DepthClipEnable = false;
        d3dDevice->CreateRasterizerState(&desc, &mShadowRasterizerState);
    }

    // Create particle rasterizer state
    {
        CD3D11_RASTERIZER_DESC desc(D3D11_DEFAULT);
        desc.CullMode = D3D11_CULL_NONE;
        desc.DepthClipEnable = false;
        d3dDevice->CreateRasterizerState(&desc, &mParticleRasterizerState);
    }

    // Create default depth-stencil state
    {
        CD3D11_DEPTH_STENCIL_DESC desc(D3D11_DEFAULT);
        desc.DepthFunc = D3D11_COMPARISON_LESS_EQUAL;
        d3dDevice->CreateDepthStencilState(&desc, &mDefaultDepthStencilState);
    }

    // Create particle depth-stencil state
    {
        CD3D11_DEPTH_STENCIL_DESC desc(D3D11_DEFAULT);
        desc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;
        desc.DepthEnable = false;
        desc.DepthFunc = D3D11_COMPARISON_ALWAYS;
        d3dDevice->CreateDepthStencilState(&desc, &mParticleDepthStencilState);
    }

    // Create particle depth-stencil state
    {
        CD3D11_DEPTH_STENCIL_DESC desc(D3D11_DEFAULT);
        desc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;
        desc.DepthEnable    = false;
        d3dDevice->CreateDepthStencilState(&desc, &mAVSMCaptureDepthStencilState);
    }

    // Create geometry phase blend state
    {
        CD3D11_BLEND_DESC desc(D3D11_DEFAULT);
        d3dDevice->CreateBlendState(&desc, &mGeometryBlendState);
    }


    // Create lighting phase blend state
    {
        CD3D11_BLEND_DESC desc(D3D11_DEFAULT);
        // Additive blending
        desc.RenderTarget[0].BlendEnable = true;
        desc.RenderTarget[0].SrcBlend = D3D11_BLEND_ONE;
        desc.RenderTarget[0].DestBlend = D3D11_BLEND_ONE;
        desc.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
        d3dDevice->CreateBlendState(&desc, &mLightingBlendState);
    }

    // Create alpha phase blend state
    {
        CD3D11_BLEND_DESC desc(D3D11_DEFAULT);
        desc.RenderTarget[0].BlendEnable = true;
        desc.RenderTarget[0].SrcBlend = D3D11_BLEND_SRC_ALPHA;
        desc.RenderTarget[0].DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
        desc.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
        d3dDevice->CreateBlendState(&desc, &mParticleBlendState);
    }

    // Create constant buffers
    {
        CD3D11_BUFFER_DESC desc(
            sizeof(PerFrameConstants),
            D3D11_BIND_CONSTANT_BUFFER,
            D3D11_USAGE_DYNAMIC,
            D3D11_CPU_ACCESS_WRITE);

        d3dDevice->CreateBuffer(&desc, 0, &mPerFrameConstants);
    }

    {
        CD3D11_BUFFER_DESC desc(
            sizeof(ParticlePerFrameConstants),
            D3D11_BIND_CONSTANT_BUFFER,
            D3D11_USAGE_DYNAMIC,
            D3D11_CPU_ACCESS_WRITE);

        d3dDevice->CreateBuffer(&desc, 0, &mParticlePerFrameConstants);
    }
    {
        CD3D11_BUFFER_DESC desc(
            sizeof(ParticlePerPassConstants),
            D3D11_BIND_CONSTANT_BUFFER,
            D3D11_USAGE_DYNAMIC,
            D3D11_CPU_ACCESS_WRITE);

        d3dDevice->CreateBuffer(&desc, 0, &mParticlePerPassConstants);
    }

    {
        CD3D11_BUFFER_DESC desc(
            sizeof(LT_Constants),
            D3D11_BIND_CONSTANT_BUFFER,
            D3D11_USAGE_DYNAMIC,
            D3D11_CPU_ACCESS_WRITE);

        d3dDevice->CreateBuffer(&desc, 0, &mListTextureConstants);
    }
    
    {
        CD3D11_BUFFER_DESC desc(
            sizeof(AVSMConstants),
            D3D11_BIND_CONSTANT_BUFFER,
            D3D11_USAGE_DYNAMIC,
            D3D11_CPU_ACCESS_WRITE);

        d3dDevice->CreateBuffer(&desc, 0, &mAVSMConstants);
    }

    // Create sampler states
    {
        CD3D11_SAMPLER_DESC desc(D3D11_DEFAULT);
        desc.Filter = D3D11_FILTER_MIN_MAG_MIP_POINT;
        desc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
        desc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
        desc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
        d3dDevice->CreateSamplerState(&desc, &mAVSMSampler);
    }
    
    {
        float borderColor[4] = {0.f, 0.f, 0.f, 0.f};
        CD3D11_SAMPLER_DESC desc(D3D11_DEFAULT);
        desc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
        desc.AddressU = D3D11_TEXTURE_ADDRESS_BORDER;
        desc.AddressV = D3D11_TEXTURE_ADDRESS_BORDER;
        desc.AddressW = D3D11_TEXTURE_ADDRESS_BORDER;
        *desc.BorderColor = *borderColor;
        d3dDevice->CreateSamplerState(&desc, &mAVSMGenSampler);
    }


    return true;
}

// Wrap specific constant buffers needed when drawing CPUT models.  By wrapping 
// them with a string name, the autobinding feature will auto-bind the constant 
// buffer for VS and PS shaders
//-----------------------------------------------------------------------------
void AVSMTechnique::WrapBuffers()
{
    // The autobinding directive is found infound in the *.mtl files
    // The string name specified here is whatever you'd like to use internally, but
    // the name specified in the .mtl file is the name that needs to match what is in the .fx file
    // so, syntax in .mtl file is:
    // [.fx buffer name] = [CPUTBufferDX11 internal buffer name]
    
    // Create the per-frame AVSM data constant buffer.
    {        
        D3D11_BUFFER_DESC bd = {0};
        bd.ByteWidth = sizeof(XMVECTOR);
        bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
        bd.Usage = D3D11_USAGE_DYNAMIC;
        bd.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

        ID3D11Buffer *pTempConstantBuffer;
        HRESULT hr = (CPUT_DX11::GetDevice())->CreateBuffer( &bd, NULL, &pTempConstantBuffer );
        ASSERT( !FAILED( hr ), _L("Error creating constant buffer.") );
        CPUTSetDebugName( pTempConstantBuffer, _L("Per-Frame Constant buffer") );
        
        cString name = _L("$cbAVSMValues");
        mAVSMConstantBufferCPUT = new CPUTBufferDX11( name, pTempConstantBuffer );
        CPUTAssetLibrary::GetAssetLibrary()->AddConstantBuffer( name, mAVSMConstantBufferCPUT );
        SAFE_RELEASE(pTempConstantBuffer); // We're done with it.  The CPUTBuffer now owns it
    }
    
    // 'cbuffer AVSMConstants' found in AVSM.hlsl, data mirrored by 'struct AVSMConstants' in AppShaderConstant.h
    {
        // Wrap the per-frame AVSM data constant buffer.
        CPUTSetDebugName( mAVSMConstants, _L("AVSMConstants constant buffer in AVSM.hlsl") );
        cString name = _L("$cbAVSMConstants");
        mAVSMConstantsCPUT = new CPUTBufferDX11( name, mAVSMConstants );
        CPUTAssetLibrary::GetAssetLibrary()->AddConstantBuffer( name, mAVSMConstantsCPUT );        
    }

    // 'cbuffer PerFrameConstants' found in Gbuffer.hlsl, data mirrored by 'struct PerFrameConstants' in AppShaderConstant.h
    {
        // Wrap the per-frame AVSM data constant buffer.
        CPUTSetDebugName( mPerFrameConstants, _L("PerFrameConstants constant buffer in Gbuffer.hlsl") );
        cString name = _L("$cbPerFrameConstants");
        mPerFrameConstantsCPUT = new CPUTBufferDX11( name, mPerFrameConstants );
        CPUTAssetLibrary::GetAssetLibrary()->AddConstantBuffer( name, mPerFrameConstantsCPUT );
    }

    // 'cbuffer ParticlePerFrameConstants' found in Gbuffer.hlsl, data mirrored by 'struct ParticlePerFrameConstants' in ConstantBuffers.hlsl
    {
        // Create the per-frame AVSM data constant buffer.
        CPUTSetDebugName( mParticlePerFrameConstants, _L("ParticlePerFrameConstants constant buffer in Gbuffer.hlsl") );
        cString name = _L("$cbParticlePerFrameConstants");
        mParticlePerFrameConstantsCPUT = new CPUTBufferDX11( name, mParticlePerFrameConstants );
        CPUTAssetLibrary::GetAssetLibrary()->AddConstantBuffer( name, mParticlePerFrameConstantsCPUT );        
    }
}

// Create the DX buffers needed by AVSM 
//-----------------------------------------------------------------------------
bool AVSMTechnique::CreateAVSMBuffers(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount,
         unsigned int avsmShadowTextureDim)
{
    // Create AVSM textures and viewport
    {
        DXGI_SAMPLE_DESC sampleDesc;
        sampleDesc.Count = 1;
        sampleDesc.Quality = 0;
        mAVSMTextures = std::shared_ptr<Texture2D>(new Texture2D(
            d3dDevice, mAVSMShadowTextureDim, mAVSMShadowTextureDim, DXGI_FORMAT_R32G32B32A32_FLOAT,
            D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE, MAX_AVSM_RT_COUNT, sampleDesc,
            D3D11_RTV_DIMENSION_TEXTURE2DARRAY, D3D11_UAV_DIMENSION_TEXTURE2DARRAY, D3D11_SRV_DIMENSION_TEXTURE2DARRAY));

        mAVSMShadowViewport.Width    = static_cast<float>(mAVSMShadowTextureDim);
        mAVSMShadowViewport.Height   = static_cast<float>(mAVSMShadowTextureDim);
        mAVSMShadowViewport.MinDepth = 0.0f;
        mAVSMShadowViewport.MaxDepth = 1.0f;
        mAVSMShadowViewport.TopLeftX = 0.0f;
        mAVSMShadowViewport.TopLeftY = 0.0f;
    }

    // Create AVSM debug textures
    {
        HRESULT hr;
        CD3D11_TEXTURE2D_DESC desc(
            DXGI_FORMAT_R32G32B32A32_FLOAT,         
            mAVSMShadowTextureDim,
            mAVSMShadowTextureDim,
            MAX_AVSM_RT_COUNT,
            1,
            0,
            D3D11_USAGE_STAGING,
            D3D10_CPU_ACCESS_READ);
        V(d3dDevice->CreateTexture2D(&desc, 0, &mAVSMTexturesDebug));
    }

    {
        HRESULT hr;
        UINT structSize = sizeof(float) * 2 * mAVSMNodeCount;

        CD3D11_BUFFER_DESC desc(
            structSize * mAVSMShadowTextureDim * mAVSMShadowTextureDim,            
            D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS,
            D3D11_USAGE_DEFAULT,
            0,
            D3D11_RESOURCE_MISC_BUFFER_STRUCTURED,
            structSize);
        V(d3dDevice->CreateBuffer(&desc, 0, &mAVSMStructBuf));

        CD3D11_UNORDERED_ACCESS_VIEW_DESC unorderedAccessResourceDesc(
            D3D11_UAV_DIMENSION_BUFFER,
            DXGI_FORMAT_UNKNOWN,
            0, mAVSMShadowTextureDim * mAVSMShadowTextureDim, 1, 0);
        V(d3dDevice->CreateUnorderedAccessView(mAVSMStructBuf, &unorderedAccessResourceDesc, &mAVSMStructBufUAV));

        CD3D11_SHADER_RESOURCE_VIEW_DESC shaderResourceDesc(
            D3D11_SRV_DIMENSION_BUFFER,
            DXGI_FORMAT_UNKNOWN,
            0, mAVSMShadowTextureDim * mAVSMShadowTextureDim, 1);
        V(d3dDevice->CreateShaderResourceView(mAVSMStructBuf, &shaderResourceDesc, &mAVSMStructBufSRV));
    }

    // Create List Texture first segment node offset texture
    {
        DXGI_SAMPLE_DESC sampleDesc;
        sampleDesc.Count = 1;
        sampleDesc.Quality = 0;
        mListTexFirstSegmentNodeOffset = std::shared_ptr<Texture2D>(new Texture2D(
            d3dDevice, mAVSMShadowTextureDim, mAVSMShadowTextureDim, DXGI_FORMAT_R32_UINT,
            D3D11_BIND_UNORDERED_ACCESS | D3D11_BIND_SHADER_RESOURCE, 1, sampleDesc,
            D3D11_RTV_DIMENSION_UNKNOWN, D3D11_UAV_DIMENSION_TEXTURE2D, D3D11_SRV_DIMENSION_TEXTURE2D));
    }

    // Create List Texture first visibility node offset texture
    {
        DXGI_SAMPLE_DESC sampleDesc;
        sampleDesc.Count = 1;
        sampleDesc.Quality = 0;
        mListTexFirstVisibilityNodeOffset = std::shared_ptr<Texture2D>(new Texture2D(
            d3dDevice, mAVSMShadowTextureDim, mAVSMShadowTextureDim, DXGI_FORMAT_R32_UINT,
            D3D11_BIND_UNORDERED_ACCESS | D3D11_BIND_SHADER_RESOURCE, 1, sampleDesc,
            D3D11_RTV_DIMENSION_UNKNOWN, D3D11_UAV_DIMENSION_TEXTURE2D, D3D11_SRV_DIMENSION_TEXTURE2D));
    }

    // Create List Texture segment nodes buffer for AVSM data capture
    {
        HRESULT hr;
        UINT structSize = sizeof(SegmentNode);

        CD3D11_BUFFER_DESC desc(
            structSize * mLisTexNodeCount,            
            D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS,
            D3D11_USAGE_DEFAULT,
            0,
            D3D11_RESOURCE_MISC_BUFFER_STRUCTURED,
            structSize);
        V(d3dDevice->CreateBuffer(&desc, 0, &mListTexSegmentNodes));

        CD3D11_UNORDERED_ACCESS_VIEW_DESC unorderedAccessResourceDesc(
            D3D11_UAV_DIMENSION_BUFFER,
            DXGI_FORMAT_UNKNOWN,
            0, mLisTexNodeCount, 1, D3D11_BUFFER_UAV_FLAG_COUNTER);
        V(d3dDevice->CreateUnorderedAccessView(mListTexSegmentNodes, &unorderedAccessResourceDesc, &mListTexSegmentNodesUAV));

        CD3D11_SHADER_RESOURCE_VIEW_DESC shaderResourceDesc(
            D3D11_SRV_DIMENSION_BUFFER,
            DXGI_FORMAT_UNKNOWN,
            0, mLisTexNodeCount, 1);
        V(d3dDevice->CreateShaderResourceView(mListTexSegmentNodes, &shaderResourceDesc, &mListTexSegmentNodesSRV));
    }

    // Create List Texture visibility nodes buffer for AVSM data capture
    {
        HRESULT hr;
        UINT structSize = sizeof(VisibilityNode);

        CD3D11_BUFFER_DESC desc(
            structSize * mLisTexNodeCount,            
            D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS,
            D3D11_USAGE_DEFAULT,
            0,
            D3D11_RESOURCE_MISC_BUFFER_STRUCTURED,
            structSize);
        V(d3dDevice->CreateBuffer(&desc, 0, &mListTexVisibilityNodes));

        CD3D11_UNORDERED_ACCESS_VIEW_DESC unorderedAccessResourceDesc(
            D3D11_UAV_DIMENSION_BUFFER,
            DXGI_FORMAT_UNKNOWN,
            0, mLisTexNodeCount, 1, D3D11_BUFFER_UAV_FLAG_COUNTER);
        V(d3dDevice->CreateUnorderedAccessView(mListTexVisibilityNodes, &unorderedAccessResourceDesc, &mListTexVisibilityNodesUAV));

        CD3D11_SHADER_RESOURCE_VIEW_DESC shaderResourceDesc(
            D3D11_SRV_DIMENSION_BUFFER,
            DXGI_FORMAT_UNKNOWN,
            0, mLisTexNodeCount, 1);
        V(d3dDevice->CreateShaderResourceView(mListTexVisibilityNodes, &shaderResourceDesc, &mListTexVisibilityNodesSRV));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////
    // AVSM Gen resources
    ////////////////////////////////////////////////////////////////////////////////////////////
    {
        DXGI_SAMPLE_DESC sampleDesc;
        sampleDesc.Count = 1;
        sampleDesc.Quality = 0;
        mAVSMGenCtrlSurface = std::shared_ptr<Texture2D>(new Texture2D(
            d3dDevice, mAVSMShadowTextureDim, mAVSMShadowTextureDim, DXGI_FORMAT_R32_FLOAT,
            D3D11_BIND_UNORDERED_ACCESS | D3D11_BIND_SHADER_RESOURCE, 1, sampleDesc,
            D3D11_RTV_DIMENSION_UNKNOWN, D3D11_UAV_DIMENSION_TEXTURE2D, D3D11_SRV_DIMENSION_TEXTURE2D));
    }

    // Create the mirrored CPUT container texture for mAVSMTextures
	if(!mAVSMGenCtrlSurfaceCPUT)
    {           
        // make an internal/system-generated texture
        cString name = _L("$gAVSMGenCtrlSurface");                                         // $ name indicates not loaded from file
        mAVSMGenCtrlSurfaceCPUT = new CPUTTextureDX11(name);                      
        
        // get needed pointers
        ID3D11Texture2D *pSourceTexture2D = mAVSMGenCtrlSurface->GetTexture();
        ID3D11ShaderResourceView* avsmGenCtrlSurface = mAVSMGenCtrlSurface->GetShaderResource();

        // wrap the previously created objects
        mAVSMGenCtrlSurfaceCPUT->SetTextureAndShaderResourceView(pSourceTexture2D, avsmGenCtrlSurface);
    }

	
	 for (UINT i = 0; i < MAX_SHADER_VARIATIONS; ++i)
	 {
        HRESULT hr;
        UINT structSize = 16 * (i+1);

        CD3D11_BUFFER_DESC desc(
            structSize * mAVSMShadowTextureDim * mAVSMShadowTextureDim,            
            D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS,
            D3D11_USAGE_DEFAULT,
            0,
            D3D11_RESOURCE_MISC_BUFFER_STRUCTURED,
            structSize);
        V(d3dDevice->CreateBuffer(&desc, 0, &mAVSMGenData[i]));

        CD3D11_UNORDERED_ACCESS_VIEW_DESC unorderedAccessResourceDesc(
            D3D11_UAV_DIMENSION_BUFFER,
            DXGI_FORMAT_UNKNOWN,
            0, mAVSMShadowTextureDim * mAVSMShadowTextureDim, 1, 0);
        V(d3dDevice->CreateUnorderedAccessView(mAVSMGenData[i], &unorderedAccessResourceDesc, &mAVSMGenDataUAV[i]));

        CD3D11_SHADER_RESOURCE_VIEW_DESC shaderResourceDesc(
            D3D11_SRV_DIMENSION_BUFFER,
            DXGI_FORMAT_UNKNOWN,
            0, mAVSMShadowTextureDim * mAVSMShadowTextureDim, 1);
        V(d3dDevice->CreateShaderResourceView(mAVSMGenData[i], &shaderResourceDesc, &mAVSMGenDataSRV[i]));       
    }
    return true;
}

//
//-----------------------------------------------------------------------------
bool AVSMTechnique::DeleteAVSMBuffers(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount)
{
	 for (UINT i = 0; i < MAX_SHADER_VARIATIONS; ++i)
	 {
		 SAFE_RELEASE(mAVSMGenDataSRV[i]);
		 SAFE_RELEASE(mAVSMGenDataUAV[i]);
		 SAFE_RELEASE(mAVSMGenData[i]);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////
    // AVSM Gen resources
    ////////////////////////////////////////////////////////////////////////////////////////////
	mAVSMGenCtrlSurface.reset();

   // Create List Texture visibility nodes buffer for AVSM data capture
    {
        SAFE_RELEASE(mListTexVisibilityNodesSRV);
        SAFE_RELEASE(mListTexVisibilityNodesUAV);
        SAFE_RELEASE(mListTexVisibilityNodes);
    }

    // Create List Texture segment nodes buffer for AVSM data capture
    {
        SAFE_RELEASE(mListTexSegmentNodesSRV);
        SAFE_RELEASE(mListTexSegmentNodesUAV);
        SAFE_RELEASE(mListTexSegmentNodes);
    }

    // Create List Texture first segment node offset texture
	mListTexFirstSegmentNodeOffset.reset();

    // Create List Texture first visibility node offset texture
	mListTexFirstVisibilityNodeOffset.reset();
  
    // Create List texture first offset debug texture
//	SAFE_RELEASE(mListTexFirstOffsetDebug);

    {
        SAFE_RELEASE(mAVSMStructBufSRV);
        SAFE_RELEASE(mAVSMStructBufUAV);
        SAFE_RELEASE(mAVSMStructBuf);
    }

 
    // Create AVSM debug textures
    {
 		SAFE_RELEASE(mAVSMTexturesDebug);
    }

    // Create AVSM textures and viewport
    mAVSMTextures.reset();

    return true;
}

// Cleanup (aka make the runtime happy)
//-----------------------------------------------------------------------------
void AVSMTechnique::Cleanup(ID3D11DeviceContext* d3dDeviceContext)
{
    d3dDeviceContext->VSSetShader(NULL, 0, 0);
    d3dDeviceContext->HSSetShader(NULL, 0, 0);
    d3dDeviceContext->DSSetShader(NULL, 0, 0);
    d3dDeviceContext->GSSetShader(NULL, 0, 0);

    d3dDeviceContext->OMSetRenderTargets(0, 0, 0);       
    d3dDeviceContext->OMSetRenderTargetsAndUnorderedAccessViews(0, 0, 0, 0, 0, 0, 0);

    // should go from 16 to 31 instead of 0 to 31 so we don't stomp on CPUT stuff but I haven't tested it
    ID3D11ShaderResourceView* nullViews[33] = {0};
    d3dDeviceContext->VSSetShaderResources(0, 33, nullViews);
    d3dDeviceContext->GSSetShaderResources(0, 33, nullViews);
    d3dDeviceContext->PSSetShaderResources(0, 33, nullViews);
    d3dDeviceContext->CSSetShaderResources(0, 33, nullViews);
    d3dDeviceContext->DSSetShaderResources(0, 33, nullViews);
    d3dDeviceContext->HSSetShaderResources(0, 33, nullViews);

    ID3D11UnorderedAccessView* nullUAViews[8] = {0, 0, 0, 0, 0, 0, 0, 0};
    d3dDeviceContext->CSSetUnorderedAccessViews(0, 8, nullUAViews, 0);
                                               
}

//
//-----------------------------------------------------------------------------
void AVSMTechnique::SetNodeCount(int nodeCount)
{
    mAVSMNodeCount = nodeCount;
    mShaderIdx = nodeCount / 4 - 1;
}

//
//-----------------------------------------------------------------------------
static void TransformBBox(D3DXVECTOR3 *min,
              D3DXVECTOR3 *max,
              const D3DXMATRIXA16 &m)
{
    const D3DXVECTOR3 &minCorner = *min;
    const D3DXVECTOR3 &maxCorner = *max;
    D3DXVECTOR3 corners[8];
    // Bottom corners
    corners[0] = D3DXVECTOR3(minCorner.x, minCorner.y, minCorner.z);
    corners[1] = D3DXVECTOR3(maxCorner.x, minCorner.y, minCorner.z);
    corners[2] = D3DXVECTOR3(maxCorner.x, minCorner.y, maxCorner.z);
    corners[3] = D3DXVECTOR3(minCorner.x, minCorner.y, maxCorner.z);
    // Top corners
    corners[4] = D3DXVECTOR3(minCorner.x, maxCorner.y, minCorner.z);
    corners[5] = D3DXVECTOR3(maxCorner.x, maxCorner.y, minCorner.z);
    corners[6] = D3DXVECTOR3(maxCorner.x, maxCorner.y, maxCorner.z);
    corners[7] = D3DXVECTOR3(minCorner.x, maxCorner.y, maxCorner.z);

    D3DXVECTOR3 newCorners[8];
    for (int i = 0; i < 8; i++) {
        D3DXVec3TransformCoord(&newCorners[i], &corners[i], &m);
    }

    D3DXVECTOR3 newMin, newMax;

    // Initialize
    for (int i = 0; i < 3; ++i) {
        newMin[i] = std::numeric_limits<float>::max();
        newMax[i] = std::numeric_limits<float>::min();
    }

    // Find new min and max corners
    for (int i = 0; i < 8; i++) {
        D3DXVec3Minimize(&newMin, &newMin, &newCorners[i]);
        D3DXVec3Maximize(&newMax, &newMax, &newCorners[i]);
    }

    *min = newMin;
    *max = newMax;
}

//
//-----------------------------------------------------------------------------
void AVSMTechnique::InitializeFrameContext( FrameContext & outContext, const Options & options, ID3D11DeviceContext* d3dDeviceContext, ParticleSystem * particleSystem, 
    const D3DXMATRIX & worldMatrix, CPUTCamera* viewerCamera, CPUTCamera* lightCamera, const D3D11_VIEWPORT & viewport )
{
    outContext.DepthBufferSRV   = NULL;
    outContext.D3dDeviceContext = d3dDeviceContext;
    outContext.Options          = options;
    outContext.ParticleSystem   = particleSystem;
    outContext.ViewerCamera     = viewerCamera;
    outContext.LightCamera      = lightCamera;
    outContext.Viewport         = viewport;
    memset( &outContext.Stats, 0, sizeof(outContext.Stats) );

    FrameMatrices & frameMatx = outContext.Matrices;

    frameMatx.worldMatrix = worldMatrix;

    assert((options.enableParticles && particleSystem) ||
        (options.enableParticles == false));

    frameMatx.cameraProj = DXUTFromCPUT( *viewerCamera->GetProjectionMatrix() );
    frameMatx.cameraView = DXUTFromCPUT( *viewerCamera->GetViewMatrix() );

    D3DXMatrixInverse(&frameMatx.cameraViewInv, 0, &frameMatx.cameraView);

    // We only use the view direction from the camera object
    // We then center the directional light on the camera frustum and set the
    // extents to completely cover it.
    frameMatx.lightView = DXUTFromCPUT( *lightCamera->GetViewMatrix() );
    {
        // NOTE: We don't include the projection matrix here, since we want to just get back the
        // raw view-space extents and use that to *build* the bounding projection matrix
        D3DXVECTOR3 min, max;
        ComputeFrustumExtents(frameMatx.cameraViewInv, frameMatx.cameraProj,
            viewerCamera->GetNearPlaneDistance(), viewerCamera->GetFarPlaneDistance(),
            frameMatx.lightView, &min, &max);

        // First adjust the light matrix to be centered on the extents in x/y and behind everything in z
        D3DXVECTOR3 center = 0.5f * (min + max);
        D3DXMATRIXA16 centerTransform;
        D3DXMatrixTranslation(&centerTransform, -center.x, -center.y, 0.0f);
        frameMatx.lightView *= centerTransform;

        // Now create a projection matrix that covers the extents when centered
        // Optimization: Again use scene AABB to decide on light far range - this one can actually clip out
        // any objects further away than the frustum can see if desired.
        D3DXVECTOR3 dimensions = max - min;
        D3DXMatrixOrthoLH(&frameMatx.lightProj, dimensions.x, dimensions.y, 0.0f, 1000.0f);
    }

    // Compute composite matrices
    frameMatx.cameraViewProj = frameMatx.cameraView * frameMatx.cameraProj;
    frameMatx.cameraWorldViewProj = frameMatx.worldMatrix * frameMatx.cameraViewProj;
    frameMatx.cameraWorldView = frameMatx.worldMatrix * frameMatx.cameraView;
    frameMatx.lightViewProj = frameMatx.lightView * frameMatx.lightProj;
    frameMatx.lightWorldViewProj = frameMatx.worldMatrix * frameMatx.lightViewProj;
    frameMatx.cameraViewToLightProj = frameMatx.cameraViewInv * frameMatx.lightViewProj;
    frameMatx.cameraViewToLightView = frameMatx.cameraViewInv * frameMatx.lightView;

    frameMatx.avsmLightProj = DXUTFromCPUT( *lightCamera->GetProjectionMatrix() );
    frameMatx.avsmLightView = DXUTFromCPUT( *lightCamera->GetViewMatrix() );

    if (options.enableAutoBoundsAVSM) {

        // doesn't work with CPUT shadows unfortunately!
        assert( false );

        // Get bounding boxes from transparent geometry
        const float bigNumf = 1e10f;
        D3DXVECTOR3 maxBB(-bigNumf, -bigNumf, -bigNumf);
        D3DXVECTOR3 minBB(+bigNumf, +bigNumf, +bigNumf);

        if (options.enableParticles) {
            // Initialize minBB, maxBB
            for (size_t p = 0; p < 3; ++p) {
                minBB[p] = std::numeric_limits<float>::max();
                maxBB[p] = std::numeric_limits<float>::min();
            }

            D3DXVECTOR3 particleMin, particleMax;
            particleSystem->GetBBox(&particleMin, &particleMax);

            for (size_t p = 0; p < 3; ++p) {
                minBB[p] = std::min(minBB[p], particleMin[p]);
                maxBB[p] = std::max(maxBB[p], particleMax[p]);
            }

            TransformBBox(&minBB, &maxBB, frameMatx.avsmLightView);
        }

        // First adjust the light matrix to be centered on the extents in x/y and behind everything in z
        D3DXVECTOR3 center = 0.5f * (minBB + maxBB);
        D3DXMATRIXA16 centerTransform;
        D3DXMatrixTranslation(&centerTransform, -center.x, -center.y, -minBB.z);
        frameMatx.avsmLightView *= centerTransform;

        // Now create a projection matrix that covers the extents when centered
        // Optimization: Again use scene AABB to decide on light far range - this one can actually clip out
        // any objects further away than the frustum can see if desired.
        D3DXVECTOR3 dimensions = maxBB - minBB;
        D3DXMatrixOrthoLH(&frameMatx.avsmLightProj, dimensions.x, dimensions.y, 0, dimensions.z);
    }

    // Compute composite matrices;
    frameMatx.avsmLightViewProj = frameMatx.avsmLightView * frameMatx.avsmLightProj;
    frameMatx.avmsLightWorldViewProj = frameMatx.worldMatrix * frameMatx.avsmLightViewProj;
    frameMatx.cameraViewToAvsmLightProj = frameMatx.cameraViewInv * frameMatx.avsmLightViewProj;
    frameMatx.cameraViewToAvsmLightView = frameMatx.cameraViewInv * frameMatx.avsmLightView;

    // Just set up some common defaults
    memset( &outContext.GPUUIConstants, 0, sizeof( outContext.GPUUIConstants ) );
    outContext.GPUUIConstants.lightingOnly              = 0;
    outContext.GPUUIConstants.faceNormals               = 0;
    outContext.GPUUIConstants.enableStats               = false;
    outContext.GPUUIConstants.volumeShadowMethod        = VOL_SHADOW_AVSM;
    outContext.GPUUIConstants.enableVolumeShadowLookup  = 1;
    outContext.GPUUIConstants.pauseParticleAnimaton     = 0;
    outContext.GPUUIConstants.particleSize              = 1.0f;
    outContext.GPUUIConstants.particleOpacity           = 33;
    outContext.GPUUIConstants.vertexShaderShadowLookup	= 0;
    outContext.GPUUIConstants.wireframe					= 0;
    outContext.GPUUIConstants.tessellate				= 1;
    outContext.GPUUIConstants.TessellationDensity		= 1.0f/14.0f;
}

//
//-----------------------------------------------------------------------------
void AVSMTechnique::UpdateParticles( FrameContext & frameContext )
{
    const Options & options                 = frameContext.Options;
    CPUTCamera * viewerCamera               = frameContext.ViewerCamera;
    CPUTCamera * lightCamera                = frameContext.LightCamera;
    ParticleSystem * particleSystem         = frameContext.ParticleSystem;
    ID3D11DeviceContext * d3dDeviceContext  = frameContext.D3dDeviceContext;
    FrameMatrices & frameMatx               = frameContext.Matrices;
    UIConstants * ui                        = &frameContext.GPUUIConstants;

    static unsigned int parframecount = 0;
    float particleDepthBounds[2];
    if (options.enableParticles) {
        if (ui->pauseParticleAnimaton == false) {
            // Update particles 
            float deltaTime = 0.009f; //currTime - mLastTime;
            if (parframecount < 100) {
                particleSystem->UpdateParticles(viewerCamera, lightCamera, deltaTime);
            }
        }

        particleSystem->SortParticles(particleDepthBounds, &frameMatx.avsmLightView, false, 1);
        particleSystem->PopulateVertexBuffers(d3dDeviceContext);

        // Fill in particle emitter (per frame) constants
        {
            D3D11_MAPPED_SUBRESOURCE mappedResource;
            d3dDeviceContext->Map(mParticlePerFrameConstantsCPUT->GetNativeBuffer(), 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
            ParticlePerFrameConstants* constants = static_cast<ParticlePerFrameConstants *>(mappedResource.pData);

            constants->mScale                        = 1.0f;
            constants->mParticleSize                 = ((float)ui->particleSize / 3.0f);
            constants->mParticleAlpha                = 1.0f;
            constants->mbSoftParticles               = 1.0f;
            constants->mParticleOpacity              = 0.8f * ((float)ui->particleOpacity / 33.0f);
            constants->mSoftParticlesSaturationDepth = 1.0f;

            d3dDeviceContext->Unmap(mParticlePerFrameConstantsCPUT->GetNativeBuffer(), 0);
        }
    }
}

//
//-----------------------------------------------------------------------------
void AVSMTechnique::CreateAVSMShadowMap( FrameContext & frameContext, CPUTAssetSet *mpAssetSet, CPUTRenderParametersDX* pRenderParams )
{
    const Options & options                 = frameContext.Options;
    ID3D11DeviceContext * d3dDeviceContext  = frameContext.D3dDeviceContext;
    ParticleSystem * particleSystem         = frameContext.ParticleSystem;
    CPUTCamera * viewerCamera               = frameContext.ViewerCamera;
    CPUTCamera * lightCamera                = frameContext.LightCamera;
    D3D11_VIEWPORT * viewport               = &frameContext.Viewport;
    UIConstants * ui                        = &frameContext.GPUUIConstants;
    FrameStats & frameStats                 = frameContext.Stats;
    FrameMatrices & frameMatx               = frameContext.Matrices;
    D3DXVECTOR4 cameraPos  = D3DXVECTOR4( DXUTFromCPUT( viewerCamera->GetPosition() ), 1.0f );

    // no lookup not supported at the moment
    assert( ui->enableVolumeShadowLookup );

    // set all the various camera, light, and per-frame constants in constant buffer
    FillInFrameConstants(d3dDeviceContext, frameMatx, cameraPos, viewerCamera, lightCamera, frameContext.mScreenWidth, frameContext.mScreenHeight, ui);

    unsigned int savedVolMethod = ui->volumeShadowMethod;
    
    // This phase computes a visibility representation for our shadow volume
    if (ui->volumeShadowMethod && options.enableVolumeShadowCreation) 
    {
        // Setup constants buffers
        FillListTextureConstants(d3dDeviceContext);
        FillAVSMConstants(d3dDeviceContext);

        ClearShadowBuffers(d3dDeviceContext, ui);

        // First pass, capture all fragments
        if (options.enableParticles) {
            
            FillParticleRendererConstants( d3dDeviceContext, 
                lightCamera, 
                frameMatx.avsmLightView, 
                frameMatx.avsmLightViewProj);

            CaptureFragments(d3dDeviceContext, particleSystem, ui, true);
        }

        // Second pass, generate visibility curves (AVSM, uncompressed, deep shadow maps, etc..)
        GenerateVisibilityCurve(d3dDeviceContext, ui);         
    }
    
    // no lookup not supported at the moment
    assert( ui->enableVolumeShadowLookup );

}

//
//-----------------------------------------------------------------------------
void AVSMTechnique::RenderShadedParticles( FrameContext & frameContext, ID3D11RenderTargetView* backBuffer, ID3D11DepthStencilView* depthBuffer )
{
    const Options & options                 = frameContext.Options;
    ID3D11DeviceContext * d3dDeviceContext  = frameContext.D3dDeviceContext;
    ParticleSystem * particleSystem         = frameContext.ParticleSystem;
    CPUTCamera * viewerCamera               = frameContext.ViewerCamera;
    CPUTCamera * lightCamera                = frameContext.LightCamera;
    D3D11_VIEWPORT * viewport               = &frameContext.Viewport;
    UIConstants * ui                        = &frameContext.GPUUIConstants;
    FrameStats & frameStats                 = frameContext.Stats;
    FrameMatrices & frameMatx               = frameContext.Matrices;
    D3DXVECTOR4 cameraPos  = D3DXVECTOR4( DXUTFromCPUT( viewerCamera->GetPosition() ), 1.0f );

    // lookup not supported 
    assert( ui->enableVolumeShadowLookup );

    if (options.enableParticles) 
    {
        // Particle Alpha Pass
        // Update particles
        particleSystem->SortParticles(NULL, &frameMatx.cameraView);
        particleSystem->PopulateVertexBuffers(d3dDeviceContext);

        // Fill particle renderer constants and shade particles
        RenderShadedParticles(d3dDeviceContext, particleSystem, backBuffer, depthBuffer, viewport, ui);            
    }

    Cleanup( d3dDeviceContext );
}

//
//-----------------------------------------------------------------------------
void AVSMTechnique::CaptureFragments(ID3D11DeviceContext* d3dDeviceContext, 
                           ParticleSystem *particleSystem,
                           const UIConstants* ui,
                           bool initCounter)
{

    d3dDeviceContext->HSSetShader( NULL, NULL, 0);
    d3dDeviceContext->DSSetShader( NULL, NULL, 0);

    switch(ui->volumeShadowMethod) {
        case VOL_SHADOW_AVSM:
        {
            ID3D11Buffer* ParticlePerFrameConstants = mParticlePerFrameConstantsCPUT->GetNativeBuffer();
            VSSetConstantBuffers(d3dDeviceContext,
                                 mParticleShadingVSReflector,
                                 "ParticlePerFrameConstants",
                                 1, 
                                 &ParticlePerFrameConstants);
            VSSetConstantBuffers(d3dDeviceContext,
                                 mParticleShadingVSReflector,
                                 "ParticlePerPassConstants",
                                 1, 
                                 &mParticlePerPassConstants);
            d3dDeviceContext->VSSetShader(mParticleShadingVS, 0, 0);

            d3dDeviceContext->RSSetState(mParticleRasterizerState);
            d3dDeviceContext->RSSetViewports(1, &mAVSMShadowViewport);


            PSSetConstantBuffers(d3dDeviceContext,
                                 mParticleAVSMCapturePSReflector,
                                 "ParticlePerFrameConstants",
                                 1, 
                                 &ParticlePerFrameConstants);
            PSSetConstantBuffers(d3dDeviceContext,
                                 mParticleAVSMCapturePSReflector,
                                 "LT_Constants",
                                 1, 
                                 &mListTextureConstants);
            d3dDeviceContext->PSSetShader(mParticleAVSMCapturePS, 0, 0);

            ID3D11UnorderedAccessView* listTexFirstSegmentNodeOffsetUAV = 
                mListTexFirstSegmentNodeOffset->GetUnorderedAccess();

            // Set List Texture UAVs (we don't need any RT!)
            static const char *paramUAVs[] = {
                "gListTexFirstSegmentNodeAddressUAV",
                "gListTexSegmentNodesUAV",
            };
            const UINT numUAVs = sizeof(paramUAVs) / sizeof(paramUAVs[0]);
            const UINT firstUAVIndex =
                GetStartBindIndex(mParticleAVSMCapturePSReflector, 
                                  paramUAVs, numUAVs);           
            UINT pUAVInitialCounts[numUAVs] = {0, 0};
            ID3D11UnorderedAccessView* pUAVs[numUAVs] = {
                listTexFirstSegmentNodeOffsetUAV, 
                mListTexSegmentNodesUAV
            };

            d3dDeviceContext->OMSetRenderTargetsAndUnorderedAccessViews(
                0, NULL, // render targets
                NULL,    // depth-stencil
                firstUAVIndex, numUAVs, pUAVs, initCounter ? pUAVInitialCounts : NULL);

            d3dDeviceContext->OMSetBlendState(mGeometryBlendState, 0, 0xFFFFFFFF);
            d3dDeviceContext->OMSetDepthStencilState(mAVSMCaptureDepthStencilState, 0x0);

            particleSystem->Draw(d3dDeviceContext, NULL, 0, particleSystem->GetParticleCount(),false);

            // Cleanup (aka make the runtime happy)
            Cleanup(d3dDeviceContext);

            break;
        }

        case VOL_SHADOW_AVSM_GEN:
        {            
            ID3D11Buffer* ParticlePerFrameConstants = mParticlePerFrameConstantsCPUT->GetNativeBuffer();
            VSSetConstantBuffers(d3dDeviceContext,
                                 mParticleShadingVSReflector,
                                 "ParticlePerFrameConstants",
                                 1, 
                                 &ParticlePerFrameConstants);
            VSSetConstantBuffers(d3dDeviceContext,
                                 mParticleShadingVSReflector,
                                 "ParticlePerPassConstants",
                                 1, 
                                 &mParticlePerPassConstants);
            d3dDeviceContext->VSSetShader(mParticleShadingVS, 0, 0);

            d3dDeviceContext->RSSetState(mParticleRasterizerState);
            d3dDeviceContext->RSSetViewports(1, &mAVSMShadowViewport);

            PSSetConstantBuffers(d3dDeviceContext,
                                 mParticleAVSM_GenPSReflector[mShaderIdx],
                                 "ParticlePerFrameConstants",
                                 1, 
                                 &ParticlePerFrameConstants);
            d3dDeviceContext->PSSetShader(mParticleAVSM_GenPS[mShaderIdx], 0, 0);

            static const char *paramUAVs[] = {
                "gAVSMGenClearMaskUAV",
                "gAVSMGenDataUAV",
            };
            const UINT numUAVs = sizeof(paramUAVs) / sizeof(paramUAVs[0]);
            const UINT firstUAVIndex =
                GetStartBindIndex(mParticleAVSM_GenPSReflector[mShaderIdx], 
                                  paramUAVs, numUAVs);           
            UINT pUAVInitialCounts[numUAVs] = {0, 0};
            ID3D11UnorderedAccessView* pUAVs[numUAVs] = {
                mAVSMGenCtrlSurface->GetUnorderedAccess(), 
                mAVSMGenDataUAV[mShaderIdx],
            };

            d3dDeviceContext->OMSetRenderTargetsAndUnorderedAccessViews(
                0, NULL, // render targets
                NULL,    // depth-stencil
                firstUAVIndex, numUAVs, pUAVs, initCounter ? pUAVInitialCounts : NULL);

            d3dDeviceContext->OMSetBlendState(mGeometryBlendState, 0, 0xFFFFFFFF);
            d3dDeviceContext->OMSetDepthStencilState(mAVSMCaptureDepthStencilState, 0x0);

            particleSystem->Draw(d3dDeviceContext, NULL, 0, particleSystem->GetParticleCount(),false);

            // Cleanup (aka make the runtime happy)
            Cleanup(d3dDeviceContext);
            break;
        }
    }
}

// clear the AVSM shadow buffers
//-----------------------------------------------------------------------------
void AVSMTechnique::ClearShadowBuffers(ID3D11DeviceContext* d3dDeviceContext, const UIConstants* ui)
{
    switch(ui->volumeShadowMethod) {
        case VOL_SHADOW_NO_SHADOW:
        case VOL_SHADOW_AVSM:
        {
            ID3D11UnorderedAccessView* listTexFirstSegmentNodeOffsetUAV = 
                mListTexFirstSegmentNodeOffset->GetUnorderedAccess();

            // Initialize the first node offset RW UAV with a NULL offset (end of the list)
            UINT clearValues[4] = { 0xFFFFFFFFUL, 0xFFFFFFFFUL, 0xFFFFFFFFUL, 0xFFFFFFFFUL };
            d3dDeviceContext->ClearUnorderedAccessViewUint(listTexFirstSegmentNodeOffsetUAV, clearValues);
            break;
        }
        case VOL_SHADOW_AVSM_GEN: {
            // Clear AVSM Gen control surface
            float clearValues[4] = {0.0f, 0.0f, 0.0f, 0.0f};
            d3dDeviceContext->ClearUnorderedAccessViewFloat(mAVSMGenCtrlSurface->GetUnorderedAccess(), clearValues);
            break;
        }
        default:
            break;
    }
}

//
//-----------------------------------------------------------------------------
void AVSMTechnique::CreatePixelShadersFromCompiledObjs(ID3D11Device* d3dDevice,
                                             int shaderCount,
                                             const char* prefix, 
                                             const char* suffix, 
                                             ID3D11PixelShader** shaderArray,
                                             ID3D11ShaderReflection** shaderReflArray)
{
    for (int i = 0; i < shaderCount; i++) 
    {
        char cNodeCount[32];
        _itoa_s((i + 1) * 4, cNodeCount, sizeof(cNodeCount), 10);  

        {
            std::string objName = std::string(prefix) + 
                                  std::string(cNodeCount) + 
                                  std::string(suffix);
            std::wstring objWName(objName.length(), L'');
            std::copy(objName.begin(), objName.end(), objWName.begin());
            CreatePixelShaderFromCompiledObj(d3dDevice,
                                            (LPCTSTR)objWName.c_str(),
                                            &shaderArray[i],
                                            &shaderReflArray[i]);
        }
    }
}   

//
//-----------------------------------------------------------------------------
void AVSMTechnique::CreateDomainShadersFromCompiledObjs(ID3D11Device* d3dDevice,
                                             int shaderCount,
                                             const char* prefix, 
                                             const char* suffix, 
                                             ID3D11DomainShader** shaderArray,
                                             ID3D11ShaderReflection** shaderReflArray)
{
    for (int i = 0; i < shaderCount; i++) 
    {
        char cNodeCount[32];
        _itoa_s((i + 1) * 4, cNodeCount, sizeof(cNodeCount), 10);  

        {
            std::string objName = std::string(prefix) + 
                                  std::string(cNodeCount) + 
                                  std::string(suffix);
            std::wstring objWName(objName.length(), L'');
            std::copy(objName.begin(), objName.end(), objWName.begin());
            CreateDomainShaderFromCompiledObj(d3dDevice,
                                            (LPCTSTR)objWName.c_str(),
                                            &shaderArray[i],
                                            &shaderReflArray[i]);
        }
    }
} 

//
//-----------------------------------------------------------------------------
void AVSMTechnique::CreateVertexShadersFromCompiledObjs(ID3D11Device* d3dDevice,
                                             int shaderCount,
                                             const char* prefix, 
                                             const char* suffix, 
                                             ID3D11VertexShader** shaderArray,
                                             ID3D11ShaderReflection** shaderReflArray)
{
    for (int i = 0; i < shaderCount; i++) 
    {
        char cNodeCount[32];
        _itoa_s((i + 1) * 4, cNodeCount, sizeof(cNodeCount), 10);  

        {
            std::string objName = std::string(prefix) + 
                                  std::string(cNodeCount) + 
                                  std::string(suffix);
            std::wstring objWName(objName.length(), L'');
            std::copy(objName.begin(), objName.end(), objWName.begin());
            CreateVertexShaderFromCompiledObj(d3dDevice,
                                            (LPCTSTR)objWName.c_str(),
                                            &shaderArray[i],
                                            &shaderReflArray[i]);
        }
    }
} 

//
//-----------------------------------------------------------------------------
void AVSMTechnique::GenerateVisibilityCurve(ID3D11DeviceContext* d3dDeviceContext,
                                 const UIConstants* ui)
{
    ID3D11ShaderResourceView*  listTexFirstSegmentNodeOffsetSRV = mListTexFirstSegmentNodeOffset->GetShaderResource();

    if (VOL_SHADOW_AVSM == ui->volumeShadowMethod) {
        // Second (full screen) pass, sort fragments and insert them in our AVSM texture(s)
        d3dDeviceContext->IASetInputLayout(0);
        d3dDeviceContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);
        d3dDeviceContext->IASetVertexBuffers(0, 0, 0, 0, 0);

        d3dDeviceContext->VSSetShader(mFullScreenTriangleVS, 0, 0);

        d3dDeviceContext->RSSetState(mRasterizerState);
        d3dDeviceContext->RSSetViewports(1, &mAVSMShadowViewport);     
        
        ID3D11ShaderReflection *shaderReflector = mAVSMInsertionSortResolvePSReflector[mShaderIdx];
        d3dDeviceContext->PSSetShader(mAVSMInsertionSortResolvePS[mShaderIdx], 0, 0); 
        
        // framework's autobind not in effect on UAV manipulation - manual bind cbuffer
        ID3D11Buffer* pAVSMConstants = mAVSMConstantsCPUT->GetNativeBuffer();
        PSSetConstantBuffers(d3dDeviceContext,
                             shaderReflector,
                             "AVSMConstants",
                             1, 
                             &pAVSMConstants);
         
        PSSetShaderResources(d3dDeviceContext,
                             shaderReflector,
                             "gListTexSegmentNodesSRV",
                             1, 
                             &mListTexSegmentNodesSRV);
        PSSetShaderResources(d3dDeviceContext,
                             shaderReflector,
                             "gListTexFirstSegmentNodeAddressSRV",
                             1, 
                             &listTexFirstSegmentNodeOffsetSRV);
        
        ID3D11RenderTargetView* pRTs[16];
        const int avsmRTCount = mAVSMNodeCount / 2;
        for (int i = 0; i < avsmRTCount; ++i) {
            pRTs[i] = mAVSMTextures->GetRenderTarget(i);
        }
        d3dDeviceContext->OMSetRenderTargets(avsmRTCount, pRTs, 0);
                                                                    
        d3dDeviceContext->OMSetBlendState(mGeometryBlendState, 0, 0xFFFFFFFF);
        d3dDeviceContext->OMSetDepthStencilState(mDefaultDepthStencilState, 0x0);

        // Full-screen triangle
        d3dDeviceContext->Draw(3, 0);

        // Cleanup (aka make the runtime happy)
        Cleanup(d3dDeviceContext);
    }
}

//  This function resets state and updates the constant buffers needed before rendering
// using the AVSM technique
//-----------------------------------------------------------------------------
void AVSMTechnique::UpdateStatesAndConstantsForAVSMUse( FrameContext & frameContext, bool forCPUTGeometry )
{
    ID3D11DeviceContext* d3dDeviceContext   = frameContext.D3dDeviceContext;
    ParticleSystem * particleSystem         = frameContext.ParticleSystem;
    D3D11_VIEWPORT * viewport               = &frameContext.Viewport;
    const UIConstants * ui                  = &frameContext.GPUUIConstants;
    CPUTCamera * viewerCamera               = frameContext.ViewerCamera;
    FrameMatrices & frameMatx               = frameContext.Matrices;
    
    FillParticleRendererConstants(d3dDeviceContext, viewerCamera, frameMatx.cameraView, frameMatx.cameraViewProj);            

    ID3D11ShaderResourceView* asvmTextureSRV = mAVSMTextures->GetShaderResource();
    ID3D11ShaderResourceView* avsmGenCtrlSurface = mAVSMGenCtrlSurface->GetShaderResource();

    if( !forCPUTGeometry )
    {
        PSSetShaderResources(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "gDepthBuffer",
            1, 
            &frameContext.DepthBufferSRV);
    }

    if(ui->vertexShaderShadowLookup && !forCPUTGeometry)
    {
        if(ui->tessellate)
        {            
            //set all hull shader constant buffers
            // current framework doesn't auto-bind GS/HS/DS buffers
            ID3D11Buffer* ParticlePerFrameConstants = mParticlePerFrameConstantsCPUT->GetNativeBuffer();
            ID3D11Buffer* PerFrameConstants = mPerFrameConstantsCPUT->GetNativeBuffer();
            HSSetConstantBuffers(d3dDeviceContext,
                mParticleShadingHSReflector,
                "PerFrameConstants",
                1, 
                &PerFrameConstants);

            HSSetConstantBuffers(d3dDeviceContext,
                mParticleShadingHSReflector,
                "ParticlePerFrameConstants",
                1, 
                &ParticlePerFrameConstants);
            HSSetConstantBuffers(d3dDeviceContext,
                mParticleShadingHSReflector,
                "ParticlePerPassConstants",
                1, 
                &mParticlePerPassConstants);

            // set all domain shader constant buffers
            DSSetConstantBuffers(d3dDeviceContext,
                mParticleShadingDSReflector[mShaderIdx],
                "ParticlePerFrameConstants",
                1, 
                &ParticlePerFrameConstants);
            DSSetConstantBuffers(d3dDeviceContext,
                mParticleShadingDSReflector[mShaderIdx],
                "ParticlePerPassConstants",
                1, 
                &mParticlePerPassConstants);
           
            DSSetConstantBuffers(d3dDeviceContext,
                mParticleShadingDSReflector[mShaderIdx],
                "PerFrameConstants",
                1, 
                &PerFrameConstants);

            ID3D11Buffer* pmAVSMConstants = mAVSMConstantsCPUT->GetNativeBuffer();
            DSSetConstantBuffers(d3dDeviceContext,
                mParticleShadingDSReflector[mShaderIdx],
                "AVSMConstants",
                1, 
                &pmAVSMConstants);

            DSSetShaderResources(d3dDeviceContext,
                mParticleShadingDSReflector[mShaderIdx],
                "NONCPUT_gAVSMTexture",
                1, 
                &asvmTextureSRV);

            DSSetShaderResources(d3dDeviceContext,
                mParticleShadingDSReflector[mShaderIdx],
                "NONCPUT_gAVSMGenClearMaskSRV",
                1, 
                &avsmGenCtrlSurface);

            DSSetShaderResources(d3dDeviceContext,
                mParticleShadingDSReflector[mShaderIdx],
                "NONCPUT_gAVSMGenDataSRV",
                1, 
                &mAVSMGenDataSRV[mShaderIdx]);

            DSSetSamplers(d3dDeviceContext,
                mParticleShadingDSReflector[mShaderIdx],
                "gAVSMSampler",
                1, 
                &mAVSMSampler);
        }
    }
    else
    {
        ID3D11Buffer* ParticlePerFrameConstants = mParticlePerFrameConstantsCPUT->GetNativeBuffer();
        VSSetConstantBuffers(d3dDeviceContext,
            mParticleShadingVSReflector,
            "ParticlePerPassConstants",
            1, 
            &mParticlePerPassConstants);

        VSSetConstantBuffers(d3dDeviceContext,
            mParticleShadingVSReflector,
            "ParticlePerFrameConstants",
            1, 
            &ParticlePerFrameConstants);
    }

    if( (ui->vertexShaderShadowLookup && !forCPUTGeometry) || (VOL_SHADOW_NO_SHADOW == ui->volumeShadowMethod))
    {
        // Buffers are auto-bound in this case for mParticleShadingPS:
        // mPerFrameConstants - PS
        // mParticlePerFrameConstants - PS
    }
    else
    {
        // Buffers are auto-bound in this case for mParticleShadingPerPixelPS[mShaderIdx]:
        // mPerFrameConstants - PS
        // mParticlePerFrameConstants - PS
        // mAVSMConstants - PS

        PSSetShaderResources(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "NONCPUT_gAVSMTexture",
            1, 
            &asvmTextureSRV);

        PSSetShaderResources(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "NONCPUT_gAVSMGenClearMaskSRV",
            1, 
            &avsmGenCtrlSurface);

        PSSetShaderResources(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "NONCPUT_gAVSMGenDataSRV",
            1, 
            &mAVSMGenDataSRV[mShaderIdx]);

        if( !forCPUTGeometry ) // no need to do this for CPUT geometry, we're setting samplers in CPUT code
        {
            PSSetSamplers(d3dDeviceContext,
                mParticleShadingPerPixelPSReflector[mShaderIdx],
                "gAVSMSampler",
                1, 
                &mAVSMSampler);

            if( ui->tessellate )
            {
                PSSetSamplers(d3dDeviceContext,
                    mParticleShadingPerPixelPSReflector[mShaderIdx],
                    "gAVSMGenCtrlSurfaceSampler",
                    1, 
                    &mAVSMGenSampler);
            }
        }
    }
}

// Render the shadowed particles
// This function, based on the settings of the technique, sets up the shaders
// and buffers and draws the shadowed particles
//-----------------------------------------------------------------------------
void AVSMTechnique::RenderShadedParticles(ID3D11DeviceContext* d3dDeviceContext,
                         ParticleSystem *particleSystem,
                         ID3D11RenderTargetView* backBuffer,
                         ID3D11DepthStencilView* depthBuffer,
                         D3D11_VIEWPORT* viewport,
                         const UIConstants* ui)
{
    ID3D11ShaderResourceView* asvmTextureSRV = mAVSMTextures->GetShaderResource();
    ID3D11ShaderResourceView* avsmGenCtrlSurface = mAVSMGenCtrlSurface->GetShaderResource();

	if(ui->vertexShaderShadowLookup)
	{
		if(ui->tessellate)
		{
            // Gen+DX version with tesselation
			d3dDeviceContext->VSSetShader(mParticleShadingTessellationVS, 0, 0);
			d3dDeviceContext->HSSetShader( mParticleShadingHS, NULL, 0);
			d3dDeviceContext->DSSetShader( mParticleShadingDS[mShaderIdx], NULL, 0);
                    
            VSSetConstantBuffers(d3dDeviceContext,
            mParticleShadingVSReflector,
            "ParticlePerFrameConstants",
            1, 
            &mParticlePerFrameConstants);
		}
		else
		{
            // this version is NEVER hit by any condition
			d3dDeviceContext->VSSetShader(mParticleShadingFinalVS[mShaderIdx], 0, 0);
			d3dDeviceContext->HSSetShader( NULL, NULL, 0);
			d3dDeviceContext->DSSetShader( NULL, NULL, 0);
		}
	}
	else
	{
        // GEN version with no tesselation
        d3dDeviceContext->VSSetShader(mParticleShadingVS, 0, 0);
        d3dDeviceContext->HSSetShader( NULL, NULL, 0);
        d3dDeviceContext->DSSetShader( NULL, NULL, 0);

        VSSetConstantBuffers(d3dDeviceContext,
            mParticleShadingVSReflector,
            "ParticlePerFrameConstants",
            1, 
            &mParticlePerFrameConstants);
    }

	if(ui->wireframe)
		d3dDeviceContext->RSSetState(mParticleWireFrameState);
	else
		d3dDeviceContext->RSSetState(mParticleRasterizerState);

    d3dDeviceContext->RSSetViewports(1, viewport);

	if(ui->vertexShaderShadowLookup || (VOL_SHADOW_NO_SHADOW == ui->volumeShadowMethod))
	{
        // Gen+DX version with tesselation
        // This path handles the cases when we are using particle tesselation
        // manually update the particle constant buffers we need since we're doing the drawing ourselves
        // and not using the framework's auto-bind features
        ID3D11Buffer* PerFrameConstants = mPerFrameConstantsCPUT->GetNativeBuffer();
        d3dDeviceContext->PSSetShader(mParticleShadingPS, 0, 0);

        PSSetConstantBuffers(d3dDeviceContext,
            mParticleShadingPSReflector,
            "PerFrameConstants",
            1, 
            &PerFrameConstants);
        ID3D11Buffer* ParticlePerFrameConstants = mParticlePerFrameConstantsCPUT->GetNativeBuffer();
        PSSetConstantBuffers(d3dDeviceContext,
            mParticleShadingPSReflector, 
            "ParticlePerFrameConstants",
            1, 
            &ParticlePerFrameConstants);
	}
    else
    {
        // GEN version with no tesselation

        // This path handles the cases when we are not using the particle tesselation features
        // manually update the particle constant buffers we need since we're doing the drawing ourselves
        // and not using the framework's auto-bind features
        ID3D11Buffer* PerFrameConstants = mPerFrameConstantsCPUT->GetNativeBuffer();
        PSSetConstantBuffers(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "PerFrameConstants",
            1, 
            &PerFrameConstants);
        
        ID3D11Buffer* ParticlePerFrameConstants = mParticlePerFrameConstantsCPUT->GetNativeBuffer();
        PSSetConstantBuffers(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "ParticlePerFrameConstants",
            1, 
            &ParticlePerFrameConstants);

        d3dDeviceContext->PSSetShader(mParticleShadingPerPixelPS[mShaderIdx], 0, 0);

        PSSetShaderResources(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "NONCPUT_gAVSMTexture",
            1, 
            &asvmTextureSRV);

        PSSetShaderResources(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "NONCPUT_gAVSMGenClearMaskSRV",
            1, 
            &avsmGenCtrlSurface);

        PSSetShaderResources(d3dDeviceContext,
            mParticleShadingPerPixelPSReflector[mShaderIdx],
            "NONCPUT_gAVSMGenDataSRV",
            1, 
            &mAVSMGenDataSRV[mShaderIdx]);
	}
    
    
    // Additively blend into back buffer
    d3dDeviceContext->OMSetRenderTargets(1, &backBuffer, NULL); //depthBuffer);
    d3dDeviceContext->OMSetBlendState(mParticleBlendState, 0, 0xFFFFFFFF);
    d3dDeviceContext->OMSetDepthStencilState(mParticleDepthStencilState, 0x0);

    particleSystem->Draw( d3dDeviceContext, NULL, 0, particleSystem->GetParticleCount(), ui->tessellate != 0 );

}

// Update the contents of the PerFrameConstantsCPUT cbuffer
//-----------------------------------------------------------------------------
void AVSMTechnique::FillInFrameConstants(ID3D11DeviceContext* d3dDeviceContext,
                               const FrameMatrices &m,
                               D3DXVECTOR4 cameraPos, 
                               CPUTCamera* viewerCamera,
                               CPUTCamera* lightCamera,
							   float ScreenWidth, float ScreenHeight,
                               const UIConstants* ui)
{
    // Compute light direction in view space
    D3DXVECTOR3 lightPosWorld = DXUTFromCPUT( lightCamera->GetPosition() );
    D3DXVECTOR3 lightTargetWorld = DXUTFromCPUT( lightCamera->GetPosition() + lightCamera->GetLook() );
    D3DXVECTOR3 lightPosView;
    D3DXVec3TransformCoord(&lightPosView, &lightPosWorld, &m.cameraView);
    D3DXVECTOR3 lightTargetView;
    D3DXVec3TransformCoord(&lightTargetView, &lightTargetWorld, &m.cameraView);
    D3DXVECTOR3 lightDirView = lightTargetView - lightPosView;
    D3DXVec3Normalize(&lightDirView, &lightDirView);
  
    D3D11_MAPPED_SUBRESOURCE mappedResource;
    d3dDeviceContext->Map(mPerFrameConstantsCPUT->GetNativeBuffer(), 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
    PerFrameConstants* constants = static_cast<PerFrameConstants *>(mappedResource.pData);

    // No world matrix for now...
    constants->mCameraWorldViewProj = m.cameraWorldViewProj;
    constants->mCameraWorldView = m.worldMatrix * m.cameraView;
    constants->mCameraViewProj = m.cameraViewProj;
    constants->mCameraProj = m.cameraProj;
    constants->mCameraPos = cameraPos;
    constants->mLightWorldViewProj = m.lightWorldViewProj;
    constants->mAvsmLightWorldViewProj = m.avmsLightWorldViewProj;
    constants->mCameraViewToLightProj = m.cameraViewToLightProj;
    constants->mCameraViewToLightView = m.cameraViewToLightView;
    constants->mCameraViewToAvsmLightProj = m.cameraViewToAvsmLightProj;
    constants->mCameraViewToAvsmLightView = m.cameraViewToAvsmLightView;
    constants->mLightDir = D3DXVECTOR4(lightDirView, 0.0f);
	constants->mScreenResolution.x = ScreenWidth;
	constants->mScreenResolution.y = ScreenHeight;

    float clipNear = viewerCamera->GetNearPlaneDistance();
    float clipFar  = viewerCamera->GetFarPlaneDistance();
    float depthHackMul = ( clipFar * clipNear) / ( clipFar - clipNear );
    float depthHackAdd = clipFar / ( clipFar - clipNear );

    constants->mScreenToViewConsts.x = depthHackMul;
    constants->mScreenToViewConsts.y = depthHackAdd;
    constants->mScreenToViewConsts.z = 0.0f;
    constants->mScreenToViewConsts.w = 0.0f;

    constants->mUI = *ui;

    d3dDeviceContext->Unmap(mPerFrameConstantsCPUT->GetNativeBuffer(), 0);
}


// Update the contents of the ParticlePerPassConstants cbuffer
//-----------------------------------------------------------------------------
void AVSMTechnique::FillParticleRendererConstants(ID3D11DeviceContext* d3dDeviceContext,
                                        CPUTCamera* camera,
                                        const D3DXMATRIXA16 &cameraView,
                                        const D3DXMATRIXA16 &cameraViewProj)
{
    // Particle renderer constants
    D3D11_MAPPED_SUBRESOURCE mappedResource;
    d3dDeviceContext->Map(mParticlePerPassConstants, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
    ParticlePerPassConstants* constants = static_cast<ParticlePerPassConstants *>(mappedResource.pData);

    D3DXMATRIXA16 lightProj     = DXUTFromCPUT( *camera->GetProjectionMatrix() );
    D3DXMATRIXA16 lightView     = DXUTFromCPUT( *camera->GetViewMatrix() );
    D3DXMATRIXA16 lightWorld    = DXUTFromCPUT( *camera->GetWorldMatrix() );

    constants->mParticleWorldViewProj = cameraViewProj;
    constants->mParticleWorldView     = cameraView;
    constants->mEyeRight              = *(( D3DXVECTOR3* )&lightWorld._11); // *camera->GetWorldRight();
    constants->mEyeUp                 = *(( D3DXVECTOR3* )&lightWorld._21); // *camera->GetWorldUp();
    
    d3dDeviceContext->Unmap(mParticlePerPassConstants, 0);
}


// Update the contents of the ListTextureConstants cbuffer
//-----------------------------------------------------------------------------
void AVSMTechnique::FillListTextureConstants(ID3D11DeviceContext* d3dDeviceContext)
{
    // List texture related constants  
    D3D11_MAPPED_SUBRESOURCE mappedResource;
    d3dDeviceContext->Map(mListTextureConstants, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
    LT_Constants* constants = static_cast<LT_Constants*>(mappedResource.pData);

    constants->mMaxNodes = mLisTexNodeCount;
    constants->mFirstNodeMapSize = (float)mAVSMShadowTextureDim;
    
    d3dDeviceContext->Unmap(mListTextureConstants, 0);
}


// Update the contents of the AVSMConstants cbuffer
//-----------------------------------------------------------------------------
void AVSMTechnique::FillAVSMConstants(ID3D11DeviceContext* d3dDeviceContext)
{
    // AVSM related constants   
    D3D11_MAPPED_SUBRESOURCE mappedResource;
    d3dDeviceContext->Map(mAVSMConstantsCPUT->GetNativeBuffer(), 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
    AVSMConstants* constants = static_cast<AVSMConstants*>(mappedResource.pData);

    constants->mMask0 = D3DXVECTOR4( 0.0f,  1.0f,  2.0f,  3.0f);
    constants->mMask1 = D3DXVECTOR4( 4.0f,  5.0f,  6.0f,  7.0f);
    constants->mMask2 = D3DXVECTOR4( 8.0f,  9.0f, 10.0f, 11.0f);
    constants->mMask3 = D3DXVECTOR4(12.0f, 13.0f, 14.0f, 15.0f);
    constants->mMask4 = D3DXVECTOR4(16.0f, 17.0f, 18.0f, 19.0f);
    constants->mEmptyNode = EMPTY_NODE;
    constants->mOpaqueNodeTrans = 1E-4f;
    constants->mShadowMapSize = (float)mAVSMShadowTextureDim;

    d3dDeviceContext->Unmap(mAVSMConstantsCPUT->GetNativeBuffer(), 0);
}

// Debug viewport
//-----------------------------------------------------------------------------
void AVSMTechnique::UpdateDebugViewD3D11( FrameContext & frameContext )
{
    ID3D11ShaderResourceView* avsmGenCtrlSurface = mAVSMGenCtrlSurface->GetShaderResource();

    PSSetShaderResources( frameContext.D3dDeviceContext,
        mParticleShadingPerPixelPSReflector[mShaderIdx],
        "NONCPUT_gAVSMGenClearMaskSRV",
        1, 
        &avsmGenCtrlSurface);
}
