//--------------------------------------------------------------------------------------
// Copyright 2011-2013 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//--------------------------------------------------------------------------------------
#ifndef __AVSMTECHNIQUE_H__
#define __AVSMTECHNIQUE_H__

//--------------------------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------------------------
#include "..\media\Shader\AVSM_def.h"
#include "AppShaderConstants.h"
#include "Texture2D.h"
#include "ParticleSystem.h"
#include "Stats.h"

#include <fstream>
#include <iostream>
#include <memory>
#include <string>
#include <vector>
#include <limits> // for std::numeric_limits<float>::max()/min()

// DirectX includes
#include <d3dx9math.h>
#include <stdio.h>
#include "CPUT_DX11.h"
#include <D3D11.h>
#include <xnamath.h>

#define NOMINMAX
#include <Windows.h>
#undef NOMINMAX
#include <d3d9.h>
#include <d3dx9.h>
#include <d3dx9math.h>
#include <D3D11Shader.h>

//--------------------------------------------------------------------------------------
// Handy utility defines
//--------------------------------------------------------------------------------------
#if defined(DEBUG) || defined(_DEBUG)
#ifndef V
#define V(x)           { hr = (x); if( FAILED(hr) ) { ::MessageBox(NULL, L"HRESULT error", L"Error in AVSMTechnique", MB_OK); } }
#endif
#ifndef V_RETURN
#define V_RETURN(x)    { hr = (x); if( FAILED(hr) ) {  ::MessageBox(NULL, L"HRESULT error", L"Error in AVSMTechnique", MB_OK); return x; } }
#endif
#else
#ifndef V
#define V(x)           { hr = (x); }
#endif
#ifndef V_RETURN
#define V_RETURN(x)    { hr = (x); if( FAILED(hr) ) { return hr; } }
#endif
#endif

#ifndef SAFE_DELETE
#define SAFE_DELETE(p)       { if (p) { delete (p);     (p)=NULL; } }
#endif
#ifndef SAFE_DELETE_ARRAY
#define SAFE_DELETE_ARRAY(p) { if (p) { delete[] (p);   (p)=NULL; } }
#endif
#ifndef SAFE_RELEASE
#define SAFE_RELEASE(p)      { if (p) { (p)->Release(); (p)=NULL; } }
#endif


//--------------------------------------------------------------------------------------
// Defines
//--------------------------------------------------------------------------------------

#define MAX_AVSM_RT_COUNT   (4)
#define MAX_SHADER_VARIATIONS (MAX_AVSM_RT_COUNT / 2)


//--------------------------------------------------------------------------------------
// Enums
//--------------------------------------------------------------------------------------
enum SCENE_SELECTION {
    POWER_PLANT_SCENE        = 0,
    GROUND_PLANE_SCENE       = 1,
};

//--------------------------------------------------------------------------------------
// Classes
//--------------------------------------------------------------------------------------
class AVSMTechnique
{
public:
    struct Options
    {
        Options() : 
        scene(GROUND_PLANE_SCENE),
        enableParticles(true),
        enableAutoBoundsAVSM(true),
        enableShadowPicking(false),
        NodeCount(8),
        enableTransmittanceCurve(false),
        enableVolumeShadowCreation(true),
        pickedX(0),
        pickedY(0)
        {}
        SCENE_SELECTION scene;
        bool enableParticles;
        bool enableAutoBoundsAVSM;
        bool enableShadowPicking;
        unsigned int NodeCount;
        bool enableTransmittanceCurve;
        bool enableVolumeShadowCreation;
        int pickedX;
        int pickedY;
    };

    struct FrameMatrices
    {    
        D3DXMATRIXA16 worldMatrix;

        D3DXMATRIXA16 cameraProj;
        D3DXMATRIXA16 cameraView;
        D3DXMATRIXA16 cameraViewInv;

        D3DXMATRIXA16 cameraWorldViewProj;
        D3DXMATRIXA16 cameraWorldView;

        D3DXMATRIXA16 cameraViewProj;
        D3DXMATRIXA16 cameraViewToLightProj;
        D3DXMATRIXA16 cameraViewToLightView;
        D3DXMATRIXA16 cameraViewToAvsmLightProj;
        D3DXMATRIXA16 cameraViewToAvsmLightView;

        D3DXMATRIXA16 lightProj;
        D3DXMATRIXA16 lightView;
        D3DXMATRIXA16 lightViewProj;
        D3DXMATRIXA16 lightWorldViewProj;

        D3DXMATRIXA16 avsmLightProj;
        D3DXMATRIXA16 avsmLightView;
        D3DXMATRIXA16 avsmLightViewProj;
        D3DXMATRIXA16 avmsLightWorldViewProj;
    };

    struct FrameContext
    {
        FrameMatrices               Matrices;
        ID3D11DeviceContext*        D3dDeviceContext;
        Options                     Options;
        ParticleSystem *            ParticleSystem;
        CPUTCamera *                ViewerCamera;
        CPUTCamera *                LightCamera;
        D3D11_VIEWPORT              Viewport;
        UIConstants                 GPUUIConstants;
		float					    mScreenWidth;
		float					    mScreenHeight;
        FrameStats                  Stats;
        ID3D11ShaderResourceView *  DepthBufferSRV;
    };

public:

    AVSMTechnique(ID3D11Device* d3dDevice,
        ID3D11DeviceContext* d3dDeviceContext,
        unsigned int nodeCount,
        unsigned int shadowTextureDim,     
        unsigned int avsmShadowTextureDim);

    ~AVSMTechnique();
    
    void OnD3D11ResizedSwapChain(ID3D11Device* d3dDevice,
                                 const D3D11_TEXTURE2D_DESC* backBufferDesc);

    void InitializeFrameContext( FrameContext & outContext, const Options & options, ID3D11DeviceContext* d3dDeviceContext, ParticleSystem * particleSystem, 
                                    const D3DXMATRIX & worldMatrix, CPUTCamera* viewerCamera, CPUTCamera* lightCamera, const D3D11_VIEWPORT & viewport );

    void UpdateParticles( FrameContext & frameContext );

    void CreateAVSMShadowMap( FrameContext & frameContext, CPUTAssetSet *mpAssetSet, CPUTRenderParametersDX* pRenderParams );

    void RenderShadedParticles( FrameContext & frameContext, ID3D11RenderTargetView* backBuffer, ID3D11DepthStencilView* depthBuffer );

    void UpdateDebugViewD3D11( FrameContext & frameContext );

    void UpdateStatesAndConstantsForAVSMUse( FrameContext & frameContext, bool forCPUTGeometry );

    void Cleanup(ID3D11DeviceContext* d3dDeviceContext);

    void DumpTransmittanceCurve() {
        mDumpTransmittanceCurve = true;
        ++mDumpTransmittanceCurveIndex;
    }

    void SetNodeCount(int nodeCount);

    // initalization of AVSM technique
    void WrapBuffers();

    bool LoadAVSMShaders(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount,
         unsigned int shadowTextureDim,
         unsigned int avsmShadowTextureDim);

    bool CreateAVSMRenderStates(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount,
         unsigned int shadowTextureDim,
         unsigned int avsmShadowTextureDim);

    bool CreateAVSMBuffers(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount,
         unsigned int avsmShadowTextureDim);

	bool DeleteAVSMBuffers(ID3D11Device *d3dDevice,
         ID3D11DeviceContext* d3dDeviceContext,
         unsigned int nodeCount);

    unsigned int mShadowTextureDim;
    unsigned int mAVSMShadowTextureDim;

private:
    struct PointVertex {
        float x, y, z;
    };
 
    void CaptureFragments(ID3D11DeviceContext* d3dDeviceContext,
                          ParticleSystem *particleSystem,
                          const UIConstants* ui,
                          bool initCounter);

    void GenerateVisibilityCurve(ID3D11DeviceContext* d3dDeviceContext,
                                 const UIConstants* ui);


    void RenderShadedParticles(ID3D11DeviceContext* d3dDeviceContext,
                        ParticleSystem *particleSystem,
                        ID3D11RenderTargetView* backBuffer,
                        ID3D11DepthStencilView* depthBuffer,
                        D3D11_VIEWPORT* viewport,
                        const UIConstants* ui);

    void FillInFrameConstants(ID3D11DeviceContext* d3dDeviceContext,
                              const FrameMatrices &m,
                              D3DXVECTOR4 cameraPos,
                              CPUTCamera* viewerCamera,
                              CPUTCamera* lightCamera,
							  float ScreenWidth, float ScreenHeight,
                              const UIConstants* ui);

    void FillParticleRendererConstants(ID3D11DeviceContext* d3dDeviceContext,
                                       CPUTCamera* lightCamera,
                                       const D3DXMATRIXA16 &cameraView,
                                       const D3DXMATRIXA16 &cameraViewProj);
    void FillListTextureConstants(ID3D11DeviceContext* d3dDeviceContext);
    void FillAVSMConstants(ID3D11DeviceContext* d3dDeviceContext);

    void ClearShadowBuffers(ID3D11DeviceContext* d3dDeviceContext, const UIConstants* ui);

    void CreatePixelShadersFromCompiledObjs(ID3D11Device* d3dDevice,
                                           int shaderCount,
                                           const char* prefix, 
                                           const char* suffix, 
                                           ID3D11PixelShader** shaderArray,
                                           ID3D11ShaderReflection** shaderReflArray);

	void CreateDomainShadersFromCompiledObjs(ID3D11Device* d3dDevice,
                                           int shaderCount,
                                           const char* prefix, 
                                           const char* suffix, 
                                           ID3D11DomainShader** shaderArray,
                                           ID3D11ShaderReflection** shaderReflArray);
	void CreateVertexShadersFromCompiledObjs(ID3D11Device* d3dDevice,
                                           int shaderCount,
                                           const char* prefix, 
                                           const char* suffix, 
                                           ID3D11VertexShader** shaderArray,
                                           ID3D11ShaderReflection** shaderReflArray);	


    unsigned int mLisTexNodeCount;


    ID3D11VertexShader* mFullScreenTriangleVS;
    ID3D11ShaderReflection* mFullScreenTriangleVSReflector;

    // Particle and AVSM shaders
    ID3D11VertexShader*  mParticleShadingVS;
    ID3D11ShaderReflection* mParticleShadingVSReflector;
    ID3D11PixelShader*   mParticleShadingPS;
    ID3D11ShaderReflection* mParticleShadingPSReflector;

    ID3D11PixelShader*   mParticleShadingPerPixelPS[MAX_SHADER_VARIATIONS];
    ID3D11ShaderReflection* mParticleShadingPerPixelPSReflector[MAX_SHADER_VARIATIONS];
    ID3D11PixelShader*   mParticleAVSMCapturePS; // Capture-all-fragments shaders
    ID3D11ShaderReflection* mParticleAVSMCapturePSReflector;
    ID3D11PixelShader*   mAVSMInsertionSortResolvePS[MAX_SHADER_VARIATIONS];
    ID3D11ShaderReflection* mAVSMInsertionSortResolvePSReflector[MAX_SHADER_VARIATIONS];
    ID3D11PixelShader*   mAVSMSinglePassInsertPS[MAX_SHADER_VARIATIONS]; // Fake render-target-read shaders
    ID3D11ShaderReflection* mAVSMSinglePassInsertPSReflector[MAX_SHADER_VARIATIONS];
    ID3D11PixelShader*   mAVSMClearStructuredBufPS;
    ID3D11ShaderReflection* mAVSMClearStructuredBufPSReflector;
    ID3D11PixelShader*   mAVSMConvertSUAVtoTex2DPS;
    ID3D11ShaderReflection* mAVSMConvertSUAVtoTex2DPSReflector;

	// New shaders for vertex texture lookup and Tessellation
    ID3D11DomainShader*  mParticleShadingDS[MAX_SHADER_VARIATIONS];
    ID3D11ShaderReflection* mParticleShadingDSReflector[MAX_SHADER_VARIATIONS];
    ID3D11HullShader*  mParticleShadingHS;
    ID3D11ShaderReflection* mParticleShadingHSReflector;
    ID3D11VertexShader*  mParticleShadingFinalVS[MAX_SHADER_VARIATIONS];
    ID3D11ShaderReflection* mParticleShadingFinalVSReflector[MAX_SHADER_VARIATIONS];
    ID3D11VertexShader*  mParticleShadingTessellationVS;
    ID3D11ShaderReflection* mParticleShadingTessellationVSReflector;

    // AVSM Gen shaders
    ID3D11PixelShader*   mParticleAVSM_GenPS[MAX_SHADER_VARIATIONS]; // "in-place" generation of AVSM visibility function
    ID3D11ShaderReflection* mParticleAVSM_GenPSReflector[MAX_SHADER_VARIATIONS];

    // Constant buffers
    ID3D11Buffer* mPerFrameConstants;    
    ID3D11Buffer* mParticlePerFrameConstants;
    ID3D11Buffer* mParticlePerPassConstants;
    ID3D11Buffer* mListTextureConstants;
    ID3D11Buffer* mAVSMConstants;
    // cput equivalents for autobinding
    CPUTBufferDX11* mAVSMConstantBufferCPUT;
    CPUTBufferDX11* mAVSMConstantsCPUT;
    CPUTBufferDX11* mPerFrameConstantsCPUT;
    CPUTBufferDX11* mParticlePerFrameConstantsCPUT;

    // Rasterizer state
    ID3D11RasterizerState* mRasterizerState;
    ID3D11RasterizerState* mDoubleSidedRasterizerState;
	ID3D11RasterizerState* mParticleWireFrameState;
    ID3D11RasterizerState* mShadowRasterizerState;
    ID3D11RasterizerState* mParticleRasterizerState;

    ID3D11DepthStencilState* mDefaultDepthStencilState;
    ID3D11DepthStencilState* mParticleDepthStencilState;
    ID3D11DepthStencilState* mAVSMCaptureDepthStencilState;

    ID3D11BlendState* mGeometryBlendState;
    ID3D11BlendState* mLightingBlendState;
    ID3D11BlendState* mParticleBlendState;

    ID3D11SamplerState* mAVSMSampler;

    std::vector< std::tr1::shared_ptr<Texture2D> > mGBuffer;

    // AVSM 
    D3D11_VIEWPORT                  mAVSMShadowViewport;
    int                             mAVSMNodeCount;
    int                             mShaderIdx;
    std::tr1::shared_ptr<Texture2D> mAVSMTextures;
    CPUTTextureDX11*                mAVSMTexturesCPUT;
    ID3D11Texture2D*                mAVSMTexturesDebug;
    ID3D11Buffer*                   mAVSMStructBuf;
    ID3D11UnorderedAccessView*      mAVSMStructBufUAV;
    ID3D11ShaderResourceView*       mAVSMStructBufSRV;

    // AVSM Gen 
    std::tr1::shared_ptr<Texture2D> mAVSMGenCtrlSurface;
    CPUTTextureDX11*                mAVSMGenCtrlSurfaceCPUT;
    ID3D11Buffer*                   mAVSMGenData[MAX_SHADER_VARIATIONS];
    ID3D11UnorderedAccessView*      mAVSMGenDataUAV[MAX_SHADER_VARIATIONS];
    ID3D11ShaderResourceView*       mAVSMGenDataSRV[MAX_SHADER_VARIATIONS];
    CPUTBufferDX11*                 mAVSMGenDataSRVCPUT[MAX_SHADER_VARIATIONS];
    ID3D11SamplerState*				mAVSMGenSampler;

    // List texture 
    std::tr1::shared_ptr<Texture2D> mListTexFirstSegmentNodeOffset;
    std::tr1::shared_ptr<Texture2D> mListTexFirstVisibilityNodeOffset;
    ID3D11Buffer*                   mListTexSegmentNodes;
    ID3D11Buffer*                   mListTexSegmentNodesDebug;
    ID3D11Buffer*                   mListTexVisibilityNodes;
    ID3D11Buffer*                   mListTexVisibilityNodesDebug;
    ID3D11UnorderedAccessView*      mListTexSegmentNodesUAV;
    ID3D11ShaderResourceView*       mListTexSegmentNodesSRV;
    ID3D11UnorderedAccessView*      mListTexVisibilityNodesUAV;
    ID3D11ShaderResourceView*       mListTexVisibilityNodesSRV;
    ID3D11Texture2D*                mListTexFirstOffsetDebug;

    ID3D11VertexShader*             mDrawTransmittanceVS;
    ID3D11ShaderReflection*         mDrawTransmittanceVSReflector;
    ID3D11PixelShader*              mDrawTransmittancePS;
    ID3D11ShaderReflection*         mDrawTransmittancePSReflector;
    ID3D11Buffer*                   mDrawTransmittanceVB;
    ID3D11InputLayout*              mDrawTransmittanceLayout;
    bool                            mDumpTransmittanceCurve;
    int                             mDumpTransmittanceCurveIndex;
    int                             mDrawTransmittanceMaxNodes;

    float                           mLastTime;
};

// Some conversion helpers
inline D3DXMATRIX DXUTFromCPUT( const float4x4 & mat )
{
    D3DXMATRIX ret( mat.r0.f );
    return ret;
}
inline D3DXVECTOR3 DXUTFromCPUT( const float3 & vec )
{
    return D3DXVECTOR3( vec.x, vec.y, vec.z );
}

#endif // #define __AVSMTECHNIQUE_H__
