// Copyright 2012 Intel Corporation
// All Rights Reserved
//



#ifndef COMMON_PARTICLE_FX_H
#define COMMON_PARTICLE_FX_H

#include "IntelExtensions.hlsl"
#include "GBuffer.hlsl"
#include "AVSM_Gen.hlsl"
#include "AVSM_Resolve.hlsl"
#include "Rendering.hlsl"
#include "ConstantBuffers.hlsl"
#include "ScreenSpace.hlsl"

//--------------------------------------------------------------------------------------

//////////////////////////////////////////////
// Resource Views
//////////////////////////////////////////////
RWStructuredBuffer<AVSMData>  gAVSMStructBufUAV;
StructuredBuffer<AVSMData>    gAVSMStructBufSRV                     : register(t22);
Texture2D                     gParticleOpacityNoiseTex              : register(t23);
Texture2D                     gDepthBuffer                          : register(t30);



////////////////////////////////////////////////////////////////////////////////////////////
// getting the viewspace depth for smooth particle - solid geometry intersection
float ScreenToViewDepth( float screenDepth )
{
   float depthHackMul = mScreenToViewConsts.x;
   float depthHackAdd = mScreenToViewConsts.y;

   // Optimised version of "-cameraClipNear / (cameraClipFar - projDepth * (cameraClipFar - cameraClipNear)) * cameraClipFar"

   // Set your depthHackMul and depthHackAdd to:
   // depthHackMul = ( cameraClipFar * cameraClipNear) / ( cameraClipFar - cameraClipNear );
   // depthHackAdd = cameraClipFar / ( cameraClipFar - cameraClipNear );

 	return depthHackMul / (depthHackAdd - screenDepth);
}
//
float LoadScreenDepthViewspace( int2 pos )
{
   return ScreenToViewDepth( gDepthBuffer.Load( int3( pos.xy, 0 ) ).x );
}
////////////////////////////////////////////////////////////////////////////////////////////




bool RaySphereIntersection(out float2 interCoeff, in float3 sphereCenter, in float sphereRadius, in float3 rayOrigin, in float3 rayNormDir, const bool snapToSphere)
{
	float3 dst = rayOrigin - sphereCenter;
	float b = dot(dst, rayNormDir);
	float c = dot(dst, dst) - (sphereRadius * sphereRadius);
	float d = b * b - c;

   if( snapToSphere )
   {
	   d = max( d, 0.00001 );
   }

	interCoeff = -b.xx + sqrt(d) * float2(-1,1);
	return d > 0;
}

// compute particles thickness assuming their billboards
// in screen space are maximal sections of a sphere.
float ParticleThickness(float2 ray)
{
	return saturate(1 - sqrt(2.0f) * length(ray));
}

// We model particles as small spheres and we intersect them with light rays
// Entry and exit points are used to defined an AVSM segment
bool IntersectDynamicParticle(in  DynamicParticlePSIn input,
                              out float3 entry,
                              out float3 exit,
                              out float  transmittance,
                              const bool snapToSphere)
{
    bool   res;
    float2 linearCoeff;    
    float3 normRayDir     = normalize(input.ViewPos);
    float3 particleCenter = input.ViewCenter;
    float  particleSize = input.UVS.z;
    float  particleRadius = mScale * particleSize / 2.0f;
                
    [flatten]if (RaySphereIntersection(linearCoeff, 
                              particleCenter, 
                              particleRadius, 
                              float3(0, 0, 0), 
                              normRayDir, snapToSphere)){  
        // compute entry and exit points along the ray direction
	    entry = linearCoeff.xxx * normRayDir;  
	    exit  = linearCoeff.yyy * normRayDir;
	    
	    // compute normalized opacity
	    float segLength = (exit.z - entry.z) / (2.0f * particleRadius); 
	    // Compute density based indirectly on distance from center
	    float densityTimesSegLength = pow(segLength, 4.0f);
	    // per particle and global opacity multipliers
        float opacity = input.Opacity * mParticleOpacity;	
        // compute transmittance
	    transmittance = exp(-opacity * densityTimesSegLength);          	    	    
	    res = true;  
    } 
    else 
    {
        entry = 0;
        exit  = 0;
        transmittance = 1;        
        res = false;
    }     

    return res;
}    

bool IntersectDynamicParticle(in  DynamicParticlePSIn input,
                              out float3 entry,
                              out float3 exit,
                              out float  transmittance)
{
   return IntersectDynamicParticle( input, entry, exit, transmittance, false );
}                    

float GetDynamicParticleAlpha(DynamicParticlePSIn Input)
{    
    float3 entry, exit;
    float  transmittance;
    IntersectDynamicParticle(Input, entry, exit, transmittance);
    
    return saturate(transmittance);
}



//--------------------------------------------------------------------------------------
DynamicParticlePSIn DynamicParticlesShading_VS(
    float4 inPosition	: POSITION,
    float3 inUV			: TEXCOORD0,
    float  inOpacity	: TEXCOORD1
)
{
    DynamicParticlePSIn	Out;
    float size		= inUV.z * mParticleSize;

    // Make screen-facing
    float4 position;
    float2 offset	= inUV.xy - 0.5f.xx;
    position.xyz	= inPosition.xyz + size * (offset.xxx * mEyeRight.xyz + offset.yyy * mEyeUp.xyz);  
    position.w		= 1.0;

    float4 projectedPosition = mul( position, mParticleWorldViewProj ); 
    
    Out.Position    = projectedPosition;
    
    Out.ObjPos      = position.xyz;
    Out.ViewPos 	= mul( position, mParticleWorldView ).xyz;
    Out.ViewCenter	= mul( float4(inPosition.xyz, 1.0f), mParticleWorldView).xyz;
    Out.UVS			= float3(inUV.xy, size);
    Out.Opacity		= inOpacity;
	Out.color		= float4(0,0,0,1);
	Out.ShadowInfo  = float2( 1, 1 );

    return Out;
}

//--------------------------------------------------------------------------------------
VS_OUTPUT_HS_INPUT DynamicParticlesShading_TessellationVS(
    float4 inPosition	: POSITION,
    float3 inUV			: TEXCOORD0,
    float  inOpacity	: TEXCOORD1
)
{
    VS_OUTPUT_HS_INPUT	Out;

   
    float size		= inUV.z * mParticleSize;

    // Make screen-facing
    float4 position;
    float2 offset	= inUV.xy - 0.5f.xx;
    position.xyz	= inPosition.xyz + size * (offset.xxx * mEyeRight.xyz + offset.yyy * mEyeUp.xyz);  
    position.w		= 1.0;
    
	float4 projectedPosition = mul( position, mParticleWorldViewProj ); 
    
    Out.vScreenPos    = projectedPosition;
	Out.vScreenPos.x = -0.5;
	Out.vWorldPos = inPosition;
	Out.texCoord = inUV;
	Out.inOpacity = inOpacity;

    return Out;
}


//--------------------------------------------------------------------------------------
// Hull shader
//--------------------------------------------------------------------------------------
HS_CONSTANT_DATA_OUTPUT ConstantsHS( InputPatch<VS_OUTPUT_HS_INPUT, 3> p, uint PatchID : SV_PrimitiveID )
{
    HS_CONSTANT_DATA_OUTPUT output  = (HS_CONSTANT_DATA_OUTPUT)0;
    float4 vEdgeTessellationFactors = float4( 0, 0, 0,0 );
 
    output.DebugColor = float3( 0, 0, 0 );

    float size		= p[0].texCoord.z * mParticleSize;

    // Make screen-facing
    float3 position;
    float2 offset	= p[0].texCoord.xy - 0.5f.xx;
    position.xyz	= p[0].vWorldPos.xyz + size * (offset.xxx * mEyeRight.xyz + offset.yyy * mEyeUp.xyz);  
    
    // Get the screen space position of each control point
    float2 f2EdgeScreenPosition0 = 
        GetScreenSpacePosition( position.xyz, mParticleWorldViewProj, mScreenResolution.x, mScreenResolution.y );

    size	= p[1].texCoord.z * mParticleSize;
    offset	= p[1].texCoord.xy - 0.5f.xx;
    position.xyz	= p[1].vWorldPos.xyz + size * (offset.xxx * mEyeRight.xyz + offset.yyy * mEyeUp.xyz);  

    float2 f2EdgeScreenPosition1 = 
        GetScreenSpacePosition( position.xyz, mParticleWorldViewProj, mScreenResolution.x, mScreenResolution.y );

    size	= p[2].texCoord.z * mParticleSize;
    offset	= p[2].texCoord.xy - 0.5f.xx;
    position.xyz	= p[2].vWorldPos.xyz + size * (offset.xxx * mEyeRight.xyz + offset.yyy * mEyeUp.xyz);  

    float2 f2EdgeScreenPosition2 = 
        GetScreenSpacePosition( position.xyz, mParticleWorldViewProj, mScreenResolution.x, mScreenResolution.y );
    
    // Calculate edge tessellation factors based on desired screen space tessellation value
    vEdgeTessellationFactors.x = mUI.TessellatioDensity * distance(f2EdgeScreenPosition2, f2EdgeScreenPosition1);
    vEdgeTessellationFactors.y = mUI.TessellatioDensity * distance(f2EdgeScreenPosition2, f2EdgeScreenPosition0);
    vEdgeTessellationFactors.z = mUI.TessellatioDensity * distance(f2EdgeScreenPosition0, f2EdgeScreenPosition1);
    float finalTessFactor = 0.33 * ( vEdgeTessellationFactors.x + vEdgeTessellationFactors.y + vEdgeTessellationFactors.z );
  
    vEdgeTessellationFactors.w =  max( 2.5, finalTessFactor );                      // Always use at least some tessellation
    vEdgeTessellationFactors.w *= saturate( finalTessFactor * 3.0 - 0.5 ) + 0.01;   // But fade away for really small particles

    // Assign tessellation levels
    output.Edges[0] = vEdgeTessellationFactors.w;
    output.Edges[1] = vEdgeTessellationFactors.w;
    output.Edges[2] = vEdgeTessellationFactors.w;
    output.Inside   = vEdgeTessellationFactors.w;

    output.DebugColor = float3( (vEdgeTessellationFactors.w-1.0) / 10.0, 0, 0 );
	
    return output;
}

[domain("tri")]
[partitioning("fractional_odd")]
[outputtopology("triangle_cw")]
[outputcontrolpoints(3)]
[patchconstantfunc("ConstantsHS")]
[maxtessfactor(15.0)]
HS_CONTROL_POINT_OUTPUT DynamicParticlesShading_HS( InputPatch<VS_OUTPUT_HS_INPUT, 3> inputPatch, uint uCPID : SV_OutputControlPointID )
{
    HS_CONTROL_POINT_OUTPUT    output = (HS_CONTROL_POINT_OUTPUT)0;
    
    // Copy inputs to outputs
    output.vWorldPos = inputPatch[uCPID].vWorldPos;
    output.vScreenPos = inputPatch[uCPID].vScreenPos;
    output.texCoord =  inputPatch[uCPID].texCoord;
    output.inOpacity =  inputPatch[uCPID].inOpacity;

    return output;
}

SurfaceData ConstructSurfaceData(float3 PosView, float3 Normal)
{
    SurfaceData Surface;
    Surface.positionView = PosView;
    Surface.positionViewDX = ddx(Surface.positionView);
    Surface.positionViewDY = ddy(Surface.positionView);
    Surface.normal = Normal;
    Surface.albedo = float4(0,0,0,1);
    Surface.lightSpaceZ = mul(float4(Surface.positionView.xyz, 1.0f), mCameraViewToLightProj).z;
    Surface.lightTexCoord = ProjectIntoLightTexCoord(Surface.positionView.xyz);
    Surface.lightTexCoordDX = ddx(Surface.lightTexCoord);
    Surface.lightTexCoordDY = ddy(Surface.lightTexCoord); 

    return Surface;
}

float ShadowContrib(SurfaceData LitSurface, DynamicParticlePSIn Input)
{
    float2 lightTexCoord = ProjectIntoAvsmLightTexCoord(LitSurface.positionView.xyz);
    float receiverDepth = mul(float4(LitSurface.positionView.xyz, 1.0f), mCameraViewToAvsmLightView).z;     
    
    return VolumeSample(mUI.volumeShadowMethod, lightTexCoord, receiverDepth);  
}

//--------------------------------------------------------------------------------------
// Domain Shader
//--------------------------------------------------------------------------------------
[domain("tri")]
DynamicParticlePSIn DynamicParticlesShading_DS( HS_CONSTANT_DATA_OUTPUT input, float3 BarycentricCoordinates : SV_DomainLocation, 
             const OutputPatch<HS_CONTROL_POINT_OUTPUT, 3> TrianglePatch )
{
    DynamicParticlePSIn output = (DynamicParticlePSIn)0;

    // Interpolate world space position with barycentric coordinates
    float4 inPosition = BarycentricCoordinates.x * TrianglePatch[0].vWorldPos + 
                       BarycentricCoordinates.y * TrianglePatch[1].vWorldPos + 
                       BarycentricCoordinates.z * TrianglePatch[2].vWorldPos;

					       
	// Interpolate other inputs with barycentric coordinates
    float3 inUV = BarycentricCoordinates.x * TrianglePatch[0].texCoord + 
                      BarycentricCoordinates.y * TrianglePatch[1].texCoord + 
                      BarycentricCoordinates.z * TrianglePatch[2].texCoord;

	// Interpolate other inputs with barycentric coordinates
    float inOpacity = BarycentricCoordinates.x * TrianglePatch[0].inOpacity + 
                      BarycentricCoordinates.y * TrianglePatch[1].inOpacity + 
                      BarycentricCoordinates.z * TrianglePatch[2].inOpacity;


    float size		= inUV.z * mParticleSize;

    // Make screen-facing
    float4 position;
    float2 offset	= inUV.xy - 0.5f.xx;
    position.xyz	= inPosition.xyz + size * (offset.xxx * mEyeRight.xyz + offset.yyy * mEyeUp.xyz);  
    position.w		= 1.0;

    float4 projectedPosition = mul( position, mParticleWorldViewProj ); 
    
    output.Position    = projectedPosition;
    
    output.ObjPos		= position.xyz;
    output.ViewPos 		= mul( position, mParticleWorldView ).xyz;
    output.ViewCenter	= mul( float4(inPosition.xyz, 1.0f), mParticleWorldView).xyz;
    output.UVS			= float3(inUV.xy, size);
    output.Opacity		= inOpacity;
    output.color        = float4( input.DebugColor.xyz, 1 );

    output.ShadowInfo = float2( 1, 1 );
    if( mUI.vertexShaderShadowLookup ) //#ifdef CALCULATE_AVSM_IN_VS
    {
       DynamicParticlePSIn	vsShadowIn = output;
       //vsShadowIn.UVS.z *= 1.6;

       float3 entry, exit;	
       float  shadowTerm = 1.0f;
       float  segmentTransmittance = 1.0f;
       [flatten]
       if( IntersectDynamicParticle( vsShadowIn, entry, exit, segmentTransmittance, true ) )
       {
          float2 lightTexCoord = ProjectIntoLightTexCoord(entry);	    
     
          SurfaceData LitSurface = ConstructSurfaceData(entry, 0.0f.xxx);	
          if (mUI.enableVolumeShadowLookup)
          {
             shadowTerm = ShadowContrib(LitSurface, vsShadowIn);
          }
          //output.ShadowInfo = LitSurface.positionView.xy;
       }
       output.ShadowInfo = float2( shadowTerm, 0.0 );
    } //#endif

    return output;
}

//--------------------------------------------------------------------------------------
DynamicParticlePSIn DynamicParticlesShadingFinal_VS(
    float4 inPosition	: POSITION,
    float3 inUV			: TEXCOORD0,
    float  inOpacity	: TEXCOORD1
)
{
    DynamicParticlePSIn	Out;
    float size		= inUV.z * mParticleSize;

    // Make screen-facing
    float4 position;
    float2 offset	= inUV.xy - 0.5f.xx;
    position.xyz	= inPosition.xyz + size * (offset.xxx * mEyeRight.xyz + offset.yyy * mEyeUp.xyz);  
    position.w		= 1.0;

    float4 projectedPosition = mul( position, mParticleWorldViewProj ); 
    
    Out.Position    = projectedPosition;
    
    Out.ObjPos      = position.xyz;
    Out.ViewPos 	= mul( position, mParticleWorldView ).xyz;
    Out.ViewCenter	= mul( float4(inPosition.xyz, 1.0f), mParticleWorldView).xyz;
    Out.UVS			= float3(inUV.xy, size);
    Out.Opacity		= inOpacity;
	Out.color        = float4(1,1,1,1);
    Out.ShadowInfo = float2( 1, 1 );

    if( mUI.vertexShaderShadowLookup ) //#ifdef CALCULATE_AVSM_IN_VS
    {
       DynamicParticlePSIn	vsShadowIn = Out;
       vsShadowIn.UVS.z *= 2.0;

       float3 entry, exit;	
       float  shadowTerm = 1.0f;
       float  segmentTransmittance = 1.0f;
       [flatten]
       if( IntersectDynamicParticle( vsShadowIn, entry, exit, segmentTransmittance ) )
       {
          float2 lightTexCoord = ProjectIntoLightTexCoord(entry);	    
     
          SurfaceData LitSurface = ConstructSurfaceData(entry, 0.0f.xxx);	
          if (mUI.enableVolumeShadowLookup)
          {
             shadowTerm = ShadowContrib(LitSurface, vsShadowIn);
          }
          //Out.ShadowInfo = LitSurface.positionView.xy;
       }
       Out.ShadowInfo = float2( shadowTerm, 0.0 );
    } //#endif

    return Out;
}



float4 DynamicParticlesShading_PS(DynamicParticlePSIn Input) : SV_Target
{
    float3 entry, exit;	
	float  shadowTerm;
	float  segmentTransmittance = 1.0f;
    [flatten]
    if (IntersectDynamicParticle(Input, entry, exit, segmentTransmittance)) 
    {
        shadowTerm = Input.ShadowInfo.x;
    }


	float depthDiff = 1.0f;
    float3 ambient = 0.01f;
	float3 diffuse = 0.95f;

	float3 LightContrib = float3(0.9,0.9,1.0) * diffuse;

    // soft blend with solid geometry    
    {
        float depthBufferViewspacePos = LoadScreenDepthViewspace( int2(Input.Position.xy) );
		
		depthDiff = (depthBufferViewspacePos - Input.ViewPos.z) / mSoftParticlesSaturationDepth;
		depthDiff = smoothstep(0.0f, 1.0f, depthDiff);
	}
 
	if(mUI.wireframe)
	{
		return Input.color;
	}
	else
	{
		float3 Color = ambient + LightContrib * shadowTerm;// * gParticleOpacityNoiseTex.Sample(gDiffuseSampler, Input.UVS.xy).xyz;
		return float4(Color, depthDiff * (1.0f - segmentTransmittance));   
	}
}

float4 DynamicParticlesPerPixelShading_PS(DynamicParticlePSIn Input) : SV_Target
{
    float3 entry, exit;	
	float  shadowTerm = 1.0f;
	float  segmentTransmittance = 1.0f;
    [flatten]
    if (IntersectDynamicParticle(Input, entry, exit, segmentTransmittance)) 
    {
	    float2 lightTexCoord = ProjectIntoLightTexCoord(entry);	   
			
        SurfaceData LitSurface = ConstructSurfaceData(entry, 0.0f.xxx);	

        //return float4( LitSurface.positionView.xy, 1, 1 );
        //return float4( ProjectIntoAvsmLightTexCoord( LitSurface.positionView.xyz ).xy, 1, 1 );

        shadowTerm = ShadowContrib(LitSurface, Input);
    }

	float depthDiff = 1.0f;
    float3 ambient = 0.01f;
	float3 diffuse = 0.95f;

	float3 LightContrib = float3(0.9,0.9,1.0) * diffuse;

    // soft blend with solid geometry    
    {
        float depthBufferViewspacePos = LoadScreenDepthViewspace( int2(Input.Position.xy) );
		
		depthDiff = (depthBufferViewspacePos - Input.ViewPos.z) / mSoftParticlesSaturationDepth;
		depthDiff = smoothstep(0.0f, 1.0f, depthDiff);
	}
 
	if(mUI.wireframe)
	{
		return Input.color;
	}
	else
	{
		float3 Color = ambient + LightContrib * shadowTerm;// * gParticleOpacityNoiseTex.Sample(gDiffuseSampler, Input.UVS.xy).xyz;
		return float4(Color, depthDiff * (1.0f - segmentTransmittance));   
	}
}

[earlydepthstencil]
void ParticleAVSMCapture_PS(DynamicParticlePSIn Input)
{
    float3 entry, exit;
    float  segmentTransmittance;
	if (IntersectDynamicParticle(Input, entry, exit, segmentTransmittance)) {
	
        // Allocate a new node
        // (If we're running out of memory we simply drop this fragment
        uint newNodeAddress;       
        
        if (LT_AllocSegmentNode(newNodeAddress)) {
            // Fill node
            ListTexSegmentNode node;            
            node.depth[0] = entry.z;
            node.depth[1] = exit.z;
            node.trans    = segmentTransmittance;
            node.sortKey  = Input.ViewPos.z;
            
	        // Get fragment viewport coordinates
            int2 screenAddress = int2(Input.Position.xy);            
            
            // Insert node!
            LT_InsertFirstSegmentNode(screenAddress, newNodeAddress, node);            
        }           
	} 
}

uint ConvertPixelPosToLinearTiledAddress(uint2 pixelPos)
{
    const uint tileSize = 8;
    const uint tileSizeSq = tileSize * tileSize;
    uint2 tileId = pixelPos / tileSize;
    uint2 tilePos = pixelPos % tileSize;
    uint address = tileId.y * ((uint)mShadowMapSize/tileSize) * tileSizeSq + 
                   tileId.x * tileSizeSq + 
                   tilePos.y * tileSize + 
                   tilePos.x;

    return address;
}

void ParticleAVSMSinglePassInsert_PS(DynamicParticlePSIn Input)
{
    float  segmentTransmittance;    
    float3 entry, exit;    
    if (IntersectDynamicParticle(Input, entry, exit, segmentTransmittance)) 
    {	    
        uint fragmentCount;

        // Get fragment viewport coordinates
        int2 screenAddress = int2(Input.Position.xy);      	  
         
        //DeviceMemoryBarrier();      	  
        
        bool cond = false;
        [allow_uav_condition]while (!cond) 
        {        
            InterlockedCompareExchange(gListTexFirstSegmentNodeAddressUAV[screenAddress], 0, 1, fragmentCount);    
            
            if (1 == fragmentCount) 
            {
                cond = true;                

                // Read AVSM from RW-UAV (RWStructuredBuffer)
                uint address = ConvertPixelPosToLinearTiledAddress((uint2)Input.Position.xy);
                
                //DeviceMemoryBarrier();
                
                AVSMData avsmData = gAVSMStructBufUAV[address];
                             
                // Insert segment into AVSM
                float segmentDepth[2] = {entry.z, exit.z};
                InsertSegmentAVSM(segmentDepth, segmentTransmittance, avsmData);
                                
                // Write AVSM back to memory
                gAVSMStructBufUAV[address] = avsmData;         

                //DeviceMemoryBarrier();

                InterlockedCompareStore(gListTexFirstSegmentNodeAddressUAV[screenAddress], 1, 0);               
            } 
        }
    }
}

void AVSMClearStructuredBuf_PS(FullScreenTriangleVSOut Input)
{
    // Compute linearized address of this pixel
    uint2 screenAddress = uint2(Input.positionViewport.xy);
    uint address = screenAddress.y * (uint)(mShadowMapSize) + screenAddress.x; 

    // Initialize AVSM data (clear)
    gAVSMStructBufUAV[address] = AVSMGetEmptyNode();
}

AVSMData_PSOut AVSMConvertSUAVtoTex2D_PS(FullScreenTriangleVSOut Input)
{
    // Read AVSM from RW-UAV (RWStructuredBuffer)
    uint address = ConvertPixelPosToLinearTiledAddress((uint2)Input.positionViewport.xy);
    AVSMData avsmData = gAVSMStructBufSRV[address]; 

    // Store final AVSM data into a 2D texture
    return avsmData;
}


//////////////////////////////////////////////
/// AVSM Gen Code
//////////////////////////////////////////////

[earlydepthstencil]
void ParticleAVSM_Gen_PS(DynamicParticlePSIn Input)
{
	// Initialize Intel Shader Extensions
	IntelExt_Init();

    float3 entry, exit;
    float  segmentTransmittance;
	IntersectDynamicParticle(Input, entry, exit, segmentTransmittance); // does the ray tracing with sphere for transmittence amount + thickness (thickness doesn't get used yet)
	if (segmentTransmittance > 0)
    {
		// From now on serialize all UAV accesses (with respect to other fragments shaded in flight which map to the same pixel)
		IntelExt_BeginPixelShaderOrdering();

        const int2 pixelAddr = int2(Input.Position.xy);

#ifndef AVSM_GEN_SOFT
        const float shadowZBias = 0.9; // shift sample a bit away from shadow source to avoid self-shadowing due to shadow sample cube size (less required for bigger shadowmap dimensions and/or more nodes, and much less in case linear shadow function is added)
        entry.z += shadowZBias;
		exit.z  += shadowZBias;
#endif
		 
        float ctrlSurface = AVSMGenLoadControlSurfaceUAV(pixelAddr);
        if (0 == ctrlSurface) {
            AVSMGenData avsmData;

            // Clear and initialize avsm data with just one fragment
#ifdef AVSM_GEN_SOFT
            AVSMGenInitDataSoft(avsmData, entry.z, exit.z, segmentTransmittance);
#else
            AVSMGenInitData(avsmData, entry.z, segmentTransmittance);
#endif

			// Store AVSM data
            AVSMGenStoreRawDataUAV(pixelAddr, avsmData);
			
			// Update control surface
			ctrlSurface = 1;		
            AVSMGenStoreControlSurface(pixelAddr, ctrlSurface);
        } else {
            AVSMGenNode nodeArray[AVSM_NODE_COUNT];
               
            AVSMGenLoadDataUAV(pixelAddr, nodeArray);
#ifdef AVSM_GEN_SOFT
            AVSMGenInsertFragmentSoft(entry.z, exit.z, segmentTransmittance, nodeArray);
#else
            AVSMGenInsertFragment(entry.z, segmentTransmittance, nodeArray);
#endif
            AVSMGenStoreDataUAV(pixelAddr, nodeArray);
        }
	} 
}


#endif // COMMON_PARTICLE_FX_H
