//--------------------------------------------------------------------------------------
// Copyright 2012 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//--------------------------------------------------------------------------------------

SamplerState         PointClampSampler    : register( s2 );
SamplerState         PointWrapSampler     : register( s3 );
SamplerState         LinearClampSampler   : register( s4 );

Texture2D            DepthSource;
Texture2D            ViewspaceDepthSource;

Texture2D            TEXTURE0;
Texture2D            TEXTURE1;


struct SSAOConstants
{
   float2               ViewportSize;              // .xy == (render target size).xy
   float2               ViewportPixelSize;         // .zw == 1.0 / ViewportSize.xy

   float2               ViewportHalfSize;          // .xy == ((int2)(render target size).xy+1)/2
   float2               ViewportHalfPixelSize;     // .zw == 1.0 / ViewportHalfSize.xy

   float2               DepthUnpackConsts;
   float2               CameraTanHalfFOV;

   float2               Viewport2xPixelSize;
   float2               RandomNoiseUVScale;

   float2               CameraNearFar;
   float2               Camera2TanHalfFOV;

   float                SSAOSettingsRadius;                          // world (viewspace) maximum size of the shadow
   float                SSAOSettingsSamplingRadiusTweak;             // scales the sampling pattern (should probably stay at 0.5 to 1.0)
   float                SSAOSettingsContrastReduction;               // reduces some minor unwanted shadows (0.01 - 0.1)
   float                SSAOSettingsShadowStrength;                  // global strength of the effect (0 - 5)

   float                SSAOSettingsFadeOutFrom;                     // effect fade out from distance (ex. 25)
   float                SSAOSettingsFadeOutTo;                       // effect fade out to distance   (ex. 100)
   float                SSAOSettingsHorizonAngleThreshold;           // limit errors on slopes and caused by insufficient geometry tessellation (0.05 to 0.5)
   float                SSAOSettingsSamplingRadiusNearLimit;         // if viewspace pixel closer than this, don't enlarge shadow sampling radius anymore (makes no sense to grow beyond some distance, not enough samples to cover everything, so just limit the shadow growth; could be SSAOSettingsFadeOutFrom * 0.1 or less)
};

cbuffer cbSSAO// : register(b4)
{
   SSAOConstants  g_SSAOConsts;
};

float ScreenToViewDepth( float screenDepth )
{
   float depthHackMul = g_SSAOConsts.DepthUnpackConsts.x;
   float depthHackAdd = g_SSAOConsts.DepthUnpackConsts.y;

   // Optimised version of "-cameraClipNear / (cameraClipFar - projDepth * (cameraClipFar - cameraClipNear)) * cameraClipFar"

   // Set your depthHackMul and depthHackAdd to:
   // depthHackMul = ( cameraClipFar * cameraClipNear) / ( cameraClipFar - cameraClipNear );
   // depthHackAdd = cameraClipFar / ( cameraClipFar - cameraClipNear );

 	return depthHackMul / (depthHackAdd - screenDepth);
}

// from [0, width], [0, height] to [-1, 1], [-1, 1]
float2 ScreenToClipPosition2D( float2 screenPos )
{
   return screenPos * g_SSAOConsts.Viewport2xPixelSize.xy - float2( 1.0f, 1.0f );
}

float3 ScreenToViewPosition( float2 screenPos, float viewspaceDepth )
{
  	return float3( g_SSAOConsts.CameraTanHalfFOV.xy * viewspaceDepth * ScreenToClipPosition2D( screenPos ), viewspaceDepth );
}

float3 ClipToViewPosition( float2 clipPos, float viewspaceDepth )
{
  	return float3( g_SSAOConsts.CameraTanHalfFOV.xy * viewspaceDepth * clipPos, viewspaceDepth );
}


// ********************************************************************************************************

struct VS_INPUT
{
    float3 Pos     : POSITION;
    float2 Uv      : TEXCOORD0;
};

struct PS_INPUT
{
    float4 Pos     : SV_POSITION;
    float2 Uv      : TEXCOORD0;
};

float4 PackEdges( float ssaoShadow, float4 edgesLRTB )
{
	return float4( ssaoShadow, 
			edgesLRTB.x * 0.5 + edgesLRTB.y * 0.25, 
			edgesLRTB.z * 0.5 + edgesLRTB.w * 0.25,
			1 );

}

float4  UnpackEdges( float4 ssaoValue )
{
	float4 edgesLRTB;
	edgesLRTB.x = ssaoValue.y > 0.45;
	edgesLRTB.z = ssaoValue.z > 0.45;
	edgesLRTB.y = (ssaoValue.y*2 - edgesLRTB.x ) > 0.1;
	edgesLRTB.w = (ssaoValue.z*2 - edgesLRTB.z ) > 0.1;
	return edgesLRTB;
}

// ********************************************************************************************************

PS_INPUT VSMain( VS_INPUT input )
{
   PS_INPUT output = (PS_INPUT)0;
   output.Pos  = float4( input.Pos, 1.0f);
   output.Uv   = input.Uv;

   output.Pos.z = g_SSAOConsts.DepthUnpackConsts.y - g_SSAOConsts.DepthUnpackConsts.x / g_SSAOConsts.SSAOSettingsFadeOutTo;

   return output;
}

// Half-res prepare depth
float PSPrepareDepth( PS_INPUT input /*float4 SVPos : SV_POSITION*/ ) : SV_Depth
{
   // move to upper left centre pixel
   float2 normalizedScreenPos = ( input.Pos.xy - float2( 0.25, 0.25 ) ) * g_SSAOConsts.ViewportHalfPixelSize;

   // add pseudorandom offset to move sampling between four underlying high-res pixels; this noise is later
   // removed by blurring passes.
   normalizedScreenPos.xy += frac( (input.Pos.yx - 0.5) / 2.0 ) * g_SSAOConsts.ViewportHalfPixelSize;
   
   float depthViewSpace = ScreenToViewDepth( DepthSource.SampleLevel( PointClampSampler, normalizedScreenPos.xy, 0.0 ).x );

   return 1.0 / depthViewSpace;
}

// ********************************************************************************************************
// Pixel shader that does smart blurring (to avoid bleeding)

void AddSample( float ssaoValue, float edgeValue, inout float sum, inout float sumWeight )
{
   // make shadows expand (up to 2x more weight to shadowed areas)
   edgeValue *= (1.5 - ssaoValue);

   sum         += ssaoValue * edgeValue;
   sumWeight   += edgeValue;
}

float4 SampleBlurred( Texture2D ssaoTex, SamplerState ssaoSampler, float2 uv, uniform bool horizontal, uniform bool vertical, const float2 pixelSize )
{
	float4 ssaoValue = ssaoTex.Sample( ssaoSampler, uv );
	float4 edgesLRTB = UnpackEdges( ssaoValue );
	
   float sum         = 0.0f;
   float sumWeight   = 0.0f;
   AddSample( ssaoValue.x, 1.0f, sum, sumWeight );

   if( horizontal )
   {
	   float ssaoValueL = ssaoTex.Sample( ssaoSampler, uv + float2( -pixelSize.x, 0 ) ).x;
	   float ssaoValueR = ssaoTex.Sample( ssaoSampler, uv + float2( +pixelSize.x, 0 ) ).x;
	
      AddSample( ssaoValueL, edgesLRTB.x, sum, sumWeight );
      AddSample( ssaoValueR, edgesLRTB.y, sum, sumWeight );
   }

   if( vertical )
   {
   	float ssaoValueT = ssaoTex.Sample( ssaoSampler, uv + float2( 0, -pixelSize.y ) ).x;
	   float ssaoValueB = ssaoTex.Sample( ssaoSampler, uv + float2( 0, +pixelSize.y ) ).x;
	
      AddSample( ssaoValueT, edgesLRTB.z, sum, sumWeight );
      AddSample( ssaoValueB, edgesLRTB.w, sum, sumWeight );
   }

   ssaoValue.x = sum / sumWeight;

	return ssaoValue;
}

// These sample patterns were generated using the Poisson Disk Generator (http://www.coderhaus.com/?p=11), modified to output
// samples in clockwise direction around 0, 0 and output 90 degrees 'pairs' for best normal generation

static const uint g_numSamples[3] = { 10, 20, 36 };

static const float2 g_samplePattern0[] =
{
    0.389000f, -0.551210f,
    0.958450f, -0.188280f,
    0.517400f,  0.139540f,
    0.584340f,  0.654050f,
    0.080010f,  0.915360f,
   -0.509720f,  0.761340f,
   -0.666420f,  0.248440f,
   -0.825540f, -0.494270f,
   -0.285580f, -0.524180f,
   -0.032960f, -0.974140f
};

static const int g_samplePatternPairs0[] =
{
   3,
   4,
   5,
   6,
   7,
   8,
   9,
   0,
   1,
   2
};

static const float2 g_samplePattern1[] =
{
    0.121830f, -0.958710f,
    0.097960f, -0.416020f,
    0.602820f, -0.765500f,
    0.507610f, -0.379090f,
    0.929010f, -0.289520f,
    0.628940f,  0.095330f,
    0.926990f,  0.344260f,
    0.301610f,  0.352000f,
    0.523730f,  0.674820f,
    0.190140f,  0.869320f,
   -0.086200f,  0.425170f,
   -0.208860f,  0.960200f,
   -0.564720f,  0.695460f,
   -0.441830f,  0.316650f,
   -0.884650f,  0.421480f,
   -0.962170f, -0.005320f,
   -0.595930f, -0.125500f,
   -0.729640f, -0.581790f,
   -0.365270f, -0.430250f,
   -0.225870f, -0.776130f
};

static const int g_samplePatternPairs1[] =
{
   5,
   6,
   7,
   8,
   9,
   10,
   11,
   13,
   14,
   15,
   16,
   17,
   18,
   19,
   0,
   1,
   2,
   3,
   4,
   4
};

static const float2 g_samplePattern2[] =
{
    0.034740f, -0.262550f,
    0.148030f, -0.972900f,
    0.127150f, -0.631000f,
    0.373860f, -0.772450f,
    0.272800f, -0.410180f,
    0.593490f, -0.547740f,
    0.426860f, -0.185570f,
    0.777860f, -0.332430f,
    0.720750f,  0.005530f,
    0.983180f,  0.020190f,
    0.907790f,  0.269250f,
    0.417850f,  0.150340f,
    0.757830f,  0.484080f,
    0.490320f,  0.401140f,
    0.585210f,  0.702190f,
    0.327000f,  0.628660f,
    0.157970f,  0.346930f,
    0.359290f,  0.887810f,
   -0.016340f,  0.878500f,
   -0.041360f,  0.598850f,
   -0.416080f,  0.895590f,
   -0.303450f,  0.609010f,
   -0.248750f,  0.315560f,
   -0.676520f,  0.611340f,
   -0.761010f,  0.350670f,
   -0.563430f,  0.071460f,
   -0.959130f,  0.058120f,
   -0.725770f, -0.162570f,
   -0.920020f, -0.377420f,
   -0.271690f, -0.113250f,
   -0.485240f, -0.322700f,
   -0.673310f, -0.522750f,
   -0.418750f, -0.639410f,
   -0.222080f, -0.420710f,
   -0.386790f, -0.912060f,
   -0.121190f, -0.888030f
};

static const int g_samplePatternPairs2[] =
{
   9,
   10,
   11,
   12,
   13,
   14,
   16,
   17,
   18,
   19,
   20,
   21,
   22,
   23,
   24,
   25,
   26,
   27,
   28,
   29,
   30,
   31,
   32,
   33,
   34,
   35,
   0,
   2,
   3,
   4,
   5,
   6,
   7,
   8,
   1,
   15
};

float LoadScreenDepth( float2 pos )
{
   return DepthSource.SampleLevel( PointClampSampler, pos.xy, 0.0 ).x;
}

float3 ConvertToViewspace( float2 pos, float depthViewSpace )
{
   float3 ret;
   
   float2 clipPos2D = pos.xy * float2( 2.0, -2.0 ) + float2( -1.0, 1.0 );
   ret.xyz = ClipToViewPosition( clipPos2D, depthViewSpace );
   
   return ret;
}

float3 LoadAndConvertToViewspace( float2 pos )
{
   return ConvertToViewspace( pos, ScreenToViewDepth( LoadScreenDepth( pos ) ) );
}

void CalculateNormal( float3 pixelCenterPos, float3 pixLPos, float3 pixRPos, float3 pixTPos, float3 pixBPos, float4 edgesLRTB, out float normalQuality, out float3 normal, bool highQuality )
{
   normal = float3( 0, 0, -0.001 );

   // these can be optimized by using vertex shader interpolants to calculate viewspace pos
   float3 deltaLViewspacePos = normalize( (pixLPos.xyz) - pixelCenterPos );
	float3 deltaRViewspacePos = normalize( (pixRPos.xyz) - pixelCenterPos );
	float3 deltaTViewspacePos = normalize( (pixTPos.xyz) - pixelCenterPos );
	float3 deltaBViewspacePos = normalize( (pixBPos.xyz) - pixelCenterPos );

   float4 acceptedNormals = float4( edgesLRTB.x*edgesLRTB.z, edgesLRTB.z*edgesLRTB.y, edgesLRTB.y*edgesLRTB.w, edgesLRTB.w*edgesLRTB.x );

   normal += cross( deltaLViewspacePos, deltaTViewspacePos ) * (acceptedNormals.x);
   normal += cross( deltaTViewspacePos, deltaRViewspacePos ) * (acceptedNormals.y);
   normal += cross( deltaRViewspacePos, deltaBViewspacePos ) * (acceptedNormals.z);
   normal += cross( deltaBViewspacePos, deltaLViewspacePos ) * (acceptedNormals.w);
	
   // a tiny bit more precision or it can be used alone for much worse precision but better performance
   if( highQuality )
   {
      float3 deltaLRViewspacePos = normalize( pixLPos.xyz - pixRPos.xyz );
      float3 deltaTBViewspacePos = normalize( pixTPos.xyz - pixBPos.xyz );
      normal += 2 * cross( deltaLRViewspacePos, deltaTBViewspacePos ) * (edgesLRTB.x*edgesLRTB.z*edgesLRTB.y*edgesLRTB.w);
   }

   normalQuality = saturate( dot( edgesLRTB, float4( 1.0, 1.0, 1.0, 1.0 ) ) - 2.5 ) + 0.00001;
	
   normal = normalize( normal );
}

void GenerateSSAOShadowsInternal( out float outShadowTerm, out float4 outEdges, float4 SVPos, uniform int quality, uniform bool halfRes, uniform bool sampleMirrored )
{
   float2 normalizedScreenPos;

   float pseudoRand = frac( sin( dot( SVPos.xy, float2(0.0129898,0.078233) ) ) * 43758.5453 );

   float2 normalizedScreenPos2;

   if( halfRes )
   {
      // move to the center of the upper left pixel (of the full res depth buffer)
      float2 normalizedScreenPosUpperLeft = ( SVPos.xy - float2( 0.25, 0.25 ) ) * g_SSAOConsts.ViewportHalfPixelSize;

      // add pseudorandom offset to move sampling between four underlying high-res pixels; this noise is later
      // removed by blurring passes.
      float2 pseudoRandomOffset = frac( (SVPos.yx - 0.5) / 2.0 );
         
      if( sampleMirrored )
      {
         pseudoRandomOffset.xy = float2( pseudoRandomOffset.y, 0.5 - pseudoRandomOffset.x );
      }
      
      normalizedScreenPos.xy           = normalizedScreenPosUpperLeft + pseudoRandomOffset * g_SSAOConsts.ViewportHalfPixelSize;
   }
   else
   {
       normalizedScreenPos = SVPos.xy * g_SSAOConsts.ViewportPixelSize;
   }


   // always use point sampler for the center (base) pixel   
   float3 pixCenterPos = LoadAndConvertToViewspace( normalizedScreenPos );

   // early out!
   if( pixCenterPos.z > g_SSAOConsts.SSAOSettingsFadeOutTo )
   {
      outShadowTerm = 1.0;
      outEdges = float4( 0, 0, 0, 0 );
      return;
   }

   const float lookupRadius = g_SSAOConsts.SSAOSettingsRadius;

   // the radius is around a point in the z plane around pixCenterPos; projected to actual sloped surface they will be further away so accept them
   const float acceptableDeviationK = 2.5;
	const float acceptRadiusThreshold = lookupRadius * acceptableDeviationK;

   // If neighbouring pixel are at more than edgeAcceptDistanceRange, it's an edge. Crude but effective.
   // (used to use lookupRadius-based range initially but that causes edge-detection problems)
   const float edgeAcceptDistanceRange = 0.08;
   float edgeAcceptDistance = min( acceptRadiusThreshold*1.25, pixCenterPos.z * edgeAcceptDistanceRange );

   float step     = (halfRes)?(2.0f):(1.0f);
   float miniStep = 0.0f;

   float3 pixLPos = LoadAndConvertToViewspace( normalizedScreenPos + float2( -step,      -miniStep ) * g_SSAOConsts.ViewportPixelSize );
   float3 pixRPos = LoadAndConvertToViewspace( normalizedScreenPos + float2(  step,       miniStep ) * g_SSAOConsts.ViewportPixelSize );
   float3 pixTPos = LoadAndConvertToViewspace( normalizedScreenPos + float2(  miniStep,  -step     ) * g_SSAOConsts.ViewportPixelSize );
   float3 pixBPos = LoadAndConvertToViewspace( normalizedScreenPos + float2( -miniStep,   step     ) * g_SSAOConsts.ViewportPixelSize );
                                                                                
	float4 edgesLRTB = float4( pixLPos.z, pixRPos.z, pixTPos.z, pixBPos.z );
	edgesLRTB = abs( edgesLRTB - pixCenterPos.zzzz );

   // These edges are used to accept normals and to prevent blurring across edges
	edgesLRTB = edgesLRTB < edgeAcceptDistance;

	float3 combinedNormal;
   float normalQualityScale;

   CalculateNormal( pixCenterPos, pixLPos, pixRPos, pixTPos, pixBPos, edgesLRTB, normalQualityScale, combinedNormal, false );

   // fit our screen sampling pattern inside viewspace 'lookupRadius'
   float2 pixLookupRadiusMod;
   {
      const float radiusScaleLimitDistance = g_SSAOConsts.SSAOSettingsSamplingRadiusNearLimit;
      float viewspaceDepthRadiusMod = (pixCenterPos.z < radiusScaleLimitDistance)?(pixCenterPos.z/radiusScaleLimitDistance):(1.0);

      float3 pixRBCenterPos = ClipToViewPosition( (normalizedScreenPos.xy + g_SSAOConsts.ViewportPixelSize) * float2( 2.0, -2.0 ) + float2( -1.0, 1.0 ), pixCenterPos.z );

      float2 viewspacePixelSizeAtCenter = float2( pixRBCenterPos.x-pixCenterPos.x, pixCenterPos.y-pixRBCenterPos.y );

      pixLookupRadiusMod = g_SSAOConsts.ViewportPixelSize * (lookupRadius.xx / viewspacePixelSizeAtCenter) * viewspaceDepthRadiusMod * g_SSAOConsts.SSAOSettingsSamplingRadiusTweak;

      // make sure we're not sampling too close to the center pixel (otherwise we'll sample it too which causes artifacts)
      pixLookupRadiusMod = max( pixLookupRadiusMod, g_SSAOConsts.ViewportPixelSize*3.0 );
   }

   // randomize sampling pattern radius slightly (sample stochastically to a degree, and blur passes smoothen the results later)
   if( quality >= 1 )
   {
      const float randRadius = 0.1;
      pixLookupRadiusMod *= 1.0 + randRadius*0.5 - randRadius * pseudoRand;
   }


   // smoothen the center point depth just a bit
   {
      //float minDepth = min( pixCenterPos.z, min( min( pixLPos.z, pixRPos.z ), min( pixTPos.z, pixBPos.z ) ) );
      //float maxDepth = max( pixCenterPos.z, max( max( pixLPos.z, pixRPos.z ), max( pixTPos.z, pixBPos.z ) ) );
      //pixCenterPos.z = pixCenterPos.z * 0.6 + minDepth * 0.3 + maxDepth * 0.1;

      pixCenterPos.z = pixCenterPos.z * 0.4 + dot( float4( pixLPos.z, pixRPos.z, pixTPos.z, pixBPos.z ), float4( 0.15, 0.15, 0.15, 0.15 ) );

      float2 clipPos2D = normalizedScreenPos.xy * float2( 2.0, -2.0 ) + float2( -1.0, 1.0 );
      pixCenterPos = ClipToViewPosition( clipPos2D, pixCenterPos.z );
   }
                                    
   //////////////////////////////////////////////////////////////////////////////////////////
   // debug section
   //return float4( 1 - edgesLRTB.x * 0.4 - edgesLRTB.y * 0.8, 1 - edgesLRTB.z * 0.4 - edgesLRTB.w * 0.8, 0, 1 );
   ////return float4( combinedNormal.zzz, 1 );
   //return float4( 0.5 + combinedNormal.xyz * 0.5, 1 );
   //return float4( normalQualityScale.xxx, 1 );
   ////return float4( 0.5 + cos( combinedNormal.xyz * 0.52359 ), 1 );
   //return float4( 0.5 + combinedNormal.zzz * 0.5, 1 );
   //return float4( edgesLRTB.xxx, 1 );
   //////////////////////////////////////////////////////////////////////////////////////////

   int additionalOccWeight = 0;

   const int numberOfSteps = g_numSamples[quality];

   float4 samplesViewspace[64];
   
   {
      [unroll]
	   for( int i = 0; i < numberOfSteps; i++ )
	   {
	      float2 sampleOffset;

         // pattern
         if( quality >= 2 )
            sampleOffset = g_samplePattern2[i];
         else if( quality == 1 )
            sampleOffset = g_samplePattern1[i];
         else if( quality == 0 )
            sampleOffset = g_samplePattern0[i];

         if( sampleMirrored )
         {
            sampleOffset.xy = sampleOffset.yx * 0.8;
         }
         else
         {
            sampleOffset.xy = sampleOffset.xy * 1.25;
         }

         // scale by lookup radius
         sampleOffset *= pixLookupRadiusMod.xy;

         float3 hitPos = LoadAndConvertToViewspace( normalizedScreenPos + sampleOffset ).xyz;
         float3 hitDelta = hitPos - pixCenterPos;
         float hitDeltaLength = length(hitDelta);
         hitDelta /= hitDeltaLength;
      
         float scale = saturate(1 - (hitDeltaLength*1.1-lookupRadius)/(acceptRadiusThreshold-lookupRadius) );
      
         samplesViewspace[i] = float4( hitDelta, scale );
      }
   }

   //combinedNormal.xyz = 0;

// calculate normals with our sampled points
#if 1
   {
      [unroll]
	   for( int i = ((sampleMirrored)?(1):(0)); i < numberOfSteps; i += 2 )
	   {
         // get our pair. all this is precompiled, not done at runtime!
         int j;
         if( quality >= 2 )
            j = g_samplePatternPairs2[i];
         else if( quality == 1 )
            j = g_samplePatternPairs1[i];
         else if( quality == 0 )
            j = g_samplePatternPairs0[i];

         float3 norm;
         if( sampleMirrored )
            norm = normalize( cross( samplesViewspace[j].xyz, samplesViewspace[i].xyz ) );
         else
            norm = normalize( cross( samplesViewspace[i].xyz, samplesViewspace[j].xyz ) );

         norm = normalize( norm );

         combinedNormal += norm * samplesViewspace[i].w * samplesViewspace[j].w;
      }
      combinedNormal = normalize( combinedNormal );
   }
#endif

   //return float4( 0.5 + combinedNormal.xyz * 0.5, 1 );

   float occlusionSum = 0.0;
   float weightSum = 0.0;
   {
      [unroll]
	   for( int i = 0; i < numberOfSteps; i++ )
	   {
         float scale = samplesViewspace[i].w;
         float3 hitDelta = samplesViewspace[i].xyz;

         float occlusion = saturate( dot( combinedNormal, hitDelta.xyz ) - g_SSAOConsts.SSAOSettingsHorizonAngleThreshold ) * scale;

         occlusionSum += occlusion;
         weightSum += 1.0;
	   }
   }
	float occlusion = occlusionSum / weightSum;

   // limit the max
   occlusion = min( occlusion, (quality==0)?(0.85):(0.87) );
   
   const float SSAOSettingsFadeOutRange = g_SSAOConsts.SSAOSettingsFadeOutTo - g_SSAOConsts.SSAOSettingsFadeOutFrom;
   float fadeOut = (pixCenterPos.z - g_SSAOConsts.SSAOSettingsFadeOutFrom) / SSAOSettingsFadeOutRange;
	fadeOut = 1 - saturate(fadeOut);

   #if 0
   // Fog-based fadeout should go here - enable if fog applied before the SSAO pass
   {
	   float	fogIntensity = YourEngineFogCalcAndHopefullyItIsViewspaceBased( pixCenterPos.z );
      // rapidly reduce the effect based on fog
	   fadeOut *= (fogIntensity * fogIntensity);
   }
   #endif
	
   #if 1
	// Reduce the SSAO shadowing if we're on the edge (there are some issues on edges that cause artifacts, especially when MSAA is enabled)
	fadeOut *= saturate( dot( edgesLRTB, float4( 0.75, 0.75, 0.75, 0.75 ) ) - 1.75 );
   #endif
	
   // contrast
	occlusion = occlusion - g_SSAOConsts.SSAOSettingsContrastReduction;

   // strength
	occlusion = saturate( occlusion * g_SSAOConsts.SSAOSettingsShadowStrength );

   // fadeout
   occlusion *= fadeOut;

   outShadowTerm  = 1.0 - occlusion;
   outEdges       = edgesLRTB;
}

float4 GenerateSSAOShadows( float4 SVPos, uniform int quality )
{
   float shadowTerm;
   float4 edges;
   
   if( quality == 0 )
   {
      GenerateSSAOShadowsInternal( shadowTerm, edges, SVPos, 0, true, false );
   }
   else if( quality == 1 )
   {
      float shadowTerm1;
      float4 edges1;
      GenerateSSAOShadowsInternal( shadowTerm, edges, SVPos, 0, true, false );

      GenerateSSAOShadowsInternal( shadowTerm1, edges1, SVPos, 0, true, true );
      shadowTerm = (shadowTerm + shadowTerm1) * 0.5;
      edges *= edges1;
   }
   else if( quality == 2 )
   {
      float shadowTerm1;
      float4 edges1;
      GenerateSSAOShadowsInternal( shadowTerm, edges, SVPos, 1, true, false );
      if( (shadowTerm < 1) )
      {
         GenerateSSAOShadowsInternal( shadowTerm1, edges1, SVPos, 1, true, true );
         shadowTerm = (shadowTerm + shadowTerm1) * 0.5;
         edges *= edges1;
      }
   }
   else if( quality == 3 )
   {
      GenerateSSAOShadowsInternal( shadowTerm, edges, SVPos, 2, false, false );
   }

   //return float4( edges.xy, 0, 1 );

   return PackEdges( shadowTerm, edges );
}

float4 PSGenerate0( float4 SVPos : SV_POSITION ) : SV_Target
{
   return GenerateSSAOShadows( SVPos, 0 );
}

float4 PSGenerate1( float4 SVPos : SV_POSITION ) : SV_Target
{
   return GenerateSSAOShadows( SVPos, 1 );
}

float4 PSGenerate2( float4 SVPos : SV_POSITION ) : SV_Target
{
   return GenerateSSAOShadows( SVPos, 2 );
}

float4 PSGenerate3( float4 SVPos : SV_POSITION ) : SV_Target
{
   return GenerateSSAOShadows( SVPos, 3 );
}


[earlydepthstencil]
float4 PSBlur( float4 SVPos : SV_POSITION ) : SV_Target
{
   float2 normalizedScreenPos = SVPos.xy * g_SSAOConsts.ViewportPixelSize;

   return SampleBlurred( TEXTURE0, PointClampSampler, normalizedScreenPos, true, true, g_SSAOConsts.ViewportPixelSize );
}

float4 PSBlurHalfRes( float4 SVPos : SV_POSITION ) : SV_Target
{
   float2 normalizedScreenPos = SVPos.xy * g_SSAOConsts.ViewportHalfPixelSize;

   return SampleBlurred( TEXTURE0, PointClampSampler, normalizedScreenPos, true, true, g_SSAOConsts.ViewportHalfPixelSize );
}

[earlydepthstencil]
float4 PSApply( float4 SVPos : SV_POSITION ) : SV_Target
{
   float2 normalizedScreenPos = SVPos.xy * g_SSAOConsts.ViewportPixelSize;

   return float4( SampleBlurred( TEXTURE0, PointClampSampler, normalizedScreenPos, true, true, g_SSAOConsts.ViewportPixelSize ).xxx, 0 );
}

[earlydepthstencil]
float4 PSApplyNoBlur( float4 SVPos : SV_POSITION ) : SV_Target
{
   float2 normalizedScreenPos = SVPos.xy * g_SSAOConsts.ViewportPixelSize;

   // show edges
   #if 0
   float4 zmz = TEXTURE0.Sample( PointClampSampler, normalizedScreenPos );
   float4 edgesLRTB = 1 - UnpackEdges( zmz );
   return float4( 1-zmz.x, edgesLRTB.x + edgesLRTB.y, edgesLRTB.z + edgesLRTB.w, 1 );
   #endif

   return float4( TEXTURE0.Sample( PointClampSampler, normalizedScreenPos ).xxx, 0 );
}

[earlydepthstencil]
float4 PSApplyHalfRes( float4 SVPos : SV_POSITION ) : SV_Target
{
   float2 normalizedScreenPos       = SVPos * g_SSAOConsts.ViewportPixelSize;
   
   // move to the upper left pixel corner
   float2 normalizedScreenPosHalf   = normalizedScreenPos - g_SSAOConsts.ViewportPixelSize * 0.5;
   
   // move to the middle of the half pixel size
   normalizedScreenPosHalf += g_SSAOConsts.ViewportHalfPixelSize * 0.5;
   
   return float4( TEXTURE0.Sample( LinearClampSampler, normalizedScreenPosHalf ).xxx, 0 );
   //return float4( TEXTURE0.Sample( LinearClampSampler, normalizedScreenPosHalf ).xyz, 0 );   
}