//--------------------------------------------------------------------------------------
// Copyright 2011,2012,2013 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//--------------------------------------------------------------------------------------


#include "SSAOTechnique.h"


struct SSAOConstants
{
    XMFLOAT2       ViewportSize;              // .xy == (render target size).xy
    XMFLOAT2       ViewportPixelSize;         // .zw == 1.0 / ViewportSize.xy

    XMFLOAT2       ViewportHalfSize;          // .xy == ((int2)(render target size).xy+1)/2
    XMFLOAT2       ViewportHalfPixelSize;     // .zw == 1.0 / ViewportHalfSize.xy

    XMFLOAT2       DepthUnpackConsts;
    XMFLOAT2       CameraTanHalfFOV;

    XMFLOAT2       Viewport2xPixelSize;
    XMFLOAT2       RandomNoiseUVScale;

    XMFLOAT2       CameraNearFar;
    XMFLOAT2       Camera2TanHalfFOV;

    float          SSAOSettingsRadius;
    float          SSAOSettingsSamplingRadiusTweak;
    float          SSAOSettingsContrastReduction;
    float          SSAOSettingsShadowStrength;

    float          SSAOSettingsFadeOutFrom;   
    float          SSAOSettingsFadeOutTo;
    float          SSAOSettingsHorizonAngleThreshold;
    float          SSAOSettingsSamplingRadiusNearLimit;
};


void SSAOTechnique::Create( ID3D11Device* d3dDevice, ID3D11DeviceContext * context )
{
    CPUTAssetLibrary *pAssetLibrary = CPUTAssetLibrary::GetAssetLibrary();

    // Create our constant buffer.
    {
        D3D11_BUFFER_DESC bd = {0};
        bd.ByteWidth = sizeof(SSAOConstants);
        bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
        bd.Usage = D3D11_USAGE_DYNAMIC;
        bd.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

        ID3D11Buffer * pOurConstants;
        HRESULT hr = (CPUT_DX11::GetDevice())->CreateBuffer( &bd, NULL, &pOurConstants );

        cString name = _L("$cbSSAO");
        mpSSAOConstantsBuffer = new CPUTBufferDX11( name, pOurConstants );

        pAssetLibrary->AddConstantBuffer( name, mpSSAOConstantsBuffer );
        SAFE_RELEASE(pOurConstants); // We're done with it.  The CPUTBuffer now owns it.

        CPUTMaterial::mGlobalProperties.AddValue( _L("cbSSAO"), _L("$cbSSAO") );
    }

    // Working render targets
    {
        int width, height;
        CPUTOSServices::GetOSServices()->GetClientDimensions(&width, &height);

        DXGI_FORMAT scratchFormat = DXGI_FORMAT_R8G8B8A8_UNORM;
        //DXGI_FORMAT_B5G6R5_UNORM uses less memory and in theory should be faster but there are various issues with it on more than one platform
#if 0
        scratchFormat = DXGI_FORMAT_B5G6R5_UNORM;
#endif
        mpScratchHalfResRenderTarget = new CPUTRenderTargetColor();
        mpScratchHalfResRenderTarget->CreateRenderTarget( cString(_L("$ScratchHalfResRenderTarget")), (width+1)/2, (height+1)/2, scratchFormat, 1, false );

        mpScratch2HalfResRenderTarget = new CPUTRenderTargetColor();
        mpScratch2HalfResRenderTarget ->CreateRenderTarget( cString(_L("$Scratch2HalfResRenderTarget")), (width+1)/2, (height+1)/2, scratchFormat, 1, false );
    }

    mpMaterialSSAOGenerate  = pAssetLibrary->GetMaterial( L"SSAO_generate" );
    mpMaterialSSAOBlurHalfResA = pAssetLibrary->GetMaterial( L"SSAO_blurHalfResA" );
    mpMaterialSSAOBlurHalfResB = pAssetLibrary->GetMaterial( L"SSAO_blurHalfResB" );
    mpMaterialSSAOApplyHalfRes = pAssetLibrary->GetMaterial( L"SSAO_applyHalfRes" );

    mpFullscreenSprite = new CPUTSprite();
    mpFullscreenSprite->CreateSprite( -1.0f, -1.0f, 2.0f, 2.0f, _L("Sprite") );
}

void SSAOTechnique::ResizeWindow( UINT width, UINT height )
{
    if( mpScratchHalfResRenderTarget != NULL ) mpScratchHalfResRenderTarget->RecreateRenderTarget( (width+1)/2, (height+1)/2 );
    if( mpScratch2HalfResRenderTarget != NULL ) mpScratch2HalfResRenderTarget->RecreateRenderTarget( (width+1)/2, (height+1)/2 );
}

void SSAOTechnique::UpdateConsts( ID3D11DeviceContext * dxContext, CPUTCamera * camera )
{
    int viewportWidth;
    int viewportHeight;
    CPUTOSServices::GetOSServices()->GetClientDimensions( &viewportWidth, &viewportHeight );
    int viewportHalfResWidth   = mpScratchHalfResRenderTarget->GetWidth();
    int viewportHalfResHeight  = mpScratchHalfResRenderTarget->GetHeight();

    if( mpSSAOConstantsBuffer )
    {
        ID3D11Buffer *pBuffer = mpSSAOConstantsBuffer->GetNativeBuffer();

        // update parameters of constant buffer
        D3D11_MAPPED_SUBRESOURCE mapInfo;
        dxContext->Map( pBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mapInfo );
        {
            SSAOConstants & consts = *((SSAOConstants*)mapInfo.pData);

            consts.ViewportSize           = XMFLOAT2( (float)viewportWidth, (float)viewportHeight );
            consts.ViewportPixelSize      = XMFLOAT2( 1.0f / (float)viewportWidth, 1.0f / (float)viewportHeight );

            consts.ViewportHalfSize       = XMFLOAT2( (float)viewportHalfResWidth, (float)viewportHalfResHeight );
            consts.ViewportHalfPixelSize  = XMFLOAT2( 1.0f / (float)consts.ViewportHalfSize.x, 1.0f / (float)consts.ViewportHalfSize.y );

            {
                float clipNear = camera->GetNearPlaneDistance();
                float clipFar  = camera->GetFarPlaneDistance();
                float depthHackMul = ( clipFar * clipNear) / ( clipFar - clipNear );
                float depthHackAdd = clipFar / ( clipFar - clipNear );
                consts.DepthUnpackConsts = XMFLOAT2( depthHackMul, depthHackAdd );
            }

            {
                // this will only work if xxxPerspectiveFovLH (or equivalent) is used to create projection matrix

                float tanHalfFOVY	= tanf( camera->GetFov() * 0.5f );
                float tanHalfFOVX	= tanHalfFOVY * camera->GetAspectRatio();
                consts.CameraTanHalfFOV = XMFLOAT2( tanHalfFOVX, tanHalfFOVY );

                consts.Viewport2xPixelSize    = XMFLOAT2( consts.ViewportPixelSize.x * 2.0f, consts.ViewportPixelSize.y * 2.0f );
                consts.RandomNoiseUVScale     = XMFLOAT2( 0.0f, 0.0f ); // unused // consts.ViewportSize.x / (float)cRandomNoiseTextureSize, consts.ViewportSize.y / (float)cRandomNoiseTextureSize );

                consts.CameraNearFar          = XMFLOAT2( camera->GetNearPlaneDistance(), camera->GetFarPlaneDistance() );
                consts.Camera2TanHalfFOV      = XMFLOAT2( 2.0f * tanHalfFOVX, 2.0f * tanHalfFOVY );
            }

            {
                consts.SSAOSettingsRadius                       = 0.22f;
                consts.SSAOSettingsSamplingRadiusTweak          = 0.60f;
                consts.SSAOSettingsContrastReduction            = 0.01f;
                consts.SSAOSettingsShadowStrength               = 3.0f;
                consts.SSAOSettingsFadeOutFrom                  = 50.0f;
                consts.SSAOSettingsFadeOutTo                    = 250.0f;
                consts.SSAOSettingsHorizonAngleThreshold        = 0.10f;
                consts.SSAOSettingsSamplingRadiusNearLimit      = 2.0f;

            }
        }
        dxContext->Unmap(pBuffer,0);

    }
}

void SSAOTechnique::GenerateSSAO( ID3D11DeviceContext * dxContext )
{
    //CPUTPerfTaskMarker cput_marker(D3DCOLOR(0xff0000), _L("SSAO Generate"));
    CPUTRenderParametersDX renderParams(dxContext);


    // renderParams.dxContext->OMSetDepthStencilState( mpDepthStencilStateFullscreen, 0 );

    mpScratchHalfResRenderTarget->SetRenderTarget( renderParams, NULL );

    mpFullscreenSprite->DrawSprite( renderParams, *mpMaterialSSAOGenerate );

    ID3D11ShaderResourceView * nullSRVs[8] = { NULL };
    dxContext->PSSetShaderResources( 0, 8, nullSRVs );

    mpScratchHalfResRenderTarget->RestoreRenderTarget( renderParams );
}

void SSAOTechnique::FullscreenApplySSAO( ID3D11DeviceContext * dxContext )
{
    CPUTRenderParametersDX renderParams( dxContext );

    // Pass: Blur
    {
        //CPUTPerfTaskMarker cput_marker(D3DCOLOR(0xff0000), _L("BLUR"));

        //renderParams.dxContext->OMSetDepthStencilState( mpDepthStencilStateFullscreen, 0 );

        const int blurPasses = 2;
        for( int i = 0; i < blurPasses; i++ )
        {
            mpScratch2HalfResRenderTarget->SetRenderTarget( renderParams, NULL );

            mpFullscreenSprite->DrawSprite( renderParams, *mpMaterialSSAOBlurHalfResA );
            ID3D11ShaderResourceView * nullSRVs[8] = { NULL };
            dxContext->PSSetShaderResources( 0, 8, nullSRVs );

            mpScratch2HalfResRenderTarget->RestoreRenderTarget( renderParams );

            mpScratchHalfResRenderTarget->SetRenderTarget( renderParams, NULL );

            mpFullscreenSprite->DrawSprite( renderParams, *mpMaterialSSAOBlurHalfResB );
            dxContext->PSSetShaderResources( 0, 8, nullSRVs );

            mpScratchHalfResRenderTarget->RestoreRenderTarget( renderParams );
        }
    }

    // Pass: Apply
    {
        mpFullscreenSprite->DrawSprite( renderParams, *mpMaterialSSAOApplyHalfRes );

        ID3D11ShaderResourceView * nullSRVs[8] = { NULL };
        dxContext->PSSetShaderResources( 0, 8, nullSRVs );
    }
}

